/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/


import java.io.*;
import java.util.*;

import org.w3c.dom.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.eval.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.prepro.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.gui.*;
//Maxent classes
import opennlp.maxent.*;
import opennlp.maxent.io.*;
//Libsvm classes
import libsvm.svm;
import libsvm.svm_model;
import libsvm.SVMScalingModel;

/**
 * <b>Description:</b><br>
 * This class is the main running class for the GuiTAR tool. It should be called
 * from a command with the desired parameters.<br>
 *
 * @author Philippe J. Goux
 * @version 1.1
 */

public class GTAR_main {
    private static Vector<FileObject> files = new Vector<FileObject>(); //List of file names to use
    private static Vector<String> evaluationTypes = new Vector<String>(); //Types to be used by GTAR_Evaluation class
    private static boolean parsedEvaluation = false; //Parsed attribute to be used by GTAR_Evaluation class
    private static boolean runEvaluation = false; //Run evaluation
    private static boolean transformFileToHTML = false; //Create HTML output
    private static boolean runPreprocess = false; //Run PreProSyntacticHeuristics
    private static boolean logInformation = false; //Run PreProSyntacticHeuristics
    private static boolean verbose = false; //Run PreProSyntacticHeuristics
    private static byte classifierType = IOXMLUtils.NODNCLASSIFIER; //values 0 - no DN classifier (default), 1 - SVM, 2 - MaxEnt
    private static svm_model svmModel = null;
    private static SVMScalingModel svmScalingModel = null;
    private static MaxentModel maxentModel;
    private static ContextGenerator contextGenerator;
    private static String classifierModelFileName = "gnmvpc.libsvm"; //default svm trained on GNOME and VPC

    private static void initialise(String args[]) {
        // Retreive and decrypt parameters
        try {
            retrieveParameters(args);
            Local.out.setVerbose(verbose);
        } catch (ArrayIndexOutOfBoundsException e) {
            Local.err.println("Invalid number of parameters.");
            displayUsage();
            System.exit(1);
        }
        if (files.size() < 1) {
            Local.err.println("No Files were found.");
            displayUsage();
            System.exit(1);
        }

        // Delete output files if necessary
        if (logInformation) {
            if (Local.FILE_LOG_PREPROCESS.exists()) Local.FILE_LOG_PREPROCESS.delete();
            if (Local.FILE_LOG_PROCESS.exists()) Local.FILE_LOG_PROCESS.delete();
        }
        if (runEvaluation) {
            if (Local.FILE_LOG_EVALUATION.exists()) Local.FILE_LOG_EVALUATION.delete();
        }
    }


    private static void retrieveParameters(String args[]) throws ArrayIndexOutOfBoundsException {
        for (int i = 0; i < args.length; i++) {
            // Display Usage and Help
            if (args[i].equalsIgnoreCase("-h") ||
                args[i].equalsIgnoreCase("-help")) {
                displayUsage();
                System.exit(0);
            }
            // Set parse attribute for evaluation
            else if (args[i].equalsIgnoreCase("-version")) {
                Local.out.println(Language.getDescription());
                System.exit(0);
            }
            // Set parse attribute for evaluation
            else if (args[i].equalsIgnoreCase("-parsed")) {
                parsedEvaluation = true;
            }
            // Create HTML output
            else if (args[i].equalsIgnoreCase("-html")) {
                transformFileToHTML = true;
            }
            // Run evaluation and Get types for evaluation
            else if (args[i].equalsIgnoreCase("-evaluate")) {
                runEvaluation = true;
                do {
                    i++;
                    evaluationTypes.add(args[i]);
                } while (i + 1 < args.length && isNextArgParameter(args[i + 1]));
            }
            // Run PreProSyntacticHeuristics
            else if (args[i].equalsIgnoreCase("-preprocess")) {
                runPreprocess = true;
            }
            // Log information
            else if (args[i].equalsIgnoreCase("-log")) {
                logInformation = true;
            }
            // Use output information
            else if (args[i].equalsIgnoreCase("-verbose")) {
                verbose = true;
                while (i + 1 < args.length && isNextArgParameter(args[i + 1])) {
                    i++;
                    Local.out.addLogType(args[i]);
                }
            }
            // Get MASXML filenames
            else if (args[i].equalsIgnoreCase("-i") ||
                     args[i].equalsIgnoreCase("-input")) {
                while (i + 1 < args.length && isNextArgParameter(args[i + 1])) {
                    i++;
                    FileObject file = new FileObject(args[i]);
                    if (!isFile(file)) {
                        System.exit(1);
                    }
                    files.add(file);
                }
            }
//New bit    //Get MASXML filenames from a file
            else if (args[i].equalsIgnoreCase("-f") ||
                     args[i].equalsIgnoreCase("-file")) {
		if (i + 1 < args.length && isNextArgParameter(args[i + 1])) {
                    i++;
                    File file = new File(args[i]);
                    if (!isFile(file)) {
                        System.exit(1);
                    }
		    try {
			    BufferedReader in = new BufferedReader(new FileReader( file ));
			    String fileName = in.readLine();
			    while (fileName != null) {
				FileObject auxFile = new FileObject( fileName );
				if (!isFile( auxFile )) {
					System.out.println("Error! " + fileName + " NOT A FILE!");
	                        	System.exit(1);
				}
				files.add( auxFile );
			    	fileName = in.readLine();
			    }
		    }
		    catch( FileNotFoundException fnfe ) {
			System.out.println("Error! " + fnfe.toString());
                       	System.exit(1);
		    }
		    catch( IOException ioe ) {
			System.out.println("Error! " + ioe.toString());
                       	System.exit(1);
		    }
                } else {
		    //Another alternative is that if no file name has been provided read file names from standard input
		    System.out.println("Error! Invalid file name argument after -f[ile].");
	            //displayUsage(); //The above message should be ideally included in the displayUsage() function
                    System.exit(1);
                }

	    } //New bit ends
	    else if (args[i].equalsIgnoreCase("-maxent")) {
		loadMaxEntClassifier( args[++i] );
	    }
	    else if (args[i].equalsIgnoreCase("-svm")) {
		loadSVMClassifier( args[++i] );
	    }
            // Get tag set
            else if (args[i].equalsIgnoreCase("-t") ||
                     args[i].equalsIgnoreCase("-tagset")) {
                if (i + 1 < args.length && isNextArgParameter(args[i + 1])) {
                    i++;
                    File file = new File(args[i]);
                    if (!isFile(file)) {
                        System.exit(1);
                    }
                    TagSet.setTagSet(file);
                } else {
                    TagSet.setCreateFile(true);
                    Local.out.println("Default tag set created.");
                    System.exit(0);
                }
            }
        }
    }


    private static boolean isFile(File file) {
        if (!file.exists()) {
            Local.err.println("Could not find file:" + Local.NEWLINE + file.getPath());
            return false;
        }
        if (file.isDirectory()) {
            Local.err.println("Could not use directory:" + Local.NEWLINE + file.getPath());
            displayUsage();
            return false;
        }
        return true;
    }


    private static boolean isNextArgParameter(String str) {
        return!(str.startsWith("-"));
    }


    private static void displayUsage() {
        String str = Language.getUsage();
        Local.out.println(str);
    }


    private static void startProcess() {
        // Work on files one by one
	int qCounter=0; //A global counter that counts how many queries have been thrown by processing each file
        for (int i = 0; i < files.size(); i++) {
            // Get file
            FileObject file = files.elementAt(i);

            // Load XML files
            Document domDocument = IOXMLUtils.load(file);

            //Producing MASXML files... (arg type: w0003_tagged.xml)
            if (runPreprocess) {
                if (logInformation)
                    Local.out.setOut(Local.FILE_LOG_PREPROCESS, true, false);
                PreProSyntacticHeuristics ppsh = new PreProSyntacticHeuristics();
                ppsh.process(domDocument, file.getName());
                domDocument = ppsh.getDomDocument();
            }

            // Running GuiTAR on MASXML files...
            if (logInformation)
                Local.out.setOut(Local.FILE_LOG_PROCESS, true, false);
            DiscourseModelImplementer dmi = new DiscourseModelImplementer(domDocument);
	    if( classifierType == IOXMLUtils.SVM )
	            dmi.setClassifier( qCounter, svmModel, svmScalingModel );
	    else if( classifierType == IOXMLUtils.MAXENT )
	            dmi.setClassifier( qCounter, maxentModel, contextGenerator );
	    else
	            dmi.setClassifier(); //No Classifier
            dmi.process();
	    qCounter = dmi.globalQueryCounter; //Update the query counter
            domDocument = dmi.getDOMDocument();

            // Running evaluation...
            if (runEvaluation) {
                Local.out.setOut(Local.FILE_LOG_EVALUATION, true, true);
                AnaphoraResolutionEvaluation are = new AnaphoraResolutionEvaluation();
                if (parsedEvaluation) {
                    File parsedFile = new File(file.getParent() + Local.FILE_SEPARATOR + "Parsed" + file.getName());
                    if (!parsedFile.exists()) {
                        Local.err.println("Could not find file for parsed evaluation:" + Local.NEWLINE + parsedFile.getPath());
                        System.exit(1);
                    }
                    Document parsedDomDocument = IOXMLUtils.load(parsedFile.getPath());
                    if (parsedDomDocument == null)
                        System.exit(1);
                    are.process(domDocument, file.getName(), parsedDomDocument, parsedFile.getName(), evaluationTypes);
                }
                are.process(domDocument, file.getName(), evaluationTypes);
                if (i + 1 >= files.size())
                    are.displayPerformance();
            }

            // Producing html files for error analysis...
            if (transformFileToHTML) {
                Local.out.setOut(file.getHtml(), false, true);
                (new GenerateErrorsInHTML()).transformDomDocument(domDocument, file.getName(), "AAante", "ante");
            }

            // Set back default output stream
            Local.out.reset();

            // Return DOM document to XML form
            IOXMLUtils.save(domDocument, file.getProcessed());
        }
    }

    private static void loadMaxEntClassifier( String fileName ) {
	/*** load the MaxEnt model ***/
	try {
		GISModel m = new SuffixSensitiveGISModelReader( new File( fileName + ".model" )).getModel();
		maxentModel = m;
        } catch (FileNotFoundException fnfe) {
            Local.err.printException(fnfe);
            System.exit(1);
        } catch (IOException ioe) {
            Local.err.printException(ioe);
            System.exit(1);
        } catch (Exception e) {
		Local.err.printException(e);
		System.exit(1);
	}
	contextGenerator = new BasicContextGenerator();
	classifierType = IOXMLUtils.MAXENT;
    }

    private static void loadSVMClassifier( String fileName ) {
        // Load gnmvpc.libsvm.* files
        svmModel = null;
        svmScalingModel = null;
        try {
            svmModel = svm.svm_load_model( fileName + ".model"); //modelFileName = "gnmvpc.libsvm"
            svmScalingModel = new SVMScalingModel( fileName + ".range");
        } catch (FileNotFoundException fnfe) {
            Local.err.printException(fnfe);
            System.exit(1);
        } catch (IOException ioe) {
            Local.err.printException(ioe);
            System.exit(1);
        }
	classifierType = IOXMLUtils.SVM;
    }

    //Main method for testing
    public static void main(String args[]) {
        // Check to see if GUI should be used
        if (args.length < 1) {
            try {
                MainFrame main = new MainFrame();
            } catch (Exception ex) {
                ex.fillInStackTrace();
            }
        } else {
            //Initialise environment depending on arguments inputted by user
            initialise(args);

            // Start processing files
            startProcess();
        }
    }
}
