/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/


package uk.ac.essex.malexa.nlp.dp.GuiTAR;

import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.*;

/**
 * A class that encapsulates the morhological features of a head noun (number,
 * person and gender).
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 1.1
 */
public class Agreement {
    //Constants to store the possible values of the agreement features
    public static final int FIRSTPERSON = 1;
    public static final int SECONDPERSON = 2;
    public static final int THIRDPERSON = 3;
    public static final int SINGULAR = 1;
    public static final int PLURAL = 2;
    public static final int FEMENINE = 1;
    public static final int MASCULINE = -1;
    public static final int IMPERSONAL = 0;
    public static final int ANY = -2;
    public static final int NONE = -3;


    //Instance variables
    private int person,
            number,
            gender;

    /**
     * Default constructor; sets features to 3rd person, singular, impersonal.
     */
    public Agreement() {
        this(Agreement.THIRDPERSON, Agreement.SINGULAR, Agreement.IMPERSONAL);
    }


    /**
     * Receives the agreement features as numeric values and stores them internally.
     * @param pPerson The person as integer
     * @param pNumber The number as integer
     * @param pGender The gender as integer
     */
    public Agreement(int pPerson, int pNumber, int pGender) {
        person = pPerson;
        number = pNumber;
        gender = pGender;
    }


    /**
     * Receives the agreement features as string values and stores them internally.
     * @param pPerson The person as String
     * @param pNumber The number as String
     * @param pGender The gender as String
     */
    public Agreement(String pPerson, String pNumber, String pGender) {
        if (pPerson.equalsIgnoreCase(IOXMLUtils.FIRSTPERSON_PROPERTY_VALUE))
            person = Agreement.FIRSTPERSON;
        else if (pPerson.equalsIgnoreCase(IOXMLUtils.SECONDPERSON_PROPERTY_VALUE))
            person = Agreement.SECONDPERSON;
        else if (pPerson.equalsIgnoreCase(IOXMLUtils.THIRDPERSON_PROPERTY_VALUE))
            person = Agreement.THIRDPERSON;
        else if (pPerson.equalsIgnoreCase(IOXMLUtils.ANY_PROPERTY_VALUE))
            person = Agreement.ANY;
        else
            person = Agreement.NONE;
        if (pNumber.equalsIgnoreCase(IOXMLUtils.SINGULAR_PROPERTY_VALUE))

            number = Agreement.SINGULAR;
        else if (pNumber.equalsIgnoreCase(IOXMLUtils.PLURAL_PROPERTY_VALUE))
            number = Agreement.PLURAL;
        else if (pNumber.equalsIgnoreCase(IOXMLUtils.ANY_PROPERTY_VALUE))
            number = Agreement.ANY;
        else
            number = Agreement.NONE;
        if (pGender.equalsIgnoreCase(IOXMLUtils.FEMENINE_PROPERTY_VALUE))
            gender = Agreement.FEMENINE;
        else if (pGender.equalsIgnoreCase(IOXMLUtils.MASCULINE_PROPERTY_VALUE))
            gender = Agreement.MASCULINE;
        else if (pGender.equalsIgnoreCase(IOXMLUtils.IMPERSONAL_PROPERTY_VALUE))
            gender = Agreement.IMPERSONAL;
        else if (pGender.equalsIgnoreCase(IOXMLUtils.ANY_PROPERTY_VALUE))
            gender = Agreement.ANY;
        else
            gender = Agreement.NONE;
    }


    //Accessors
    public int getPerson() {
        return (person);
    }


    public int getNumber() {
        return (number);
    }


    public int getGender() {
        return (gender);
    }


    //Override the equals method
    public boolean equals(Object obj) {
        if (obj != null) {
            if (obj instanceof Agreement) {
                int oPerson = ((Agreement) obj).getPerson();
                int oNumber = ((Agreement) obj).getNumber();
                int oGender = ((Agreement) obj).getGender();
                return (
                        (person == ANY || oPerson == ANY || person == oPerson) &&
                        (number == ANY || oNumber == ANY || number == oNumber) &&
                        (gender == ANY || oGender == ANY || gender == oGender)
                        );

            } else
                throw new ClassCastException("Agreement != " + obj.getClass().getName());
        } else
            return (false);
    }


    //Override toString method
    public String toString() {
        String p, n, g;
        switch (getPerson()) {
        case FIRSTPERSON:
            p = new String("FirstPerson");
            break;
        case SECONDPERSON:
            p = new String("SecondPerson");
            break;
        case THIRDPERSON:
            p = new String("ThirdPerson");
            break;
        default:
            p = new String("Any");
        } //End person

        switch (getNumber()) {
        case SINGULAR:
            n = new String("Singular");
            break;
        case PLURAL:
            n = new String("Plural");
            break;
        default:
            n = new String("Any");
        } //End number

        switch (getGender()) {
        case FEMENINE:
            g = new String("Femenine");
            break;
        case MASCULINE:
            g = new String("Masculine");
            break;
        case IMPERSONAL:
            g = new String("Impersonal");
            break;
        default:
            g = new String("Any");
        } //End gender

        return (" -Person: " + p + " -Number: " + n + " -Gender: " + g);
    }


    /*** STATIC METHODS ***/

    /**
     * Determines the agreement features using hand coded heuristics based on part-of-speech.
     * @param posTag The part-of-speech
     * @param word The word
     * @return Agreement An object holding the resulting agreement features
     */
    public static Agreement newInstance(String posTag, String word) {
        int person = -3, number = -3, gender = -3;
        if (Local.contains(TagSet.COMMON_NOUN_SINGULAR, posTag)) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.IMPERSONAL;
        } else if (Local.contains(TagSet.COMMON_NOUN_PLURAL, posTag)) {
            person = Agreement.THIRDPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.IMPERSONAL;
        } else if (Local.contains(TagSet.PROPER_NOUN_SINGULAR, posTag)) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.ANY;
            //gender = Agreement.MASCULINE;
            //***A CRITERIA FOR FEMENINE GENDER NEEDED
        } else if (Local.contains(TagSet.PROPER_NOUN_PLURAL, posTag)) {
            person = Agreement.THIRDPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("I") == 0 ||
                   word.compareToIgnoreCase("Me") == 0) { //This was the beginning of personal pronoun block
            person = Agreement.FIRSTPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("You") == 0) {
            person = Agreement.SECONDPERSON;
            number = Agreement.ANY;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("He") == 0 ||
                   word.compareToIgnoreCase("Him") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.MASCULINE;
        } else if (word.compareToIgnoreCase("She") == 0 ||
                   word.compareToIgnoreCase("Her") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.FEMENINE;
        } else if (word.compareToIgnoreCase("It") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.IMPERSONAL;
        } else if (word.compareToIgnoreCase("We") == 0 ||
                   word.compareToIgnoreCase("Us") == 0) {
            person = Agreement.FIRSTPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("They") == 0 ||
                   word.compareToIgnoreCase("Them") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("Mine") == 0) { //This was the end of personal pronoun block, This was the beginning of possessive pronoun block
            person = Agreement.FIRSTPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("Yours") == 0 ||
                   word.compareToIgnoreCase("Your") == 0) {
            person = Agreement.SECONDPERSON;
            number = Agreement.ANY;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("His") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.MASCULINE;
        } else if (word.compareToIgnoreCase("Hers") == 0 ||
                   word.compareToIgnoreCase("Her") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.FEMENINE;
	} else if( word.compareToIgnoreCase( "Its" ) == 0 ) {
	    person = Agreement.THIRDPERSON;
	    number = Agreement.SINGULAR;
	    gender = Agreement.IMPERSONAL;
        } else if (word.compareToIgnoreCase("Ours") == 0 ||
                   word.compareToIgnoreCase("Our") == 0) {
            person = Agreement.FIRSTPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("Theirs") == 0 ||
                   word.compareToIgnoreCase("Their") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("myself") == 0) { //This was the end of possessive pronoun block, This was the beginning of reflexive pronoun block
            person = Agreement.FIRSTPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("yourself") == 0) {
            person = Agreement.SECONDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("himself") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.MASCULINE;
        } else if (word.compareToIgnoreCase("herself") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.FEMENINE;
        } else if (word.compareToIgnoreCase("itself") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.IMPERSONAL;
        } else if (word.compareToIgnoreCase("ourselves") == 0) {
            person = Agreement.FIRSTPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("yourselves") == 0) {
            person = Agreement.SECONDPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("themselves") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.ANY;
        } else if (word.compareToIgnoreCase("This") == 0 //This was the end of reflexive pronoun block
                   || word.compareToIgnoreCase("That") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.IMPERSONAL;
        } else if (word.compareToIgnoreCase("These") == 0
                   || word.compareToIgnoreCase("Those") == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.PLURAL;
            gender = Agreement.IMPERSONAL;
        } else if (Local.contains(TagSet.TOBE_PAST_TENSE_SINGULAR, posTag)) {
            person = Agreement.ANY;
            number = Agreement.SINGULAR;
            gender = Agreement.ANY;
        } else if (Local.contains(TagSet.TOBE_PRESENT_1_SINGULAR, posTag)) {
            person = Agreement.FIRSTPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.ANY;
        } else if (posTag.compareTo(TagSet.VERB_PRESENT_3_SINGULAR[0]) == 0 ||
                   posTag.compareTo(TagSet.VERB_PRESENT_3_SINGULAR[1]) == 0 ||
                   posTag.compareTo(TagSet.VERB_PRESENT_3_SINGULAR[2]) == 0 ||
                   posTag.compareTo(TagSet.VERB_PRESENT_3_SINGULAR[3]) == 0) {
            person = Agreement.THIRDPERSON;
            number = Agreement.SINGULAR;
            gender = Agreement.ANY;
        } else {
            person = Agreement.ANY;
            number = Agreement.ANY;
            gender = Agreement.ANY;
        }

        return (new Agreement(person, number, gender));

    } //End method newInstance


    /* ** MAIN ** */
    public static void main(String args[]) {

        Agreement ag1 = new Agreement(3, 1, Agreement.ANY);
        Agreement ag2 = new Agreement();
        Local.out.println("The value is: " + ag1.equals(ag2));
        Local.out.println("The value is: " + ag1);
        Local.out.println("The value is: " + ag2);
    }


} // End of class Agreement


