/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/


package uk.ac.essex.malexa.nlp.dp.GuiTAR;

import java.util.Vector;
import org.w3c.dom.Node;
import org.w3c.dom.Element;

/**
 * A class that encapsulates the general functionality of a forward-looking centre (Cf). This is an
 * abstract class that enables the creation and manipulation of a polymorphic list of Cfs. It implements
 * the interface AnaphoraResolution, hence forces its subclasses to provide an implementation of the method
 * resolveAnaphor().
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 3.0.1
 */
public abstract class Cf implements AnaphoraResolution {

	public int queryCounter;	//Stores the number of queries for this Cf

	//Instance variables
	private Node		domNode;		//Pointer to the DOM node
	private Utterance	utterance;		//Pointer to parent utterance
	private byte		type;			//The type
	private boolean		anaphoric;		//A flag

	/**
	 * Constructor. Sets the DOM node that holds the Cf, the pointer to the Utterance of the Cf
	 * and the identified type of the Cf (see NominalGroup for types of NPs).
	 * @param theNode The node that holds the Cf
	 * @param theUtterance The Utterance of this Cf
	 * @param theType The type of the Cf
	 */
	public Cf( Node theNode, Utterance theUtterance, byte theType ) {
		setNode( theNode );
		setUtterance( theUtterance );
		setType( theType );
		setAnaphoric( false ); //By default a Cf is considered non-anaphoric
	}


	//Accessors

	/**
	 * Returns the type of the Cf.
	 * @return byte the type of the Cf
	 */
	public byte getType() {
		return type;
	}

	/**
	 * Returns the node holding the Cf.
	 * @return Node the node holding the Cf
	 */
	public Node getNode() {
		return domNode;
	}

	/**
	 * Returns the utterance of the Cf.
	 * @return Utterance The utterance of this Cf
	 */
	public Utterance getUtterance() {
		return utterance;
	}

	/**
	 * Checks whether this Cf is anaphoric or not.
	 * @return boolean true if this Cf is anaphoric, false otherwise
	 */
	public boolean isAnaphoric() {
		return anaphoric;
	}

	/**
	 * Looks for the Cf whose neId of its DOM node is equal to cfId.
	 * @param cfId The cfId of the Cf to be searched for
	 * @param dm The Dicourse Model to be searched
	 * @return Cf The Cf whose neId of its DOM node is equal to cfId
	 */
	public Cf findCfById( String cfId, DiscourseModel dm ) {
		Cf result = null;
		if( cfId != null && cfId.length() > 0 ) {
			Utterance utterance = getUtterance();
			while( utterance != null ) { //loop over utterances backwards
			   Vector vCfs = dm.getCfs( utterance );
			   for( int j = 0; j < vCfs.size(); j++ ) { //loop over Cfs forwards
				Cf cfAnte = ( Cf )vCfs.elementAt( j );
				Node cfNode = cfAnte.getNode();
				String auxId = ((Element)cfNode).getAttribute( IOXMLUtils.ID_PROPERTY_NAME );
				if( auxId.equalsIgnoreCase( cfId ) ) result = cfAnte;
			   } //end inner for
			   utterance = dm.getPrevUtterance( utterance );
			} //end outer loop
		} //end if not null
		return result;
	}

	/**
	 * Returns the parent Cf of this Cf if any.
	 * @param dm The Dicourse Model to be searched
	 * @return Cf The parent Cf of this Cf or null if none
	 */
	public Cf getParentCf( DiscourseModel dm ) {
		Cf result = null;
		Node nodeThisCf = this.getNode();
		Node uttThisCf = this.getUtterance().getNode();
		Node auxResult = null;
		Node parent = IOXMLUtils.getParentNodeByName( nodeThisCf, uttThisCf, IOXMLUtils.NP_TAG );
		while( parent != uttThisCf ) {
			auxResult = parent;
			parent = IOXMLUtils.getParentNodeByName( parent, uttThisCf, IOXMLUtils.NP_TAG );
		}

		if( auxResult != null ) {
			String parentId = ((Element)auxResult).getAttribute( IOXMLUtils.ID_PROPERTY_NAME );
			result = findCfById( parentId, dm ); //a method in class Cf
		}

		return result;
	}


	//Mutators

	/**
	 * Stores the node that holds this Cf.
	 * @param theNode the node that holds this Cf
	 */
	public void setNode( Node theNode ) {
		domNode = theNode;
	}

	/**
	 * Stores the pointer to the Utterance that holds this Cf.
	 * @param utt The Utterance
	 */
	public void setUtterance( Utterance utt ) {
		utterance = utt;
		utterance.addChild( this );
	}

	/**
	 * Stores the identified type this Cf.
	 * @param theType The identified type this Cf
	 */
	public void setType( byte theType ) {
		type = theType;
	}

	/**
	 * Sets the anaphoric attribute of this Cf as given by the parameter flag.
	 * @param flag The parameter that sets this Cf as anaphoric or not (true/false)
	 */
	public void setAnaphoric( boolean flag ) {
		anaphoric = flag;
	}
} //End class Cf
