/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/


package uk.ac.essex.malexa.nlp.dp.GuiTAR;

import java.util.Iterator;
import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Arrays;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;
import com.google.soap.search.GoogleSearch;
import com.google.soap.search.GoogleSearchResult;
import com.google.soap.search.GoogleSearchFault;

//Libsvm classes
import libsvm.svm_node;

import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.Local;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.TagSet;

/**
 THIS CLASS PREPARES TRAINING PATTERNS FOR DISCOURSE NEW CLASSIFICATION/DD Resolution

 * A class that encapsulates the functionality for processing Definite Descriptions (i.e. "the red car").
 * It's mainly meant to provide the implementation for resolving anaphoric definite descriptions,
 * identifying discourse new ones, and resolving bridging references.
 * The algorithms employed to process a definite description are covered in detail in:
 * - Renata Vieira and Massimo Poesio, 2000, "An Empirically-Based System for Processing Definite Descriptions", Computational Linguistics, v. 26, n.4, p. 539-593.
 * - Renata Vieira and Massimo Poesio, 2000, "Corpus-based development and evaluation of a system for processing definite descriptions", Proc. of the 18th COLING, Saarbruecken, August.
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 3.0.1
 */
public class DefiniteDescription extends NominalGroup {

    //It would be good to move these two somewhere so they can be parametricised
    public static final String COPULA_LIST[] = {"appear", "appeared", "appears", "are", "became", "become", "becomes", "been", "is", "remain", "remained", "remains", "seem", "seemed", "seems", "was", "were"};
    public static final int WINDOW_SIZE = 4;

    private Cf antecedent;
    private svm_node[] inputVector;
    private boolean googleFeaturesOn;
    private byte dnClassifierType;

    /**
     * Constructor. Just calls the constructor of NominalGroup.
     */
    public DefiniteDescription(Node theNode, Utterance theUtterance, byte theType, byte dnClassType ) {
        super(theNode, theUtterance, theType);
        queryCounter = 0; //declared in class Cf
	dnClassifierType = dnClassType;
	googleFeaturesOn = false; //default
        //smaller model -- inputVector = new svm_node[19];
	if( dnClassifierType == IOXMLUtils.SVM ) {
	        inputVector = new svm_node[23];
	        for (int i = 0; i < inputVector.length; i++) inputVector[i] = new svm_node();
	}
    }


    /**
     * A gold standard algorithm for processing Definite Descriptions based on corpus annotation.
     * @param dm The Discourse Model to which this Cf belongs
     * @return Cf The antecedent
     */
    //public Cf goldStandardresolveAnaphor( DiscourseModel dm ) {
    public Cf resolveAnaphor(DiscourseModel dm) {
        Cf result = null;
        result = preparePattern(dm); //For training
        setAnaphoric(result != null);
        return result; //This is the proposed antecedent not the real one
    }


    private boolean isEmbedded(Node node, Node topNode) {
        boolean result = false;
        String name = node.getNodeName();
        while (node != null && (node = node.getParentNode()) != topNode) {
            if (node != null && name.equalsIgnoreCase(node.getNodeName())) {
                result = true;
                break;
            }
        }
        return result;
    }


    /**
     * This method extracts a number of input features and sets a targeted output for them (true/false).
     * In other words it prepares a file with training data for Discourse-New Classification.
     * @param anteId The antecedent Id of this DD according to the annotation
     * @param equivClass The Set of Ids of the members of the equivalence class of this DD
     * @param dm The Discourse Model to which this Cf belongs
     */
    private Cf preparePattern(DiscourseModel dm) {
        Cf proposedAntecedent = null;
        byte target = 0;
        byte correctAntecedent = 0;
        String sRealAnteId = "";
        String anteFeatures[] = new String[] {"-1", "-1", "-1", "-1", "-1", "-1", "-1", "-1", "-1"};
        DiscourseEntity deAnaphor = null;
        if (getHeadNode() != null) deAnaphor = dm.getDiscourseEntity(this); //DE corresponding to this Cf
        else {
            Local.out.println("DefiniteDescription", "The DD has no head!");
            return null;
        }
        String ddFeatures[] = null;

	if( dnClassifierType != IOXMLUtils.NODNCLASSIFIER ) {
		ddFeatures = readFeatures();
	        String bkpDDFeatures[] = ddFeatures;
	        if (ddFeatures[4] == null || ddFeatures[4].length() == 0) { //if google features not in XML file collect them
		    googleFeaturesOn = true;
	        }
	        ddFeatures = extractFeatures(deAnaphor);
		if(googleFeaturesOn) storeFeatures( ddFeatures );
		else {
		    ddFeatures[4] = bkpDDFeatures[4];
		    ddFeatures[5] = bkpDDFeatures[5];
		    ddFeatures[6] = bkpDDFeatures[6];
		    ddFeatures[7] = bkpDDFeatures[7];
		    ddFeatures[8] = bkpDDFeatures[8];
		    ddFeatures[9] = bkpDDFeatures[9];
		}
	}
        Utterance utterance = getUtterance();
        //Get the parent Cf of this Cf if any
        Cf thisCf = this; //default
        Cf parentCf = getParentCf(dm); //a method in class Cf
        if (parentCf != null) thisCf = parentCf;
        stop: {
            for (int distance = 0; utterance != null; distance++) { //loop over utterances backwards
                Vector vCfs = dm.getCfs(utterance);
                for (int j = 0; j < vCfs.size(); j++) { //loop over Cfs forwards
                    Cf cfAnte = (Cf) vCfs.elementAt(j);
                    if (cfAnte == thisCf) break; //Only for the first utterance, which holds this cf
                    byte cfType = cfAnte.getType();
                    if (Arrays.binarySearch(NominalGroup.DEFINITE, cfType) >= 0 ||
                        Arrays.binarySearch(NominalGroup.INDEFINITE, cfType) >= 0 ||
                        cfType == NominalGroup.POSS_NP ||
                        cfType == NominalGroup.PERS_PRO ) {
                       //|| cfType == NominalGroup.PN) { // allow only definites, indefinites, possessives and personal pronouns
                        DiscourseEntity deAnte = dm.getDiscourseEntity(cfAnte);
                        if (deAnte.getTypes().containsAll(deAnaphor.getTypes())) { //If Type Properties are compatible
                            //First try to reject the potential antecedent
                            if (distance <= DefiniteDescription.WINDOW_SIZE) { //Within segment
                                if (deAnte.getAttributes().size() > 0 &&
                                    !deAnte.getAttributes().containsAll(deAnaphor.getAttributes())) {
                                    continue; //The antecedent is rejected, so skip it and keep on looking
                                }
                            } else { //outside segment
                                Cf lastNonReducedCf = getLastNonReducedCf(deAnte);
                                if (!isSubsequentMention(deAnte, deAnaphor) && !isIdentical(lastNonReducedCf)) {
                                    continue; //The antecedent is rejected, so skip it and keep on looking
                                }
                            }
                            anteFeatures = extractFeatures(cfAnte, deAnte, deAnaphor, distance);
                            //PASS IT ON TO CLASSIFIER
                            /*** Note: For the SVM classifier pattern is not used at all ***/
                            String pattern = compilePattern(ddFeatures, anteFeatures, sRealAnteId, target, correctAntecedent);
                            //ASK THE ORACLE, i.e. classifier
			    if( dnClassifierType == IOXMLUtils.MAXENT ) {
				Local.out.println("DefiniteDescription", "DDid:" + ddFeatures[0] + ":FEATURES:\n" + pattern);
				double dnProbability = 0;
				if( ( dnProbability = dm.getClassification( pattern ) ) < 0.74 ) { //classified as not DN
					Local.out.println("DefiniteDescription", "***MAXENT:ANTE ACCEPTED! (Pdn=" + dnProbability + ")***");
					proposedAntecedent = cfAnte;
				}
				else
					Local.out.println("DefiniteDescription", "***MAXENT:ANTE REJECTED! (Pdn=" + dnProbability + ")***");
			    }
			    else if( dnClassifierType == IOXMLUtils.SVM ) {
				Local.out.println("DefiniteDescription", "DDid:" + ddFeatures[0] + ":FEATURES:\n" + printLIBSVMInputVector( inputVector) );
				if (dm.getClassification(inputVector) != 0.0) { //if not DN, i.e., not 0, i.e., =1
					Local.out.println("DefiniteDescription", "***SVM:ANTE ACCEPTED!***");
					proposedAntecedent = cfAnte;
                            	}
				else
					Local.out.println("DefiniteDescription", "***SVM:ANTE REJECTED!***");
			    }
			    else  //No Classifier, hence accept the antecedent
				proposedAntecedent = cfAnte;
                            break stop; //stop looping
                        } //end if the heads are equal
                    } //end if cf is a definite, indefinite or possessive
                } //end inner for - Cfs
                utterance = dm.getPrevUtterance(utterance);
            } //end outer for - utterances
        } //end stop
        return proposedAntecedent;
    }


    private String compilePattern(String ddFeatures[], String anteFeatures[], String realAnteId, byte target, byte correctAntecedent) {
	String result = "";
        int directAnaphora = -1;
        String anteId = "";
        if (anteFeatures != null) {
            directAnaphora = Integer.parseInt(anteFeatures[8]); //The distance to the proposed antecedent
            anteId = anteFeatures[0];
        }
        /*** NUMBER OF FEATURES IS 29 (including targeted output) ***/

	if( dnClassifierType == IOXMLUtils.MAXENT ) {
		//Google features preprocessed 
		result = "1=" + directAnaphora +
			" 2=" + ddFeatures[3] +
			" 3=" + (Double.parseDouble( ddFeatures[4] )>=2 ? 2 : (Double.parseDouble( ddFeatures[4] )<=0.5 ? 0 : 1 ) ) +
			" 4=" + (Double.parseDouble( ddFeatures[5] )>=2 ? 2 : (Double.parseDouble( ddFeatures[5] )<=0.5 ? 0 : 1 ) ) +
			" 5=" + (Double.parseDouble( ddFeatures[6] )>=2 ? 2 : (Double.parseDouble( ddFeatures[6] )<=0.5 ? 0 : 1 ) ) +
			" 6=" + (Double.parseDouble( ddFeatures[7] )>=2 ? 2 : (Double.parseDouble( ddFeatures[7] )<=0.5 ? 0 : 1 ) ) +
			" 7=" + (Double.parseDouble( ddFeatures[8] )>=2 ? 2 : (Double.parseDouble( ddFeatures[8] )<=0.5 ? 0 : 1 ) ) +
			" 8=" + (Double.parseDouble( ddFeatures[9] )>=2 ? 2 : (Double.parseDouble( ddFeatures[9] )<=0.5 ? 0 : 1 ) ) +
			" 9=" + ddFeatures[10] +
			" 10=" + ddFeatures[11] +
			" 11=" + ddFeatures[13] +
			" 12=" + ddFeatures[14] +
			" 13=" + ddFeatures[15] +
			" 14=" + ddFeatures[16] +
			" 15=" + ddFeatures[17] +
			" 16=" + ddFeatures[18] +
			" 17=" + ddFeatures[19] +
			" 18=" + anteFeatures[2] +
			" 19=" + anteFeatures[3] +
			" 20=" + anteFeatures[4] +
			" 21=" + anteFeatures[5] +
			" 22=" + anteFeatures[6] +
			" 23=" + anteFeatures[7];
	}
	else if( dnClassifierType == IOXMLUtils.SVM ) { //HERE THE INPUT VECTOR FOR LIBSVM IS FILLED IN WITH VALUES
	        inputVector[0].index = 2;
	        inputVector[0].value = directAnaphora;
	        inputVector[1].index = 3;
	        inputVector[1].value = getDouble(ddFeatures[3]);
	        inputVector[2].index = 4;
	        inputVector[2].value = getDouble(ddFeatures[4]);
	        inputVector[3].index = 5;
	        inputVector[3].value = getDouble(ddFeatures[5]);
	        inputVector[4].index = 6;
	        inputVector[4].value = getDouble(ddFeatures[6]);
	        inputVector[5].index = 7;
	        inputVector[5].value = getDouble(ddFeatures[7]);
	        inputVector[6].index = 8;
	        inputVector[6].value = getDouble(ddFeatures[8]);
	        inputVector[7].index = 9;
	        inputVector[7].value = getDouble(ddFeatures[9]);
	        inputVector[8].index = 10;
	        inputVector[8].value = getDouble(ddFeatures[10]); //This is superlative
	        inputVector[9].index = 11;
	        inputVector[9].value = getDouble(ddFeatures[11]);
	        inputVector[10].index = 12;
	        inputVector[10].value = getDouble(ddFeatures[13]);
	        inputVector[11].index = 13;
	        inputVector[11].value = getDouble(ddFeatures[14]);
	        inputVector[12].index = 14;
	        inputVector[12].value = getDouble(ddFeatures[15]);
	        inputVector[13].index = 15;
	        inputVector[13].value = getDouble(ddFeatures[16]);
	        inputVector[14].index = 16;
	        inputVector[14].value = getDouble(ddFeatures[17]);
	        inputVector[15].index = 17;
	        inputVector[15].value = getDouble(ddFeatures[18]); //not in for smaller model
	        inputVector[16].index = 18;
	        inputVector[16].value = getDouble(ddFeatures[19]); //not in for smaller model
	        inputVector[17].index = 19;
	        inputVector[17].value = getDouble(anteFeatures[2]);
	        inputVector[18].index = 20;
	        inputVector[18].value = getDouble(anteFeatures[3]);
	        inputVector[19].index = 21;
	        inputVector[19].value = getDouble(anteFeatures[4]);
	        inputVector[20].index = 22;
	        inputVector[20].value = getDouble(anteFeatures[5]); //not in for smaller model
	        inputVector[21].index = 23;
	        inputVector[21].value = getDouble(anteFeatures[6]);
	        inputVector[22].index = 24;
	        inputVector[22].value = getDouble(anteFeatures[7]);
	}
	return result;
    }

    /**
     * Get double from string value.
     * @param str String
     * @return double
     */
    private double getDouble(String str) {
        try {
            return Double.parseDouble(str);
        }
        catch(NumberFormatException e) {
            return -1;
        }
    }

    /**
     * Extracts the features for a proposed antecedent.
     * @param cfAnte The Cf holding the proposed antecedent
     * @param deAnte The DiscourseEntity mapped to the Cf of the antecedent
     * @param distance The distance in number of utterances to the antecedent
     * @return The vector of features
     */
    private String[] extractFeatures(Cf cfAnte, DiscourseEntity deAnte, DiscourseEntity deAnaphor, int distance) {
        //0 - false; 1 - true
        String sAnteId = ((Element) cfAnte.getNode()).getAttribute("id");
        byte fDDAnteTypeMatch = 1; //There is a type match by default - not used for the moment
        byte fDDAnteAttrMatch = 0;
        byte fAnteHasAttributes = 0;
        byte fAnteIsSubsequentMention = 0;
        byte fAnteIsEmbeddedNe = 0;
        byte fAnteIsIdentical = 0;
        byte fAnteType = cfAnte.getType();
        if (deAnte.getAttributes().containsAll(deAnaphor.getAttributes())) fDDAnteAttrMatch++;
        if (deAnte.getAttributes().size() > 0) fAnteHasAttributes++;
        //OLD VERSION: if( deAnte.getEquivalenceClass().size() > 1 ) fAnteIsSubsequentMention++;
        if (isSubsequentMention(deAnte, deAnaphor)) fAnteIsSubsequentMention++;
        if (isIdentical(getLastNonReducedCf(deAnte))) fAnteIsIdentical++;
        if (isEmbedded(cfAnte.getNode(), cfAnte.getUtterance().getNode())) fAnteIsEmbeddedNe++;
        int nDDAnteDistance = distance;
        /**** The type match seems not to be a useful feature **********/
        String result[] = new String[] {sAnteId,
                          Byte.toString(fDDAnteTypeMatch),
                          Byte.toString(fDDAnteAttrMatch),
                          Byte.toString(fAnteHasAttributes),
                          Byte.toString(fAnteIsSubsequentMention),
                          Byte.toString(fAnteIsEmbeddedNe),
                          Byte.toString(fAnteIsIdentical),
                          Byte.toString(fAnteType),
                          Integer.toString(nDDAnteDistance)};
        return result;

    }


    /*********************************************************************/
    /******* AN AUXILIARY METHOD TO COMPUTE ONLY CERTAIN FEATURES ********/
    /*********************************************************************/
    private String[] AUXextractFeatures(DiscourseEntity deAnaphor) {
        /**** Predicative NPs ********************/
        //Appositions
        String isAppositive = getIsAppositive(this.getNode(), this.getUtterance().getNode());
        String containsAppositive = getContainsAppositive(this.getNode(), this.getUtterance().getNode());

        String result[] = new String[] {isAppositive, containsAppositive};

        return result;
    }


    /**
     * Extracts the features for this DD.
     * GOOGLE FEATURES HAVE BEEN TURNED OFF LOCALLY AS WELL!
     * @param deAnaphor The DiscourseEntity of this DD
     * @return String[] The vector of features
     */
    private String[] extractFeatures(DiscourseEntity deAnaphor) {
        String sDDId = ((Element) getNode()).getAttribute("id");
        /**** Additional DD features not included in the paper *******/
        byte fDDHasAttributes = 0;
        if (deAnaphor.getAttributes().size() > 0) fDDHasAttributes++;
        byte fDDIsEmbeddedNe = 0;
        if (isEmbedded(getNode(), getUtterance().getNode())) fDDIsEmbeddedNe++;
        NodeList nl = ((Element) getNode()).getElementsByTagName(IOXMLUtils.WORD_TAG);
        int nDDNumberOfWords = nl.getLength();

        /**** Predicative NPs ********************/
        //Appositions
        String isAppositive = getIsAppositive(this.getNode(), this.getUtterance().getNode());
        String containsAppositive = getContainsAppositive(this.getNode(), this.getUtterance().getNode());
        //Copular
        String copular = getCopular(this.getNode(), this.getUtterance().getNode());
        /**** Functionality detection ************/
        //the-head/a(n)-head
        //Iterator it = deAnaphor.getTypes().iterator();
        //String head = ((Property)it.next()).getPredicate().getEntry();
        String head = "";
        Node headNode = this.getHeadNode();
        if (headNode != null) {
            NodeList words = ((Element) headNode).getElementsByTagName(IOXMLUtils.WORD_TAG);
            StringBuffer sb = new StringBuffer();
            Node wordNode = null;
            for (int i = 0; i < words.getLength(); i++) {
                wordNode = words.item(i);
                sb.append(wordNode.getFirstChild().getNodeValue().trim() + " ");
            }
            head = sb.substring(0, sb.length() - 1); //Ignore last white space
        } //end if head is null
        Local.out.println("DefiniteDescription", ">>>Head: " + head);
        double headRatio[] = getDefIndefRatio(head.toLowerCase());
        String features[] = getNPnoDetFirstAdjSuperlative(this.getPreModifiers(), head);
        String npNoDet = features[0];
        //the-NPnoDet/a(n)-NPnoDet, where NPnoDet is the NP with no determiner
        Local.out.println("DefiniteDescription", ">>>NPnoDet: " + npNoDet);
        double nillRatio[] = new double[] {0.0, 0.0};
        double npNoDetRatio[] = ((npNoDet != null && npNoDet.length() > 0) ? (getDefIndefRatio(npNoDet.toLowerCase())) : nillRatio);
        //the-mod/a(n)-mod. For the first adjective only
        Local.out.println("DefiniteDescription", ">>>First Adjective word: " + features[1]);
        double firstAdjRatio[] = ((features[1] != null && features[1].length() > 0) ? (getDefIndefRatio(features[1].toLowerCase())) : nillRatio);
        String superlative = features[2];
        /***** Establishing Relative **************/
        String relative = getRelative(this.getUtterance().getNode());
        /***** Text Position *********************/
        String title = getTitle(this.getNode(), this.getUtterance().getNode());
        String firstPar = getFirstParagraph(this.getNode());
        String firstSent = getFirstSentence();
        /***** Proper Names *********************/
        String cHead = "0";
        if (head.charAt(0) >= 'A' && head.charAt(0) <= 'Z') cHead = "1"; //Capitalised
        String cPremod = "0";
        if (npNoDet != null) {
            StringTokenizer st = new StringTokenizer(npNoDet);
            for (int i = 0; i < st.countTokens() - 1; i++) { //do not consider the head
                String premod = st.nextToken();
                if (premod.charAt(0) >= 'A' && premod.charAt(0) <= 'Z') {
                    cPremod = "1"; //Capitalised
                    break; //stop looping
                }
            }
        }

        String result[] = new String[] {sDDId,
                          isAppositive,
                          containsAppositive,
                          copular,
                          Double.toString(headRatio[0]),
                          Double.toString(headRatio[1]),
                          Double.toString(npNoDetRatio[0]),
                          Double.toString(npNoDetRatio[1]),
                          Double.toString(firstAdjRatio[0]),
                          Double.toString(firstAdjRatio[1]),
                          superlative,
                          relative,
                          title,
                          firstPar,
                          firstSent,
                          cHead,
                          cPremod,
                          Byte.toString(fDDHasAttributes),
                          Byte.toString(fDDIsEmbeddedNe),
                          Integer.toString(nDDNumberOfWords)
        };
        return result;
    }


    /**
     * Receives a vector of premodifiers (<mod> nodes in terms of DOM) and a head of an NP.
     * Forms an array of Strings of three elements:
     * 1. A String of all premodifiers followed by the head (that is the original NP with no determiner),
     * 2. A String of the first adjective word if any,
     * 3. If any of the pre-modifiers is a Superlative than this feature is set to 1, otherwise it is 0.
     * @param preModifiers The vector of pre-modifiers
     * @param head The head of the NP
     * @return String[] The array of Strings
     */
    private String[] getNPnoDetFirstAdjSuperlative(Vector preModifiers, String head) {
        String result[] = new String[3];
        result[2] = "0"; //Default value for superlative
        if (preModifiers.size() > 0) {
            StringBuffer npNoDet = new StringBuffer();
            for (int i = 0; i < preModifiers.size(); i++) {
                Element mod = (Element) preModifiers.elementAt(i);
                NodeList words = mod.getElementsByTagName(IOXMLUtils.WORD_TAG);
                for (int j = 0; j < words.getLength(); j++) {
                    Node wordNode = words.item(j);
                    String strMod = wordNode.getFirstChild().getNodeValue().trim();
                    npNoDet.append(strMod + " ");
                    String strPOS = ((Element) wordNode).getAttribute(TagSet.POS_PROPERTY_NAME);
                    if ( Local.containsIgnoreCase(TagSet.ADJECTIVE_SUPERLATIVE, strPOS) || Local.containsIgnoreCase(TagSet.ADVERB_SUPERLATIVE, strPOS) ) result[2] = "1"; //set superlative
                    else {
                        if (i == 0 & j == 0) //only for the first word
                            if (Local.containsIgnoreCase(TagSet.ADJECTIVE_GENERAL, strPOS)) result[1] = strMod;
                    }

                } //end inner loop
            } //end outer loop
            result[0] = npNoDet.toString().trim() + " " + head;
        } //end if premodified

        return result;
    }


    /**
     * Accesses the google API and returns two ratios:
     * 1. #"the phrase"/#"a(n) phrase"
     * 2. #"the phrase"/#"phrase"
     * @param phrase The string to be used to form the "the, a(n), and bare" queries
     * @return double[] The two-element array of doubles containing the two ratios
     */
    private double[] getDefIndefRatio(String phrase) {
        double result[] = new double[2];
        double theQ = 0, aQ = 0, bareQ = 0;
        String the_Query = "\"the " + phrase + "\"";
        String bare_Query = "\"" + phrase + "\"";
        String a_Query = "";
        char firstCharacter = phrase.charAt(0);
        if (firstCharacter == 'A' || firstCharacter == 'a' || firstCharacter == 'O' || firstCharacter == 'o' || firstCharacter == 'E' || firstCharacter == 'e' || firstCharacter == 'I' || firstCharacter == 'i')
            a_Query = "\"an " + phrase + "\"";
        else
            a_Query = "\"a " + phrase + "\"";
        /** send to google */
        /** the_Query, a_Query, bare_Query **/
        Local.out.println("DefiniteDescription", ">>>>> The Query: " + the_Query);
        Local.out.println("DefiniteDescription", ">>>>> A(n) Query: " + a_Query);
        Local.out.println("DefiniteDescription", ">>>>> Bare Query: " + bare_Query);
        // Create a Google Search object, set authorization key

	if(googleFeaturesOn) {
          GoogleSearch gs = new GoogleSearch();
          gs.setKey( TagSet.GOOGLE_KEY );

          int faultCounter = 0;
          while( faultCounter < 2 ) { //safe loop
           try {
                   gs.setQueryString( the_Query );
            queryCounter++; //declared in class Cf
         Local.out.println("DefiniteDescription", "DD:qCounter: " + queryCounter);
            GoogleSearchResult gsr = gs.doSearch();
            theQ = ( double )gsr.getEstimatedTotalResultsCount();
                   gs.setQueryString( a_Query );
            queryCounter++; //declared in class Cf
         Local.out.println("DefiniteDescription", "DD:qCounter: " + queryCounter);
                   gsr = gs.doSearch();
            aQ = ( double )gsr.getEstimatedTotalResultsCount();
                   gs.setQueryString( bare_Query );
            queryCounter++; //declared in class Cf
         Local.out.println("DefiniteDescription", "DD:qCounter: " + queryCounter);
                   gsr = gs.doSearch();
            bareQ = ( double )gsr.getEstimatedTotalResultsCount();
            break; //exit safe loop
           } catch ( GoogleSearchFault f ) {
            faultCounter++;
            Local.out.println("DefiniteDescription", "The call to the Google Web APIs failed:");
            Local.out.println("DefiniteDescription",  f.toString() );
           }
          } //end safe loop
	} //end if googleFeaturesOn

        Local.out.println("DefiniteDescription", "theQ=" + theQ + " aQ=" + aQ + " bareQ=" + bareQ);
        //Avoid division by zero
        aQ = (aQ == 0 ? 0.001 : aQ);
        bareQ = (bareQ == 0 ? 0.001 : bareQ);
        result[0] = theQ / aQ;
        result[1] = theQ / bareQ;

        return result;
    }


    /**
     * Checks whether the dd is an appositive.
     * @param ddNode The DOM node holding the DD
     * @param utterance The utterance node
     * @return String "1" is appositive "0" otherwise
     */
    private String getIsAppositive(Node ddNode, Node utterance) {
        String result = "0"; //Default value
        Node parent = IOXMLUtils.getParentNodeByName(ddNode, utterance, IOXMLUtils.UNIT_TAG);
        if (parent != utterance) {
            String unitType = ((Element) parent).getAttribute("utype");
            //check if unit is paren-app or appositive
            if (unitType.equalsIgnoreCase("paren-app")
                || unitType.equalsIgnoreCase("appositive"))
                result = "1";
        }

        return result;
    }


    /**
     * Checks whether the dd contains an appositive.
     * @param ddNode The DOM node holding the DD
     * @param utterance The utterance node
     * @return String "1" is appositive "0" otherwise
     */
    private String getContainsAppositive(Node ddNode, Node utterance) {
        String result = "0"; //Default value
        //Any child UNIT?
        NodeList nl = ((Element) ddNode).getElementsByTagName(IOXMLUtils.UNIT_TAG);
        for (int i = 0; i < nl.getLength(); i++) {
            String unitType = ((Element) nl.item(i)).getAttribute("utype");
            if (unitType.equalsIgnoreCase("paren-app")
                || unitType.equalsIgnoreCase("appositive")) {
                result = "1";
                break;
            }
        }

        return result;
    }


    /**
     * Checks whether the dd is part of a copular construction.
     * @param dd The DOM node holding the DD
     * @param utterance The utterance node
     * @return String "1" DD to the left of copula, "2" DD to the right, "0" otherwise
     */
    private String getCopular(Node dd, Node utterance) {
        String result = "0"; //Default value
        //Look for copula to the right
        String verb = IOXMLUtils.getRightCollocate(dd, utterance);
        if (verb != null && Arrays.binarySearch(COPULA_LIST, verb) >= 0) result = "1";
        else {
            //Look for copula to the left
            verb = IOXMLUtils.getLeftCollocate(dd, utterance);
            if (verb != null && Arrays.binarySearch(COPULA_LIST, verb) >= 0) result = "2";
        }

        return result;
    }


    /**
     * Checks whether the dd is postmodified by a relative clause or a PP.
     * @param utterance The utterance node
     * @return String "0" not postmodified, "1" post- PP, "2" post- relative clause, "3" other postmodifier
     */
    private String getRelative(Node utterance) {
        String result = "0";
        Vector postMod = this.getPostModifiers();
        if (postMod.size() > 0) { //if postmodified
            Node head = this.getHeadNode();
            Element postmod = (Element) postMod.elementAt(0); //The first post-modifier
            NodeList nl = postmod.getElementsByTagName(IOXMLUtils.WORD_TAG);
            Node firstWordAfterHead = nl.item(0);
            String theWord = firstWordAfterHead.getFirstChild().getNodeValue().trim().toLowerCase();
            String thePOS = ((Element) firstWordAfterHead).getAttribute(TagSet.POS_PROPERTY_NAME);
            if (Local.containsIgnoreCase(TagSet.COMMA, thePOS) || Local.containsIgnoreCase(TagSet.PUNCTUATION, thePOS)) {
                firstWordAfterHead = nl.item(1);
                if (firstWordAfterHead == null)return result;
                theWord = firstWordAfterHead.getFirstChild().getNodeValue().trim().toLowerCase();
                thePOS = ((Element) firstWordAfterHead).getAttribute(TagSet.POS_PROPERTY_NAME);
            }
            Vector relativePronouns = new Vector();
            relativePronouns.add("that");
            relativePronouns.add("who");
            relativePronouns.add("which");
            if (Local.containsIgnoreCase(TagSet.PREPOSITION, thePOS)) result = "1";
            else if (relativePronouns.contains(theWord)) result = "2";
            else if (!Local.containsIgnoreCase(TagSet.POSSESSIVE, thePOS)) result = "3";
            /** Supposedly 3 is for any other post-mod, then why the possessive constraint?
                ANSWER: There were problems with poss-np using apostrophe, so they are excluded from postmodification**/
        }

        return result;
    }


    /**
     * Checks whether the DD is part of a Title node.
     * @param ddNode The DOM node holding the DD
     * @param utterance The utterance node
     * @return String "1" it is, "0" otherwise
     */
    private String getTitle(Node ddNode, Node utterance) {
        String result = "0"; //Default value
        Node parent = IOXMLUtils.getParentNodeByName(ddNode, utterance, IOXMLUtils.TITLE_TAG);
        if (parent.getNodeName().equalsIgnoreCase(IOXMLUtils.TITLE_TAG)) result = "1";
        return result;
    }


    /**
     * Checks whether the DD is part of the first paragraph of the Document.
     * @param ddNode The DOM node holding the DD
     * @return String "1" it is, "0" otherwise
     */
    private String getFirstParagraph(Node ddNode) {
        String result = "0"; //Default value
        //Get paragraph node in a similar way as the title
        Node rootNode = ddNode.getOwnerDocument().getDocumentElement();
        Node parent = IOXMLUtils.getParentNodeByName(ddNode, rootNode, IOXMLUtils.PARAGRAPH_TAG);
        //If it has a previous sibling paragraph node set to 0, otherwise to 1
        if (parent != rootNode) {
            Node prevSibling = parent.getPreviousSibling();
            if (prevSibling == null || !prevSibling.getNodeName().equalsIgnoreCase(IOXMLUtils.PARAGRAPH_TAG)) result = "1";
        }
        return result;
    }


    /**
     * Checks whether the DD is part of the first sentence of the Document.
     * @return String "1" it is, "0" otherwise
     */
    private String getFirstSentence() {
        String result = "0"; //Default value
        //Get the utterance node of this dd
        Utterance utterance = this.getUtterance();
        Node uttNode = utterance.getNode();
        //Get the previous utterance
        Segment prevSegment = utterance.getPrevSibling();
        if (prevSegment != null) {
            if (prevSegment instanceof Utterance) {
                Utterance prevUtterance = (Utterance) prevSegment;
                Node prevUttNode = prevUtterance.getNode();
                Node rootNode = uttNode.getOwnerDocument().getDocumentElement();
                Node uttDivNode = IOXMLUtils.getParentNodeByName(uttNode, rootNode, IOXMLUtils.DIV_TAG);
                Node prevUttDivNode = IOXMLUtils.getParentNodeByName(prevUttNode, rootNode, IOXMLUtils.DIV_TAG);
                //if both are in the same <div> -> 0, otherwise 1
                if (uttDivNode != prevUttDivNode) result = "1";
            }
        } else
            result = "1";

        return result;
    }


    /**
     * A method that reads the DN classification features from the XML file.
     * @return String [] The array of values of the DN classification features
     */
    private String[] readFeatures() {
        Element thisDD = (Element) getNode();
        String result[] = new String[20];
        //These are the original indexes of the features
        //result[3] = thisDD.getAttribute( "Cop" );
        result[4] = thisDD.getAttribute("theHaH");
        result[5] = thisDD.getAttribute("theHH");
        result[6] = thisDD.getAttribute("theYaY");
        result[7] = thisDD.getAttribute("theYY");
        result[8] = thisDD.getAttribute("theAaA");
        result[9] = thisDD.getAttribute("theAA");
        //result[11] = thisDD.getAttribute( "Rel" );
        //result[13] = thisDD.getAttribute( "FstP" );
        //result[14] = thisDD.getAttribute( "FstS" );
        //result[15] = thisDD.getAttribute( "cH" );
        //result[16] = thisDD.getAttribute( "cP" );
        return result;
    }


    /**
     * A method that stores all the DN classification features to the XML file.
     * @param ddFeatures The array containing the values of the remaining features
     */
    private void storeFeatures(String[] ddFeatures) {
        /*********** STORING THE FEATURES TO THE XML FILE ***********/
        Element thisDD = (Element) getNode();
        //	thisDD.setAttribute( "Cop", ddFeatures[3] );
        thisDD.setAttribute("theHaH", ddFeatures[4]);
        thisDD.setAttribute("theHH", ddFeatures[5]);
        thisDD.setAttribute("theYaY", ddFeatures[6]);
        thisDD.setAttribute("theYY", ddFeatures[7]);
        thisDD.setAttribute("theAaA", ddFeatures[8]);
        thisDD.setAttribute("theAA", ddFeatures[9]);
        //	thisDD.setAttribute( "Rel", ddFeatures[11] );
        //	thisDD.setAttribute( "FstP", ddFeatures[13] );
        //	thisDD.setAttribute( "FstS", ddFeatures[14] );
        //	thisDD.setAttribute( "cH", ddFeatures[15] );
        //	thisDD.setAttribute( "cP", ddFeatures[16] );
        /*********** END ***********/
    }


    /***************************************************************/
    /******** Methods from the original VP algorithm ***************/
    /***************************************************************/


    /**
     * Implements the subsequent mention heuristic.
     * Checks whether the potential antecedent is a subsequent mention (anaphoric) itself. If it is
     * then the same heuristics as if the antecedent were within the same segment of the given DD are applied
     * plus a check for compatible post-modification (if both, DD and a potential antecedent,
     * are post modified, then their postmodification should be identical in order to accept the antecedent).
     * @param deAnte The DiscourseEntity of the potential antecedent
     * @param deAnaphor The DiscourseEntity of the anaphor
     * @return boolean True if the check succeeds, false otherwise
     */
    private boolean isSubsequentMention(DiscourseEntity deAnte, DiscourseEntity deAnaphor) {
        boolean result = false;
        if (deAnte.getEquivalenceClass().size() > 1) {
            //Apply the same check as within segment
            result = (deAnte.getAttributes().size() == 0 || deAnte.getAttributes().containsAll(deAnaphor.getAttributes())
                      && deAnte.getRelations().containsAll(deAnaphor.getRelations()));
        }
        return result;
    }


    /**
     * Implements the identical NP heuristic.
     * Checks whether the potential antecedent is an identical NP, including the article, to the DD.
     * It converts both, the antecedent NP and the DD, to normalised strings (all letters are
     * lowercase and one white space between words) and casts the problem to string comparison.
     * @param antecedent The Cf of the potential antecedent
     * @return boolean True if the check succeeds, false otherwise
     */
    private boolean isIdentical(Cf antecedent) {
        boolean result = false; //enabled by default
        if (antecedent != null) {
            String ante = IOXMLUtils.getPhraseString((Element) antecedent.getNode());
            String dd = IOXMLUtils.getPhraseString((Element) getNode());
            result = dd.equalsIgnoreCase(ante);
        }
        return result;
    }


    private Cf getLastNonReducedCf(DiscourseEntity deAnte) {
        Cf result = null;
        Vector cfs = deAnte.getEquivalenceClass();
        for (int i = cfs.size() - 1; i >= 0; i--) { //go backwards
            Cf auxCf = (Cf) cfs.elementAt(i);
            if (Arrays.binarySearch(NominalGroup.PRONOUNS, auxCf.getType()) < 0) { //if not a pronoun
                result = auxCf;
                break;
            }
        }
        return result;
    }

    private String printLIBSVMInputVector( svm_node[] iArray ) {
	String result="";
	for(int i=0; i<iArray.length; i++ ) result = result + iArray[i].index + ":" + iArray[i].value + " ";
	return result;
    }


} //End class DefiniteDescription
