/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/


package uk.ac.essex.malexa.nlp.dp.GuiTAR;

import java.util.*;

import org.w3c.dom.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.*;

//import uk.ac.essex.malexa.nlp.dp.GuiTAR.la.ImplementerWSD;
//import uk.ac.essex.malexa.nlp.dp.GuiTAR.la.WordSenseDisambiguator;


/**
 * A class that encapsulates the general functionality of a Discourse Entity.
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 1.2
 */
public class DiscourseEntity {

    //Instance variables
    private Integer id; //A unique identifier of this DE
    private Set<Property> types; //set of heads used to refer to this entity (set of Property objects)
    private Set<Property> attributes; //set of attributes (set of Property objects type=attribute)
    private Set<Property> relations; //set of relations to other DEs (set of Property objects type=relation)
    private Set<Set<Property>> names; //set of names to be used by PN and The_PN(set of Property objects type=name)

    //private Set		kind;		//set of kinds such as company, person...(set of Property objects type=kind)
    private Vector<Cf> equivalenceClass; //The co-reference chain (list of Cf objects)
    private DiscourseModel dModel; //Pointer to the parent DiscourseModel

    //private WordSenseDisambiguator wsd;	//Provides disambiguation service


    /**
     * Discourse Entity constructor. Evokes a new entity in a Discourse Model.
     * @param identifier The identifier of this entity
     * @param cf The Cf evoking this entity
     * @param dm The pointer to the parent Discourse Model
     */
    public DiscourseEntity(Integer identifier, Cf cf, DiscourseModel dm) {
        //wsd = new ImplementerWSD();
        id = identifier;
        dModel = dm;
        types = new HashSet<Property>();
        attributes = new LinkedHashSet<Property>(); // Keep order of inputed elements
        relations = new HashSet<Property>();
        names = new HashSet<Set<Property>>();
        equivalenceClass = new Vector<Cf>();
        addCf(cf);
    }


    //Accessors

    /**
     * Gets the id of this Discourse Entity.
     * @return Integer The id
     */
    public Integer getId() {
        return id;
    }


    /**
     * Gets the heads used to refer to this Discourse Entity.
     * Returns a Set of objects of class Property.
     * @return Set The Set of heads
     */
    public Set<Property> getTypes() {
        return types;
    }


    /**
     * Gets the attributes of this Discourse Entity.
     * Returns a Set of objects of class Property.
     * @return Set The Set of attributes
     */
    public Set<Property> getAttributes() {
        return attributes;
    }


    /**
     * Gets the relations of this Discourse Entity with other Discourse Entities.
     * Returns a Set of objects of class Property.
     * @return Set The Set of relations
     */
    public Set<Property> getRelations() {
        return relations;
    }


    /**
     * Gets the names of this Discourse Entity.
     * Returns a Set of objects of class Property.
     * @return Set The Set of names
     */
    public Set<Set<Property>> getNames() {
        return names;
    }


    /**
     * Gets the co-reference chain (list of Cfs) of this Discourse Entity.
     * Returns a Set of objects of class Cf.
     * @return Set The Set of Cfs
     */
    public Vector getEquivalenceClass() {
        return equivalenceClass;
    }


    //Mutators

    /**
     * A method that updates the state of a Discourse Entity. It stands for an "access" of the DE, and
     * hence it is called after anaphora resolution has taken place. It expects a Cf object as a parameter and
     * then updates the coreference chain of this DE, and reads head and modifiers of this Cf to update this information of the DE.
     * @param cf The new Cf (realisation of this entity)
     */
    private void addCf(Cf cf) {
        equivalenceClass.add(cf);
        if (Arrays.binarySearch(NominalGroup.PRONOUNS, cf.getType()) < 0) { //if not a pronoun
            Property head = null;
            Set<Property> attrs = null;
            Set<Property> rels = null;
            Set<Set<Property>> nms = null;
            if (cf instanceof NominalGroup) {
                Local.out.print("DiscourseEntity", "\n- Creating Discourse Entity for:");
                Local.out.println("DiscourseEntity", " \"" + IOXMLUtils.getPhraseString((Element) cf.getNode()) + "\"");

                NominalGroup np = (NominalGroup) cf;
                head = getHead(np); //Get head
                attrs = getAttributes(np); //Get attrs
                nms = getNames(np, head, attrs); //Get nms
                /*** Still to do: post-modifiers as well as some premodifiers involve embedded NEs,
                     which should be treated as Property.RELATION ***/

                Local.out.println("DiscourseEntity", "\t- Head:\t\t\t\t" + head);
                Local.out.println("DiscourseEntity", "\t- Byte Type:\t\t" + cf.getType());
                Local.out.println("DiscourseEntity", "\t- Type:\t\t\t\t" + NominalGroup.getStringType(cf.getType()));
                if (attrs.size() > 0)
                    Local.out.println("DiscourseEntity", "\t- Attribute Set:\t" + attrs);
                if (nms.size() > 0)
                    Local.out.println("DiscourseEntity", "\t- Name Set:\t\t\t" + nms);
            }
            addHead(head);
            addAttributes(attrs);
            addRelations(rels);
            addNames(nms);
        } //end if not a pronoun
    }


    /**
     * Merges a given Discourse Entity with this one. Adds the types, attributes, and relations
     * of the former to the latter, and updates the Discourse Model by deleting the former and
     * updating the necessary data structures.
     * @param de The Discourse Entity to be merged into this one
     * @param cfAnte The Cf proposed as an antecedent by the resolveAnaphor() method
     */
    public void merge(DiscourseEntity de, Cf cfAnte) {
        types.addAll(de.getTypes());
        attributes.addAll(de.getAttributes());
        relations.addAll(de.getRelations());
        names.addAll(de.getNames());
        equivalenceClass.add(cfAnte);
    }


    /**
     * Adds a new (if new) head to this Discourse Entity.
     * @param head The head of a Cf
     */
    private void addHead(Property head) {
        if (head != null) types.add(head);
    }


    /**
     * Adds (if new) attribute(s) to this Discourse Entity.
     * @param attrs The Set of attributes to be added
     */
    private void addAttributes(Set<Property> attrs) {
        if (attrs != null)
            attributes.addAll(attrs);
    }


    /**
     * Adds (if new) relation(s) to this Discourse Entity.
     * @param rels The Set of relations to be added
     */
    private void addRelations(Set<Property> rels) {
        if (rels != null)
            relations.addAll(rels);
    }


    /**
     * Adds (if new) name(s) to this Discourse Entity.
     * @param nms The Set of names to be added
     */
    private void addNames(Set<Set<Property>> nms) {
        if (nms != null)
            names.addAll(nms);
    }


    /**
     * Reads the necessary information from a given NP and creates a Property object for its head.
     * @param np The NP to be processed
     * @return Property An object holding the head of the NP
     */
    private Property getHead(NominalGroup np) {
        Node headNode = np.getHeadNode();
        if (headNode != null) {
            NodeList words = ((Element) headNode).getElementsByTagName(IOXMLUtils.WORD_TAG);
            StringBuffer sb = new StringBuffer();
            Node wordNode = null;
            for (int i = 0; i < words.getLength(); i++) {
                wordNode = words.item(i);
                sb.append(wordNode.getFirstChild().getNodeValue().trim().toLowerCase() + " ");
            }
            String strHead = sb.substring(0, sb.length() - 1); //Ignore last white space
            String pos = ((Element) wordNode).getAttribute(TagSet.POS_PROPERTY_NAME);
            String lemma = ((Element) wordNode).getAttribute(IOXMLUtils.LEXEME_PROPERTY_NAME);
            //return ( new Property( Property.TYPE, wsd.chooseSense( strHead, lemma, pos ) ) );
            return (new Property(Property.TYPE, new LexicalWord(strHead, pos)));
        } //end if head not null
        return null; //only if head is null
    }


    /**
     * Reads the premodifiers from a given NP and creates Property objects as attributes for every premodifier
     * not part of an embedded NE. Embedded NEs are to be treated as relations to other Discourse Entities.
     * @param np The NP to be processed
     * @return Set A Set of Property Objects; one for every premodifier (attribute)
     */
    private Set<Property> getAttributes(NominalGroup np) {
        Set<Property> result = new LinkedHashSet<Property>();
        Vector preModifiers = np.getPreModifiers();
        if (preModifiers.size() > 0) {
            for (int i = 0; i < preModifiers.size(); i++) {
                Node mod = (Node) preModifiers.elementAt(i);
                //Skip embedded NEs
                Vector words = IOXMLUtils.getElementsByTagNameNoEmbedding(mod, IOXMLUtils.NP_TAG, IOXMLUtils.WORD_TAG);
                for (int j = 0; j < words.size(); j++) {
                    Node wordNode = (Node) words.elementAt(j);
                    String strMod = wordNode.getFirstChild().getNodeValue().trim().toLowerCase();
                    String strPOS = ((Element) wordNode).getAttribute(TagSet.POS_PROPERTY_NAME);
                    String lemma = ((Element) wordNode).getAttribute(IOXMLUtils.LEXEME_PROPERTY_NAME);
                    if (Local.containsIgnoreCase(TagSet.VALID_ATTR_TAGS, strPOS)) {
                        //result.add( new Property( Property.ATTRIBUTE, wsd.chooseSense( strMod, lemma, strPOS ) ) );
                        result.add(new Property(Property.ATTRIBUTE, new LexicalWord(strMod, strPOS)));
                    }
                } //end inner loop
            } //end outer loop
        } //end if premodified
        return result;
    }


    /**
     * Reads the premodifiers from a given NP and creates Property objects as names for every premodifier
     * not part of an embedded NE. Embedded NEs are to be treated as relations to other Discourse Entities.
     * @param np The NP to be processed
     * @param head An object holding the head of the NP
     * @param attributes Set A Set of Property Objects; one for every premodifier (attribute)
     * @return Set A Set of Property Objects; one for every name
     */
    private Set<Set<Property>> getNames(NominalGroup np, Property head, Set<Property> attributes) {
        Set<Property> nms = new LinkedHashSet<Property>(); // Keep order of inputed elements
        if (NominalGroup.PN == np.getType() || NominalGroup.THE_PN == np.getType()) {
            if (attributes != null && attributes.size() > 0) {
                for (Iterator<Property> i = attributes.iterator(); i.hasNext(); ) {
                    Property attribute = i.next();
                    LexicalWord word = (LexicalWord) attribute.getPredicate();
                    if (Local.containsIgnoreCase(TagSet.VALID_PROPER_NOUN_TAGS, word.getPOSTag())) {
                        nms.add(new Property(Property.NAME, new LexicalWord(word.getWord(), word.getPOSTag())));
                    }
                }
                if (head != null) {
                    LexicalWord word = (LexicalWord) head.getPredicate();
                    nms.add(new Property(Property.NAME, new LexicalWord(word.getWord(), word.getPOSTag())));
                }
            } else {
                // Get Cf phrase
                //Vector<Node> phrase = IOXMLUtils.getElementsByTagNameNoEmbedding((Element) np.getNode(), IOXMLUtils.NP_TAG, IOXMLUtils.WORD_TAG);
                Vector<Node> phrase = IOXMLUtils.getElementsByTagNameNoEmbedding((Element) np.getNode(), "", IOXMLUtils.WORD_TAG);
                boolean isLastPN = false;
                Property antecedant = null;
                // Search each word in phrase
                for (Iterator<Node> i = phrase.iterator(); i.hasNext(); ) {
                    Node wordNode = (Node) i.next();
                    String strMod = wordNode.getFirstChild().getNodeValue().trim().toLowerCase();
                    String strPOS = ((Element) wordNode).getAttribute(TagSet.POS_PROPERTY_NAME);

                    // If current word is a proper noun, store it and continue
                    if (Local.containsIgnoreCase(TagSet.VALID_PROPER_NOUN_TAGS, strPOS)) {
                        if (antecedant != null) {
                            nms.add(antecedant);
                            antecedant = null;
                        }
                        nms.add(new Property(Property.NAME, new LexicalWord(strMod, strPOS)));
                        isLastPN = true;
                    }
                    // Else If last word was a PN, continue searching
                    else if (isLastPN) {
                        antecedant = new Property(Property.NAME, new LexicalWord(strMod, strPOS));
                        isLastPN = false;
                    }
                    // Else If current word is not a proper nouns
                    // but some have have already been found but, exit loop
                    else if (Local.containsIgnoreCase(TagSet.COMMA, strPOS) ||
                            !isLastPN && nms.size() > 0)
                        break;
                }
            }
        }
        Set<Set<Property>> results = new HashSet<Set<Property>>();
        if (nms.size() > 0)
            results.add(nms);
        return results;
    }

} //end class DiscourseEntity
