/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR;

import java.util.Set;
import java.util.Map;
import java.util.Vector;

import org.w3c.dom.Document;

import libsvm.svm_node;

/**
 * This interface defines the API of the Discourse Model employed by GuiTAR.
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 3.0.1
 */
public interface DiscourseModel {

    //Accessors

    /**
     * Classifies a definite description as Discourse-new or Anaphoric given a vector of input feature values.
     * The classification is performed using a previously trained maximum entropy model.
     * 93% of GNOME and VPC corpora were used for training.
     * @param pattern The vector of input feature values (each feature-value pair separated by a space)
     * @return The resulting classification (i.e. 0.0 or 1.0, meaning discourse-new or anaphoric respectively)
     */
    public double getClassification(String pattern);


    /**
     * Classifies a definite description as Discourse-new or Anaphoric given a vector of input feature values.
     * The classification is performed using a previously trained SVM (support vector machine) model via the
     * LIBSVM package.
     * 93% of GNOME and VPC corpora were used for training.
     * @param inputVector The vector of input feature values
     * @return The resulting classification (i.e. 0.0 or 1.0, meaning discourse-new or anaphoric respectively)
     */
    public double getClassification(svm_node[] inputVector);


    /**
     * Checks whether the referential expression provided is anaphoric.
     * @param refId The id of a referential expression
     * @return boolean True if the NP is anaphoric, false otherwise
     */
    public boolean isAnaphoric(Object refId);


    /**
     * Retrieves the equivalence class to which refId belongs.
     * If none it returns null.
     * @param refId The id of a referential expression
     * @return Set The set of refIds, members of the equivalence class being retrieved
     */
    public Set getEquivalenceClass(Object refId);


    /**
     * Retrieves the equivalence class corresponding to classIndex.
     * If an invalid classIndex has been provided it returns null.
     * @param classIndex The index of the equivalence class to be returned
     * @return Set The set of refIds, members of the equivalence class being retrieved
     */
    public Set getEquivalenceClass(int classIndex);


    /**
     * Retrieves the Discourse Entity, which cf is a realization of.
     * If no such Discourse Entity, it returns null.
     * @param cf The Cf for which a DE is to be retrieved
     * @return DiscourseEntity The Discourse Entity corresponding to cf
     */
    public DiscourseEntity getDiscourseEntity(Cf cf);


    /**
     * Returns the number of mentions of a given Discourse Entity, that is
     * number of Cfs in its equivalence class.
     * @param de The Discourse Entity
     * @return int The number of mentions
     */
    public int getTimesMentioned(DiscourseEntity de);


    /**
     * Retrieves the index of the equivalence class to which refId belongs.
     * If there is no matching class, it returns -1.
     * @param refId The id of a referential expression
     * @return int The index of the equivalence class to which refId belongs
     */
    public int getEquivalenceClassIndex(Object refId);


    /**
     * Retrieves the refId of the antecedent of anaphorId.
     * @param anaphorId The id of a referential expression used anaphorically
     * @return Object The id of the expression which is coreferential with the one identified by anaphorId
     */
    public Object getAntecedent(Object anaphorId);


    /**
     * Returns the forward-looking centers within the given utterance.
     * @param utt The Utterance
     * @return Vector The list of Cfs within the utterance
     */
    public Vector<Cf> getCfs(Utterance utt);


    /**
     * Returns the utterance following the reference utterance in this Discourse Model.
     * @param uttRef The reference utterance
     * @return Utternace The next utterance
     */
    public Utterance getNextUtterance(Utterance uttRef);


    /**
     * Returns the utterance preceeding the reference utterance in this Discourse Model.
     * @param uttRef The reference utterance
     * @return Utternace The previous utterance
     */
    public Utterance getPrevUtterance(Utterance uttRef);


    /**
     * Returns the distance between the referential expression identified by refId1, and the one
     * identified by refId2. The distance can be measured in terms of number of intermediate words (tagName=W),
     * number of intermediate NEs (tagName=ne) or number of intermediate utterances
     * @param refId1 The id of a referential expression
     * @param refId2 The id of a referential expression
     * @param tagName The tag name of the nodes to be accounted for between refId1 and refId2
     * @return int The distance between the anaphor and the antecedent
     */
    public int getDistance(Object refId1, Object refId2, String tagName);


    /**
     * Returns the number of entities contained in this discourse model.
     * @return int the number of entities in the discourse model
     */
    public int getNumberOfEntities();


    /**
     * Returns the number of (anaphoric) referential expressions in the discourse model.
     * @return int the number of anaphoric references
     */
    public int getNumberOfAnaphoricReferences();


    /**
     * Returns the set of anaphors stored in this discourse model.
     * @return Set The set of anaphors
     */
    public Set getSetOfAnaphors();


    /**
     * Returns a vector of vectors in which this referential expression features.
     * (Every vector of those vectors matches a specific corpus NE)
     * @param refId An Id of a referentail expression
     * @return Vector The vector of vectors in which this referential expression features
     */
    public Vector getVectors(Object refId);


    /**
     * Returns the data structure (Map) that holds the anaphor-to-antecedent mappings.
     * @return Map The anaphor-to-antecedent mappings
     */
    public Map getAnaphorToAntecedentMap();


    /**
     * Returns all the members of the corresponding equivalence class separated by commas.
     * @param refId The id of a referential expression by which an equivalence class will be retrieved
     * @return String Co-referential expressions separated by commas
     */
    public String getEquivalenceClassString(Object refId);


    /**
     * Returns all the members of the corresponding equivalence class separated by commas.
     * @param index The index of the equivalence class to be retrieved
     * @return String
     */
    public String getEquivalenceClassString(int index);


    /**
     * Computes P/R per equivalence class on the basis of class intersection and with reference
     * to this discourse model. Equivalence classes which did not match any of the corpus classes are printed
     * off at the end.
     * @param dm The Discourse Model with which this dicourse model will be intersected
     */
    public void printEquivalenceClassesStatistics(DiscourseModel dm);


    /**
     * Converts the Discourse Model contained in this object into a String.
     * @return String A String representation of this Discourse Model
     */
    public String toString();


    /**
     * Returns the root segment of the physical Discourse Model.
     * @return Segment The root segment
     */
    public Segment getRootSegment();


    /**
     * Returns a flat Discourse Model built from the annotation.
     * (Used by the gold standard algorithm)
     * @return DiscourseModel The flat DM
     */
    public DiscourseModel getFlatDM();


    /**
     * Returns the pointer to the DOM Document of this DiscourseModel.
     * (Used by the gold standard algorithm)
     * @return Document The DOM Document
     */
    public Document getDOMDocument();


    //Setters

    /**
     * Inserts a new anaphoric link into the discourse model.
     * In the flat version of the discourse model, which is constructed out of the annotation,
     * anaphorId and anteId are usually Strings (i.e. ne237), whereas in the full version of the discourse model,
     * which is to be constructed incrementally, they are object references of type Cf.
     * @param anaphorId The id of the anaphor
     * @param anteId The id of the antecedent of the anaphor
     * @param relation The type of relation that holds between the anaphor and the antecedent (ident, poss-inv, etc.)
     */
    public void setAnaphoricLink(Object anaphorId, Object anteId, String relation);


    /**
     * Inserts a new anaphoric link into the discourse model.
     * This is a default version which assumes an "ident" relationship between the anaphor and the antecedent.
     * @param anaphorId The id of the anaphor
     * @param anteId The id of the antecedent of the anaphor
     */
    public void setAnaphoricLink(Object anaphorId, Object anteId);


    /**
     * Sets the pointer to a flat Discourse Model built from the annotation.
     * (Used by the gold standard algorithm)
     * @param dm The flat DM
     */
    public void setFlatDM(DiscourseModel dm);


} //end DiscourseModel interface
