/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR;

//For DOM Tree manipulation
import java.util.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.Local;
import org.w3c.dom.*;
import uk.ac.essex.malexa.nlp.NLPEvaluationMeasures;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

//Maxent classes
import opennlp.maxent.*;
import opennlp.maxent.io.*;
//Libsvm classes
import libsvm.svm;
import libsvm.svm_node;
import libsvm.svm_model;
import libsvm.SVMScalingModel;
import libsvm.SVMScaling;

/**
 * This class provides an implementation for GuiTAR's Discourse Model interface.
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 3.0.3
 */
public class DiscourseModelImplementer implements DiscourseModel {
    //instance variables
    private Vector vecEquivalenceClasses; //A vector of Sets for the coreference classes
    private Map<Object, Integer> mapEquivalenceClasses; //neId -> index in the vector of coreference classes (above)
    private Map<Object, Object> mapAnaphorToAntecedent; //anaphorId -> antecedentId
    private Map<String, Element> mapIdToNode; //for efficiency - neId -> DOM Node - Only for flat DM
    private Map mapNodeToSegment; //Keeps track of what segments/utterances have been constructed already - only for incremental DM
    private Map<String, Vector<Vector<String>>> mapNeIdToVectors; //Vector nex -> Vector {nex, ney, nez}, {ner, nex, nep}
    //Used only for constructing a DM of aligned nes
    private Segment rootSegment; //The root Segment of the physical Discourse Model
    private Document domDocument; //The main pointer to the DOM tree

    private DiscourseModel flatDM; //This only used for the gold standard and it holds a pointer to a flat DM built from annotation

    /******* CLASSIFIERS ********/
    MaxentModel maxentModel;
    ContextGenerator contextGenerator;
    svm_model svmModel;
    SVMScalingModel svmScalingModel;
    byte dnClassifierType;
    /******* QUERY COUNTER ********/
    public int globalQueryCounter;

    //Constructors
    /**
     * Constructs a discourse model for a specific type of NP with reference to an annotation.
     * @param document The DOM document containing the annotation
     * @param anteTag The tag name of the nodes containing the annotation (i.e. ante)
     * @param catAttributeName The attribute name which stores the type of NP (i.e. cat, AAcat)
     * @param typeOfNP The type(s) of NP to be considered (the-np, pers-pro, etc.)
     */
    public DiscourseModelImplementer(Document document, String anteTag,
                                     String catAttributeName, Vector typeOfNP) {
        initialiseDataStructures();
        constructEquivalenceClasses(document, anteTag, catAttributeName, typeOfNP);
    }


    /**
     * Constructs a discourse model for a specific type of NP with reference to an annotation.
     * An annotation generated automatically is aligned with a corpus one.
     * @param documentCorpus The DOM document containing the Corpus annotation
     * @param documentParsed The DOM document containing the automatically generated annotation
     * @param anteTag The tag name of the nodes containing the annotation (i.e. ante)
     * @param catAttributeName The attribute name which stores the type of NP (i.e. cat, AAcat)
     * @param typeOfNP The type(s) of NP to be considered (the-np, pers-pro, etc.)
     */
    public DiscourseModelImplementer(Document documentCorpus, Document documentParsed,
                                     String anteTag, String catAttributeName, Vector typeOfNP) {
        initialiseDataStructures();
        mapNeIdToVectors = new HashMap<String, Vector<Vector<String>>>(); //For NE alignment
        constructEquivalenceClasses(documentCorpus, documentParsed, anteTag, catAttributeName, typeOfNP);
    }


    /**
     * Constructs a discourse model for all types of NP with reference to an annotation.
     * @param document The DOM document containing the annotation
     * @param anteTag The tag name of the nodes containing the annotation (i.e. ante)
     */
    public DiscourseModelImplementer(Document document, String anteTag) {
        initialiseDataStructures();
        constructEquivalenceClasses(document, anteTag);
    }


    /**
     * Constructs a discourse model incrementally (GuiTAR's main processing).
     * @param document Document
     * @param queryCounter int
     * @param svmModel svm_model
     * @param svmScalingModel SVMScalingModel
     */
    public DiscourseModelImplementer(Document document, int queryCounter, svm_model svmModel, SVMScalingModel svmScalingModel) {
        initialiseDataStructures();
        setDomDocument(document);
        // Set SVM
        setClassifier(queryCounter, svmModel, svmScalingModel);
    }

    /**
     * Constructs a discourse model incrementally (GuiTAR's main processing).
     * @param document Document
     * @param queryCounter int The google query counter
     * @param MaxentModel model The maxent model
     */
    public DiscourseModelImplementer(Document document, int queryCounter, MaxentModel model, ContextGenerator cg ) {
        initialiseDataStructures();
        setDomDocument(document);
        // Set MaxEnt
        setClassifier(queryCounter, model, cg );
    }

    /**
     * Constructs a discourse model incrementally (GuiTAR's main processing).
     * @param document Document
     * @param queryCounter int The google query counter
     */
    public DiscourseModelImplementer(Document document, int queryCounter ) {
	//queryCounter is a dummy parameter just to make this method's signature unique
        initialiseDataStructures();
        setDomDocument(document);
        // Set No DN classifier
        setClassifier();
    }

    /**
     * Constructs a discourse model incrementally (GuiTAR's main processing).
     * @param domDoc Document
     */
    public DiscourseModelImplementer(Document domDoc) {
        initialiseDataStructures();
        mapNodeToSegment = new HashMap(); //For Segments and Utterances
        setDomDocument(domDoc);
    }


    private void setDomDocument(Document domDoc) {
        this.domDocument = domDoc;
    }


    private void setDomDocument(String inputFileName) {
        Document domDoc = IOXMLUtils.load(inputFileName); //Load XML file
        setDomDocument(domDoc);
    }


    /**
     * Initialises the data structures of the Discourse Model.
     */
    private void initialiseDataStructures() {
        vecEquivalenceClasses = new Vector();
        mapEquivalenceClasses = new HashMap<Object, Integer>();
        mapAnaphorToAntecedent = new HashMap<Object, Object>();
        mapNodeToSegment = new HashMap(); //For Segments and Utterances
        mapIdToNode = new HashMap<String, Element>(); //For efficiency in retrieving the DOM node out of a neId
    }


    //Accessors
    /**
     * Classifies a definite description as Discourse-new or Anaphoric given a vector of input feature values.
     * The classification is performed using a previously trained maximum entropy model.
     * The complete GNOME and VPC corpora were used for training.
     * @param pattern The vector of input feature values (each feature-value pair separated by a space)
     * @return The resulting classification (i.e. 0.0 or 1.0, meaning discourse-new or anaphoric respectively)
     */
    public double getClassification(String pattern) {
        double[] ocs = maxentModel.eval( contextGenerator.getContext( pattern) );
        return ocs[0];
    }


    /**
     * Classifies a definite description as Discourse-new or Anaphoric given a vector of input feature values.
     * The classification is performed using a previously trained LIBSVM model (support vector machine).
     * The complete GNOME and VPC corpora were used for training.
     * @param inputVector The vector of input feature values
     * @return The resulting classification (i.e. 0.0 or 1.0, meaning discourse-new or anaphoric respectively)
     */
    public double getClassification(svm_node[] inputVector) {
        SVMScaling.performScaling(svmScalingModel, inputVector); //First scale the vector
        return (svm.svm_predict(svmModel, inputVector)); //Return predicted class
    }


    /**
     * Checks whether the referential expression provided is anaphoric.
     * @param refId The id of a referential expression
     * @return boolean True if the NP is anaphoric, false otherwise
     */
    public boolean isAnaphoric(Object refId) {
        return (mapAnaphorToAntecedent.containsKey(refId));
    }


    /**
     * Retrieves the equivalence class to which refId belongs.
     * If none it returns null.
     * @param refId The id of a referential expression
     * @return Set The set of refIds, members of the equivalence class being retrieved
     */
    public Set getEquivalenceClass(Object refId) {
        Set result = null;
        int classIndex = getEquivalenceClassIndex(refId);
        if (classIndex >= 0 && classIndex < vecEquivalenceClasses.size())
            result = (Set) vecEquivalenceClasses.elementAt(classIndex);
        return result;
    }


    /**
     * Retrieves the equivalence class corresponding to classIndex.
     * If an invalid classIndex has been provided it returns null.
     * @param classIndex The index of the equivalence class to be returned
     * @return Set The set of refIds, members of the equivalence class being retrieved
     */
    public Set getEquivalenceClass(int classIndex) {
        Set result = null;
        if (classIndex >= 0 && classIndex < vecEquivalenceClasses.size())
            result = (Set) vecEquivalenceClasses.elementAt(classIndex);
        return result;
    }


    /**
     * Retrieves the Discourse Entity, which cf is a realization of.
     * If no such Discourse Entity, it returns null.
     * @param cf The Cf for which a DE is to be retrieved
     * @return DiscourseEntity The Discourse Entity corresponding to cf
     */
    public DiscourseEntity getDiscourseEntity(Cf cf) {
        DiscourseEntity result = null;
        int classIndex = getEquivalenceClassIndex(cf);
        if (classIndex >= 0 && classIndex < vecEquivalenceClasses.size())
            result = (DiscourseEntity) vecEquivalenceClasses.elementAt(classIndex);
        return result;
    }


    /**
     * Returns the number of mentions of a given Discourse Entity, that is
     * number of Cfs in its equivalence class.
     * @param de The Discourse Entity
     * @return int The number of mentions
     */
    public int getTimesMentioned(DiscourseEntity de) {
        return (de.getEquivalenceClass().size());

    }


    /**
     * Retrieves the index of the equivalence class to which refId belongs.
     * If there is no matching class, it returns -1.
     * @param refId The id of a referential expression
     * @return int The index of the equivalence class to which refId belongs
     */
    public int getEquivalenceClassIndex(Object refId) {
        int result = -1;
        Integer indx = (Integer) mapEquivalenceClasses.get(refId);
        if (indx != null) result = indx.intValue();
        return result;
    }


    /**
     * Retrieves the refId of the antecedent of anaphorId.
     * @param anaphorId The id of a referential expression used anaphorically
     * @return Object The id of the expression which is coreferential with the one identified by anaphorId
     */
    public Object getAntecedent(Object anaphorId) {
        return (mapAnaphorToAntecedent.get(anaphorId));
    }


    /**
     * Returns the forward-looking centers within the given utterance.
     * @param utt The Utterance
     * @return Vector The list of Cfs within the utterance
     */
    public Vector getCfs(Utterance utt) {
        return (utt.getChildren());
    }


    /**
     * Returns the utterance following the reference utterance in this Discourse Model.
     * @param uttRef The reference utterance
     * @return Utternace The next utterance
     */
    public Utterance getNextUtterance(Utterance uttRef) {
        /*** NOT NEEDED FOR THE MOMENT ***/
        return null;
    }


    /**
     * Returns the utterance preceeding the reference utterance in this Discourse Model.
     * @param uttRef The reference utterance
     * @return Utternace The previous utterance
     */
    public Utterance getPrevUtterance(Utterance uttRef) {
        Segment result = uttRef.getPrevSibling();
        Segment parent = uttRef.getParent();
        while (result == null) {
            result = parent.getPrevSibling();
            parent = parent.getParent();
            if (parent == null)break; //reached the root
        } while (result != null && !(result instanceof Utterance)) { //if not an utterance
            Vector children = result.getChildren();
            result = (Segment) children.lastElement(); //Go down the tree always getting the rightmost child
        }
        return ((Utterance) result);
    }


    /**
     * Returns the distance between the referential expression identified by refId1, and the one
     * identified by refId2. The distance can be measured in terms of number of intermediate words (tagName=W),
     * number of intermediate NEs (tagName=ne) or number of intermediate utterances
     * @param refId1 The id of a referential expression
     * @param refId2 The id of a referential expression
     * @param tagName The tag name of the nodes to be accounted for between refId1 and refId2
     * @return int The distance between the anaphor and the antecedent
     */
    public int getDistance(Object refId1, Object refId2, String tagName) {
        Node node1 = IOXMLUtils.getElementById((String) refId1, mapIdToNode);
        Node node2 = IOXMLUtils.getElementById((String) refId2, mapIdToNode);

        return (IOXMLUtils.getNumberOfIntermediateNodes(node1, node2, tagName));
    }


    /**
     * Returns the number of entities contained in this discourse model.
     * @return int the number of entities in the discourse model
     */
    public int getNumberOfEntities() {
        return (vecEquivalenceClasses.size());
    }


    /**
     * Returns the number of (anaphoric) referential expressions in the discourse model.
     * @return int the number of anaphoric references
     */
    public int getNumberOfAnaphoricReferences() {
        return (mapAnaphorToAntecedent.size());
    }


    /**
     * Returns the set of anaphors stored in this discourse model.
     * @return Set The set of anaphors
     */
    public Set<Object> getSetOfAnaphors() {
        return (mapAnaphorToAntecedent.keySet());
    }


    /**
     * Returns a vector of vectors in which this referential expression features.
     * (Every vector of those vectors matches a specific corpus NE)
     * @param refId An Id of a referentail expression
     * @return Vector The vector of vectors in which this referential expression features
     */
    public Vector<Vector<String>> getVectors(Object refId) {
        return (mapNeIdToVectors.get(refId));
    }


    /**
     * Returns the data structure (Map) that holds the anaphor-to-antecedent mappings.
     * @return Map The anaphor-to-antecedent mappings
     */
    public Map<Object, Object> getAnaphorToAntecedentMap() {
        return mapAnaphorToAntecedent;
    }


    /**
     * Returns all the members of the corresponding equivalence class separated by commas.
     * @param refId The id of a referential expression by which an equivalence class will be retrieved
     * @return String Co-referential expressions separated by commas
     */
    public String getEquivalenceClassString(Object refId) {
        String result = "";
        Set eqClass = getEquivalenceClass(refId);
        if (eqClass != null) {
            Iterator it = eqClass.iterator();
            while (it.hasNext()) result = new String(result + it.next() + ",");
        }
        return result;
    }


    /**
     * Returns all the members of the corresponding equivalence class separated by commas.
     * @param index The index of the equivalence class to be retrieved
     * @return String
     */
    public String getEquivalenceClassString(int index) {
        String result = "";
        Set eqClass = getEquivalenceClass(index);
        if (eqClass != null) {
            for (Iterator it = eqClass.iterator(); it.hasNext(); )
                result = new String(result + it.next() + ",");
        }
        return result;
    }


    /**
     * Computes P/R per equivalence class on the basis of class intersection and with reference
     * to this discourse model. Equivalence classes which did not match any of the corpus classes are printed
     * off at the end.
     * @param dm The Discourse Model with which this dicourse model will be intersected
     */
    public void printEquivalenceClassesStatistics(DiscourseModel dm) {
        Local.out.println(Local.NEWLINE + "C. Precision and Recall per Equivalence Class:");
        int classCounter = 1;
        Set<Integer> printedAlgorithmClasses = new HashSet<Integer>();
        //Copy anaphor-to-antecedent map into a new map
        Map<Object, Object> anaphoricNEsCopy = new HashMap<Object, Object>();
        Set keys = this.getSetOfAnaphors();
        for (Iterator it = keys.iterator(); it.hasNext(); ) {
            Object key = it.next();
            Object value = this.getAntecedent(key);
            anaphoricNEsCopy.put(key, value);
        }

        //Loop over the original anaphoric NEs
        while (!anaphoricNEsCopy.isEmpty()) {
            Object anaphorNeId = null;
            anaphorNeId = anaphoricNEsCopy.keySet().iterator().next(); //get any remaining anaphoric NE
            int index = this.getEquivalenceClassIndex(anaphorNeId); //Get equivalence class index
            Local.out.println("-------------------------------------------------");
            Local.out.println("Coreference chain " + classCounter);
            Local.out.println("From corpus," + "\"" + this.getEquivalenceClassString(index) + "\"");
            Set thisEqClass = this.getEquivalenceClass(index); //Get equivalence class
            Set dmEqClass = null;
            Set<Integer> intersectingClasses = new HashSet<Integer>();
            for (Iterator it = thisEqClass.iterator(); it.hasNext(); ) {
                Object classMember = it.next();
                int classIndex = dm.getEquivalenceClassIndex(classMember);
                if (classIndex >= 0) { //Is there a corresponding equivalence class from the algorithm
                    printedAlgorithmClasses.add(new Integer(classIndex)); //update printed classes set
                    dmEqClass = dm.getEquivalenceClass(classIndex); //Get equivalence class
                    //Print class if not already printed and
                    if (!intersectingClasses.contains(new Integer(classIndex))) {
                        Local.out.println("From algorithm," + "\"" + dm.getEquivalenceClassString(classIndex) + "\"");
                        Local.out.println("Precision," + NLPEvaluationMeasures.precision(dmEqClass, thisEqClass));
                        Local.out.println("Recall," + NLPEvaluationMeasures.recall(dmEqClass, thisEqClass));
                        Local.out.println("F Measure," + NLPEvaluationMeasures.fMeasure(dmEqClass, thisEqClass));
                        intersectingClasses.add(new Integer(classIndex));
                    }
                }
                anaphoricNEsCopy.remove(classMember); //delete class member from original map for anaphoric NEs
            } //end inner while
            classCounter++;
        } //end outer while
        //Print unprinted equivalence classes from algorithm (if any)
        if (printedAlgorithmClasses.size() != dm.getNumberOfEntities()) {
            Local.out.println("-------------------------------------------------");
            Local.out.println("Unprinted (so far) equivalence classes from algorithm:");
            for (int i = 0; i < dm.getNumberOfEntities(); i++) {
                if (!printedAlgorithmClasses.contains(new Integer(i)))
                    Local.out.println("class " + i + ":," +
                                      "\"" + dm.getEquivalenceClassString(i) + "\"");
            }
        }
    }


    /**
     * Converts the Discourse Model contained in this object into a String.
     * @return String A String representation of this Discourse Model
     */
    public String toString() {
        StringBuffer strBuff = new StringBuffer();
        strBuff.append(Local.NEWLINE + "-- DISCOURSE MODEL BEGINS --" + Local.NEWLINE + Local.NEWLINE);
        strBuff.append("Equivalence Classes:" + Local.NEWLINE);
        for (int i = 0; i < vecEquivalenceClasses.size(); i++) {
            strBuff.append(">> Equivalence Class -> " + i + " : " +
                           ((Set) vecEquivalenceClasses.elementAt(i)).toString() + "" + Local.NEWLINE);
        }

        strBuff.append("Direct Anaphoric Links (antecedent <- anaphor):" + Local.NEWLINE + Local.NEWLINE);
        Iterator it = this.getSetOfAnaphors().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            Object value = this.getAntecedent(key);
            strBuff.append(">> " + value + " <- " + key + "" + Local.NEWLINE);
        }
        strBuff.append(Local.NEWLINE + "-- DISCOURSE MODEL ENDS --" + Local.NEWLINE);

        return (strBuff.toString());
    }


    /**
     * Returns the root segment of the physical Discourse Model.
     * @return Segment The root segment
     */
    public Segment getRootSegment() {
        return rootSegment;
    }


    /**
     * Returns a flat Discourse Model built from the annotation.
     * (Used by the gold standard algorithm)
     * @return DiscourseModel The flat DM
     */
    public DiscourseModel getFlatDM() {
        return flatDM;
    }


    /**
     * Returns the pointer to the DOM Document of this DiscourseModel.
     * (Used by the gold standard algorithm)
     * @return Document The DOM Document
     */
    public Document getDOMDocument() {
        return domDocument;
    }


    //Setters

    /*** The functionality to support several relations has not been implemented yet, hence at the moment
     the third parameter is unused.
     Data structures must be modified for that purpose ***/
    /**
     * Inserts a new anaphoric link into the discourse model.
     * In the flat version of the discourse model, which is constructed out of the annotation,
     * anaphorId and anteId are usually Strings (i.e. ne237), whereas in the full version of the discourse model,
     * which is to be constructed incrementally, they are object references of type Cf.
     * @param anaphorId The id of the anaphor
     * @param anteId The id of the antecedent of the anaphor
     * @param relation The type of relation that holds between the anaphor and the antecedent (ident, poss-inv, etc.)
     */
    public void setAnaphoricLink(Object anaphorId, Object anteId,
                                 String relation) {
        Integer classIndex = (Integer) mapEquivalenceClasses.get(anteId);
        Set<Object> eqClass = new HashSet<Object>();
        if (classIndex == null) {
            int newIndex = vecEquivalenceClasses.size();
            classIndex = new Integer(newIndex);
            eqClass.add(anteId);
            vecEquivalenceClasses.add(eqClass);
            mapEquivalenceClasses.put(anteId, classIndex);
        } else
            eqClass = (Set<Object>) vecEquivalenceClasses.elementAt(classIndex.intValue());
        eqClass.add(anaphorId);
        mapEquivalenceClasses.put(anaphorId, classIndex);
        mapAnaphorToAntecedent.put(anaphorId, anteId);
    }


    /**
     * Inserts a new anaphoric link into the discourse model.
     * This is a default version which assumes an "ident" relationship between the anaphor and the antecedent.
     * @param anaphorId The id of the anaphor
     * @param anteId The id of the antecedent of the anaphor
     */
    public void setAnaphoricLink(Object anaphorId, Object anteId) {
        this.setAnaphoricLink(anaphorId, anteId, "ident");
    }


    /**
     * Creates a new Discourse Entity for the given Cf.
     * @param cf The Cf
     */
    public void createDiscourseEntity(Cf cf) {
        int newIndex = vecEquivalenceClasses.size();
        Integer classIndex = new Integer(newIndex);
        DiscourseEntity discourseEntity = new DiscourseEntity(classIndex, cf, this);
        vecEquivalenceClasses.add(discourseEntity);
        mapEquivalenceClasses.put(cf, classIndex);
    }


    /**
     * Sets the pointer to a flat Discourse Model built from the annotation.
     * (Used by the gold standard algorithm)
     * @param dm The flat DM
     */
    public void setFlatDM(DiscourseModel dm) {
        flatDM = dm;
    }


    /***************************************************************/
    /********************* PRIVATE METHODS *************************/
    /***************************************************************/


    /**
     * Builds the Discourse Model out of the annotation provided in the XML file.
     * This version of the method is tailored for a specific type of NP, which is given as a parameter.
     * @param document The DOM document containing the annotation
     * @param anteTag The tag name of the nodes containing the annotation (i.e. ante)
     * @param catAttributeName The attribute name which stores the type of NP (i.e. cat, AAcat)
     * @param typeOfNP The type(s) of NP to be considered (the-np, pers-pro, etc.)
     */
    private void constructEquivalenceClasses(Document document, String anteTag, String catAttributeName, Vector typeOfNP) {
        NodeList anteList = document.getElementsByTagName(anteTag);
        Local.out.println("antelist: " + anteList.getLength());
        for (int i = 0; i < anteList.getLength(); i++) {
            Element ante = (Element) anteList.item(i);
            String relation = ante.getAttribute(IOXMLUtils.RELATION_PROPERTY_NAME);
            //Local.out.println("relation " + i + ": " + relation);
            if (relation.equalsIgnoreCase("ident")) {
                String anaphorNeId = ante.getAttribute(IOXMLUtils.CURRENT_PROPERTY_NAME);
                NodeList aux = ante.getElementsByTagName(IOXMLUtils.ANCHOR_TAG);
                Element anchor = (Element) aux.item(0); //Though there might be more than one anchors for ambiguous anaphors
                String anteNeId = anchor.getAttribute(IOXMLUtils.ANTECEDENT_PROPERTY_NAME);
                Integer classIndex = (Integer) mapEquivalenceClasses.get(anteNeId);
                Set eqClass = new HashSet();
                if (classIndex == null) {
                    int newIndex = vecEquivalenceClasses.size();
                    classIndex = new Integer(newIndex);
                    eqClass.add(anteNeId);
                    vecEquivalenceClasses.add(eqClass);
                    mapEquivalenceClasses.put(anteNeId, classIndex);
                } else
                    eqClass = (Set) vecEquivalenceClasses.elementAt(classIndex.intValue());
                eqClass.add(anaphorNeId);
                if (mapEquivalenceClasses.containsKey(anaphorNeId))
                    Local.out.println("***Overwriting: " + anaphorNeId + ", Class index: " + classIndex);
                mapEquivalenceClasses.put(anaphorNeId, classIndex);
                if (isAnaphorOfType(document, anaphorNeId, typeOfNP, catAttributeName)) { //check if the anaphor is of the desired type
                    if (mapAnaphorToAntecedent.containsKey(anaphorNeId))
                        Local.out.println("*******Trying to overwrite: " + anaphorNeId + ", antecedent" + anteNeId);
                    else
                        mapAnaphorToAntecedent.put(anaphorNeId, anteNeId);
                }
            } //end if rel="ident"
        } //end for
    }


    /*** NEW - For aligned nes ***/
    /**
     * Builds the Discourse Model out of the annotation provided in the XML file.
     * This version of the method is tailored for a specific type of NP, which is given as a parameter.
     * Additionally a translation from the neIds in the corpus file to the neIds in the parsed file is performed.
     * @param documentCorpus The DOM document containing the Corpus annotation
     * @param documentParsed The DOM document containing the automatically generated annotation
     * @param anteTag The tag name of the nodes containing the annotation (i.e. ante)
     * @param catAttributeName The attribute name which stores the type of NP (i.e. cat, AAcat)
     * @param typeOfNP The type(s) of NP to be considered (the-np, pers-pro, etc.)
     */
    private void constructEquivalenceClasses(Document documentCorpus, Document documentParsed, String anteTag, String catAttributeName, Vector typeOfNP) {
        Map neIdTranslatorMap = NeAlignment.constructCorpusIdToParsedId(documentCorpus, documentParsed);
        NodeList anteList = documentCorpus.getElementsByTagName(anteTag);
        Local.out.println("antelist: " + anteList.getLength());
        for (int i = 0; i < anteList.getLength(); i++) {
            Element ante = (Element) anteList.item(i);
            String relation = ante.getAttribute(IOXMLUtils.RELATION_PROPERTY_NAME);
            //Local.out.println("relation " + i + ": " + relation);
            if (relation.equalsIgnoreCase("ident")) {
                String anaphorNeId = ante.getAttribute(IOXMLUtils.CURRENT_PROPERTY_NAME);
                NodeList aux = ante.getElementsByTagName(IOXMLUtils.ANCHOR_TAG);
                Element anchor = (Element) aux.item(0); //Though there might be more than one anchors for ambiguous anaphors
                String anteNeId = anchor.getAttribute(IOXMLUtils.ANTECEDENT_PROPERTY_NAME);
                //Translate
                Vector vecAnaphorNeId = (Vector) neIdTranslatorMap.get(anaphorNeId);
                Vector vecAnteNeId = (Vector) neIdTranslatorMap.get(anteNeId);
                if (vecAnaphorNeId == null || vecAnteNeId == null) {
                    Local.out.println("*** Problems finding a matching neId!");
                    Local.out.println("anaphorNeId: " + vecAnaphorNeId + ", anaphorNeIdCorpus:" + anaphorNeId);
                    Local.out.println("anteNeId: " + vecAnteNeId + ", anteNeIdCorpus:" + anteNeId);
                } else {
                    updateNeToVectorsMap(vecAnaphorNeId);
                    updateNeToVectorsMap(vecAnteNeId);
                    Integer classIndex = (Integer) mapEquivalenceClasses.get(vecAnteNeId);
                    Set eqClass = new HashSet();
                    if (classIndex == null) {
                        int newIndex = vecEquivalenceClasses.size();
                        classIndex = new Integer(newIndex);
                        eqClass.addAll(vecAnteNeId);
                        vecEquivalenceClasses.add(eqClass);
                        mapEquivalenceClasses.put(vecAnteNeId, classIndex);
                    } else
                        eqClass = (Set) vecEquivalenceClasses.elementAt(classIndex.intValue());
                    eqClass.addAll(vecAnaphorNeId);
                    if (mapEquivalenceClasses.containsKey(vecAnaphorNeId))
                        Local.out.println("***Overwriting: " + vecAnaphorNeId + ", Class index: " + classIndex);
                    mapEquivalenceClasses.put(vecAnaphorNeId, classIndex);
                    if (isAnaphorOfType(documentCorpus, anaphorNeId, typeOfNP, catAttributeName)) { //check if the anaphor is of the desired type
                        if (mapAnaphorToAntecedent.containsKey(vecAnaphorNeId))
                            Local.out.println("*******Trying to overwrite: " + vecAnaphorNeId + ", antecedent" + vecAnteNeId);
                        else
                            mapAnaphorToAntecedent.put(vecAnaphorNeId, vecAnteNeId);
                    }
                } //end else if anaphor or ante is null
            } //end if rel="ident"
        } //end for
    }


    /**
     * Updates the mapping of ne-to-(vector of vectors of nes) with a new vector of keys.
     * Every member of vec will be mapped to vec.
     * Here, the intuituion is that a given ne from the corpus could have been mapped to more than one
     * sets (vectors) of nes from the parsed file.
     * @param vec The new vector of keys (i.e. neIds)
     */
    private void updateNeToVectorsMap(Vector<String> vec) {
        for (int k = 0; k < vec.size(); k++) {
            String key = (String) vec.elementAt(k);
            Vector<Vector<String>> neVectors = mapNeIdToVectors.get(key);
            if (neVectors == null) {
                neVectors = new Vector<Vector<String>>();
                mapNeIdToVectors.put(key, neVectors);
            }
            neVectors.add(vec);
        }
    }


    /*** This version of the method has not been used yet ***/
    /**
     * Builds the Discourse Model out of the annotation provided in the XML file.
     * This version of the method is generic in the sense that it takes into account anaphors of
     * all types of NP.
     * @param document The DOM document containing the annotation
     * @param anteTag The tag name of the nodes containing the annotation (i.e. ante)
     */
    private void constructEquivalenceClasses(Document document, String anteTag) {
        NodeList anteList = document.getElementsByTagName(anteTag);
        for (int i = 0; i < anteList.getLength(); i++) {
            Element ante = (Element) anteList.item(i);
            String relation = ante.getAttribute(IOXMLUtils.
                                                RELATION_PROPERTY_NAME);
            if (relation.equalsIgnoreCase("ident")) {
                String anaphorNeId = ante.getAttribute(IOXMLUtils.CURRENT_PROPERTY_NAME);
                NodeList aux = ante.getElementsByTagName(IOXMLUtils.ANCHOR_TAG);
                Element anchor = (Element) aux.item(0); //Though there might be more than one anchors for ambiguous anaphors
                String anteNeId = anchor.getAttribute(IOXMLUtils.ANTECEDENT_PROPERTY_NAME);
                Integer classIndex = (Integer) mapEquivalenceClasses.get(anteNeId);
                Set eqClass = new HashSet();
                if (classIndex == null) {
                    int newIndex = vecEquivalenceClasses.size();
                    classIndex = new Integer(newIndex);
                    eqClass.add(anteNeId);
                    vecEquivalenceClasses.add(eqClass);
                    mapEquivalenceClasses.put(anteNeId, classIndex);

                } else
                    eqClass = (Set) vecEquivalenceClasses.elementAt(classIndex.intValue());
                eqClass.add(anaphorNeId);
                mapEquivalenceClasses.put(anaphorNeId, classIndex);
                mapAnaphorToAntecedent.put(anaphorNeId, anteNeId);
            }
        } //end for
    }


    /**
     * This method checks whether a given neId is of a certain type or set of types. The method
     * returns true if the neId is of the desired type, and false otherwise. For efficiency the
     * first time this method is called an ne-to-node map is constructed.
     * @param doc The DOM document containing the annotation
     * @param anaphorNeId The id of the ne, whose type will ve checked
     * @param typeOfNP The type(s) of NP to be considered (the-np, pers-pro, etc.)
     * @param catAttributeName The attribute name which stores the type of NP (i.e. cat, AAcat)
     * @return boolean True if the type corresponding to anaphorNeId is one of typeOfNP, false otherwise
     */
    private boolean isAnaphorOfType(Document doc, String anaphorNeId,
                                    Vector typeOfNP, String catAttributeName) {
        if (mapIdToNode.isEmpty()) { //done only the first time
            //Get the name of the elements to be searched
            /***StringTokenizer st = new StringTokenizer( anaphorNeId, "0123456789"); //eliminate all the digits
               String elementName = st.nextToken();
             mapIdToNode = IOXMLUtils.buildIdToNodeMap( doc, elementName );***/
            //Corrected because I added a letter before the neId to make them unique for segmentation
            mapIdToNode = IOXMLUtils.buildIdToNodeMap(doc, "ne");
        }
        Element anaphorNode = IOXMLUtils.getElementById(anaphorNeId, mapIdToNode);
        String type = anaphorNode.getAttribute(catAttributeName);
        return typeOfNP.contains(type);
    }


    /***********************************************************************/
    /******** METHODS FOR PROCESSING INCREMENTALLY MAS-XML FILES ***********/
    /***********************************************************************/

    /**
     * Processes a file in XML-in XML-out fashion, identifying and annotating anaphoric links.
     * This is the main method for constructing incrementally a Discourse Model.
     * Essentially it does the following:
     * 1. Extracts all the NPs (NEs) in pre-order and processes ( processCf() ) them sequentially one at a time
     * 2. Resolved anaphors are stored in a Discourse Model and annotated directly into the DOM tree
     */
    public void process() {
        //Retrieve Cfs (NEs) in preorder (equivalent to a sequential order in the original text)
        NodeList cfList = domDocument.getElementsByTagName(IOXMLUtils.NP_TAG);

        //Process Cfs
        int numberOfCfs = cfList.getLength();
        for (int i = 0; i < numberOfCfs; i++)
            processCf(cfList.item(i));
    }


    /**
     * This method does the following 3 things:
     * 1. Updates the physical Discourse Model by processing a Cf held in a DOM Node and creating a corresponding Cf object for it
     * 2. Resolves anaphoric links (if any)
     * 3. Updates the logical Discourse Model and DOM tree accordingly.
     * @param theNode The DOM node holding a Cf to be processed
     */
    private void processCf(Node theNode) {
        //Update physical DM
        Utterance utterance = createUtterance(theNode);
        Cf currentCf = createCf(theNode, utterance); //Create and attach Cf to physical structure of DM

        /******** SLEEPING BIT BEGINS **********/
        if (globalQueryCounter >= 4990) { //query limit 5000
            //if( globalQueryCounter >= 3500 ) { //query limit 5000
            try {
                Thread.sleep(86400000);
            } catch (InterruptedException e) { //This is in millisecs, 86400000 = 24hr
                Local.err.println("ERROR: Could not get to sleep!");
                Local.err.printException(e);
            }
            globalQueryCounter = 0; //reset counter!
        }
        /******** SLEEPING BIT ENDS **********/

        //The corresponding implementation will be executed
        Cf antecedent = currentCf.resolveAnaphor(this); //Pass on this Discourse Model
        globalQueryCounter = globalQueryCounter + currentCf.queryCounter;
        updateLogicalDiscourseModel(currentCf, antecedent);
        //Add antecedent information (if any) to DOM tree
        if (currentCf.isAnaphoric()) {
            Node sentence = utterance.getNode();
            sentence.appendChild(createAnteNode(currentCf.getNode(), antecedent.getNode(), "ident"));
        }
    }


    /**
     * Creates an appropriate Cf object for a given syntactic phrase held in a DOM Node.
     * @param theNode The DOM node that holds the syntactic phrase
     * @param utt Utterance
     * @return Cf The Cf object appropriate for the syntactic phrase
     */
    private Cf createCf(Node theNode, Utterance utt) {
        String name = theNode.getNodeName();
        Cf result = null;
        if (name.equalsIgnoreCase(IOXMLUtils.NP_TAG))
            result = NominalGroup.newInstance(theNode, utt, dnClassifierType );
        //else if( name.equalsIgnoreCase( VP_TAG ) ) result = VerbalGroup.newInstance( theNode );
        createDiscourseEntity(result); //Create a Discourse Entity for the Cf
        return result;
    }


    /**
     * Creates an appropriate Utterance object for a given Cf held in a DOM Node.
     * It also updates the physical Discourse Model (segments<<utternaces<<Cfs).
     * @param theNode The DOM node that holds the Cf
     * @return Utterance The Utterance of the Cf held in theNode
     */
    private Utterance createUtterance(Node theNode) {
        Node uttNode = DiscourseModelImplementer.findUtteranceNode(theNode);
        Utterance utterance = (Utterance) mapNodeToSegment.get(uttNode);
        if (utterance == null) {
            Vector segmentNodes = findSegmentNodes(uttNode);
            Segment parent = (Segment) mapNodeToSegment.get((Node) segmentNodes.elementAt(0)); //upmost segment
            if (parent == null) {
                Node sgNode = (Node) segmentNodes.elementAt(0);
                parent = new Segment(sgNode, null); //Create upmost segment - root
                mapNodeToSegment.put(sgNode, parent);
                rootSegment = parent; //The main pointer to the physical DM
            }
            Segment segment = parent;
            for (int i = 1; i < segmentNodes.size(); i++) {
                segment = (Segment) mapNodeToSegment.get((Node) segmentNodes.
                        elementAt(i));
                if (segment == null) {
                    Node sgNode = (Node) segmentNodes.elementAt(i);
                    segment = new Segment(sgNode, parent);
                    //parent.addChild( segment );
                    mapNodeToSegment.put(sgNode, segment);
                }
                parent = segment;
            }
            utterance = new Utterance(uttNode, segment);
            //segment.addChild( utterance );
            mapNodeToSegment.put(uttNode, utterance);
        }
        return utterance;
    }


    /**
     * Updates the logical part of the ongoing DiscourseModel.
     * This involves updating three data structure:
     * 1. The Cf-to-EntityId map
     * 2. The Vector of Entities (EntityId is equal to the index in which a given Entity resides in this Vector)
     * 3. The Cf_anaphor-to-Cf_antecedent map
     * @param anaphor The Cf holding the anaphor
     * @param antecedent The Cf holding the antecedent
     */
    private void updateLogicalDiscourseModel(Cf anaphor, Cf antecedent) {
        if (antecedent != null) {
            DiscourseEntity deAnte = getDiscourseEntity(antecedent);
            DiscourseEntity deAnaphor = getDiscourseEntity(anaphor);
            deAnte.merge(deAnaphor, anaphor);
            mapEquivalenceClasses.put(anaphor, deAnte.getId()); //Replace existing mapping anaphor->deAnaphor
            mapAnaphorToAntecedent.put(anaphor, antecedent);
            vecEquivalenceClasses.remove(deAnaphor.getId().intValue());
        }
    }


    /**
     * Creates the ante node (<ante...><anchor.../></ante>) to be attached to the DOM tree.
     * @param current The node of the current Cf
     * @param antecedent The node of the antecedent
     * @param relation The relation between current Cf and antecedent
     * @return Node The DOM node that holds the anaphoric information
     */
    private Node createAnteNode(Node current, Node antecedent, String relation) {
        String currentId = ((Element) current).getAttribute(IOXMLUtils.
                ID_PROPERTY_NAME);
        String antecedentId = ((Element) antecedent).getAttribute(IOXMLUtils.
                ID_PROPERTY_NAME);
        Element ante = domDocument.createElement(IOXMLUtils.ANTE_TAG);
        Element anchor = domDocument.createElement(IOXMLUtils.ANCHOR_TAG);
        ante.setAttribute(IOXMLUtils.CURRENT_PROPERTY_NAME, currentId);
        ante.setAttribute(IOXMLUtils.RELATION_PROPERTY_NAME, relation);
        anchor.setAttribute(IOXMLUtils.ANTECEDENT_PROPERTY_NAME, antecedentId);
        ante.appendChild(anchor);
        return (ante);
    }


    /**
     * Finds the DOM node corresponding to the sentence in which a node holding a Cf is located.
     * Starts from node (first parameter) and goes up the tree until the sentence is found.
     * If no sentence is found then the parent of the upmost UNIT element is returned.
     * @param node Node
     * @return Node
     */
    public static Node findUtteranceNode(Node node) {
        Node domTreeUtteranceNode = null;
        Node parent = node.getParentNode();
        String parentName = null;
        Node auxParent = null; //is used to store the parent node of a Unit in case there is no sentence
        while (parent != null &&
               !(parentName = parent.getNodeName()).
                              equalsIgnoreCase(IOXMLUtils.SENTENCE_TAG)) {
            if (parentName.equalsIgnoreCase(IOXMLUtils.UNIT_TAG))
                auxParent = parent; //stores the uppermost unit node
            //Go up one level
            parent = parent.getParentNode();
        }
        if (parent == null) {
            if (auxParent == null) domTreeUtteranceNode = node.getParentNode(); //If both heuristics fail just take the parent of the NP
            else domTreeUtteranceNode = auxParent.getParentNode(); //the parent of the upmost unit node
        } else domTreeUtteranceNode = parent;
        return domTreeUtteranceNode;
    }


    /**
     * Finds the DOM nodes corresponding to the path of segments, starting from the utterance node given as a parameter and
     * continuing to the root Segment of the document that is being processed.
     * @param utteranceNode The utterance node
     * @return Vector The segment nodes above the utterance node
     */
    private Vector<Node> findSegmentNodes(Node utteranceNode) {
        Vector<Node> segmentNodes = new Vector<Node>();
        Node parent = utteranceNode.getParentNode();
        while (parent != null) {
            String parentName = parent.getNodeName();
            if (parentName.equalsIgnoreCase(IOXMLUtils.SEGMENT_TAG))
                segmentNodes.add(0, parent); //Insert at the beginning - reverse order of traversal
            //Go up one level
            parent = parent.getParentNode();
        }
        return segmentNodes;
    }

    public void setClassifier( int queryCounter, svm_model svmModel, SVMScalingModel svmScalingModel ) {
        globalQueryCounter = queryCounter;
        this.svmModel = svmModel;
        this.svmScalingModel = svmScalingModel;
	dnClassifierType = IOXMLUtils.SVM;
    }

    public void setClassifier( int queryCounter, MaxentModel model, ContextGenerator cg ) {
        globalQueryCounter = queryCounter;
        this.maxentModel = model;
        this.contextGenerator = cg;
	dnClassifierType = IOXMLUtils.MAXENT;
    }

    public void setClassifier() {
	dnClassifierType = IOXMLUtils.NODNCLASSIFIER;
    }

} //end DiscourseModelImplementer
