/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR;

//For read operation (XML -> DOM Tree)
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;
import org.w3c.dom.DOMException;

//For write operation (DOM Tree -> XML)
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerConfigurationException;

import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import java.io.*;
import java.util.Vector;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.HashMap;
import java.util.Map;
import java.util.Arrays;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.*;


/**
 * A wrapper class for hiding XML low level reading/writing.
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 1.1
 */
public abstract class IOXMLUtils {

    /*** PARAMETERS ***/
    public static final String SECTION_TAG = "section";
    public static final String UNIT_TAG = "unit";
    //public static final String SEGMENT_TAG = "AAsegment";
    public static final String SEGMENT_TAG = "gnomedoc";
    public static final String DIV_TAG = "div";
    public static final String PARAGRAPH_TAG = "p";
    public static final String SENTENCE_TAG = "s";
    public static final String PARSER_SENTENCE_TAG = "S"; //for txt-to-xml pipeline
    public static final String ANTE_TAG = "AAante"; //AA stands for automatically annotated
    public static final String ANCHOR_TAG = "anchor";
    public static final String NP_TAG = "ne";
    public static final String VE_TAG = "ve";
    public static final String WORD_TAG = "W";
    public static final String NP_HEAD_TAG = "nphead";
    public static final String NP_MODIFIER_TAG = "mod";
    public static final String TITLE_TAG = "title";

    public static final String LEXEME_PROPERTY_NAME = "lemma";
    public static final String ID_PROPERTY_NAME = "id";
    public static final String NEVEID_PROPERTY_NAME = "gId"; //global id over NEs and VEs
    public static final String CURRENT_PROPERTY_NAME = "current";
    public static final String RELATION_PROPERTY_NAME = "rel";
    public static final String ANTECEDENT_PROPERTY_NAME = "antecedent";
    public static final String CATEGORY_PROPERTY_NAME = "AAcat";
    public static final String PERSON_PROPERTY_NAME = "AAper";
    public static final String NUMBER_PROPERTY_NAME = "AAnum";
    public static final String GENDER_PROPERTY_NAME = "AAgen";
    public static final String DEPTHSCORE_LEFT_PROPERTY_NAME = "dsL";
    public static final String DEPTHSCORE_CENTRE_PROPERTY_NAME = "dsC";
    public static final String DEPTHSCORE_RIGHT_PROPERTY_NAME = "dsR";
    public static final String IDENTITY_RELATION = "ident";

    public static final String UTTERANCEID_PROPERTY_VALUE_PREFIX = "s";
    public static final String NEID_PROPERTY_VALUE_PREFIX = "ne";
    public static final String VEID_PROPERTY_VALUE_PREFIX = "ve";
    public static final String NEVEID_PROPERTY_VALUE_PREFIX = "nv";
    public static final String FIRSTPERSON_PROPERTY_VALUE = "per1";
    public static final String SECONDPERSON_PROPERTY_VALUE = "per2";
    public static final String THIRDPERSON_PROPERTY_VALUE = "per3";
    public static final String SINGULAR_PROPERTY_VALUE = "sing";
    public static final String PLURAL_PROPERTY_VALUE = "plur";
    public static final String FEMENINE_PROPERTY_VALUE = "fem";
    public static final String MASCULINE_PROPERTY_VALUE = "masc";
    public static final String IMPERSONAL_PROPERTY_VALUE = "neut";
    public static final String ANY_PROPERTY_VALUE = "any";

    public final static String ENCODING_UTF8 = "UTF8";
    public final static String ENCODING_LATIN1 = "8859_1";

    public static byte NODNCLASSIFIER = 0; //constant for the type of classifier - probably should move it to Local...something
    public static byte SVM = 1; //constant for the type of classifier
    public static byte MAXENT = 2; //constant for the type of classifier



    /**
     * Loads an XML file and parses it into a DOM Tree.
     * @param inputFileName The name of the XML file.
     * @return Document
     */
    public static Document load(String inputFileName) {
        return load(new File(inputFileName));
    }


    /**
     * Loads an XML file and parses it into a DOM Tree.
     * @param inputFile The name of the XML file.
     * @return Document
     */
    public static Document load(File inputFile) {
        Document document = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            //factory.setNamespaceAware( true );
            //factory.setValidating( true );
            DocumentBuilder builder = factory.newDocumentBuilder();
            document = builder.parse(inputFile);
        } catch (FactoryConfigurationError fce) {
            System.err.println("ERROR: " + fce.toString());
        } catch (ParserConfigurationException pce) {
            System.err.println("ERROR: " + pce.toString());
        } catch (SAXException saxe) {
            System.err.println("ERROR: " + saxe.toString());
        } catch (IOException ioe) {
            System.err.println("ERROR: " + ioe.toString());
        }

        return document;
    }


    /**
     * Stores an XML document (DOM Tree) to a file assuming and encoding of UTF8.
     * @param xmlDocument The XML document to be stored
     * @param outputFileName The name of the new XML file
     */
    public static void save(Document xmlDocument, String outputFileName) {
        save(xmlDocument, new File(outputFileName));
    }


    /**
     * Stores an XML document (DOM Tree) to a file assuming and encoding of UTF8.
     * @param xmlDocument The XML document to be stored
     * @param outputFile The name of the new XML file
     */
    public static void save(Document xmlDocument, File outputFile) {
        save(xmlDocument, outputFile, IOXMLUtils.ENCODING_UTF8);
        //save( xmlDocument, outputFile, IOXMLUtils.ENCODING_LATIN1 );
    }


    /**
     * Stores an XML document (DOM Tree) to a file.
     * @param xmlDocument The XML document to be stored
     * @param outputFile The name of the new XML file
     * @param encoding The encoding to be used (i.e. UTF8, LATIN-1, etc.)
     */
    public static void save(Document xmlDocument, File outputFile, String encoding) {
        try {
            //Create a transformer
            TransformerFactory transFactory = TransformerFactory.newInstance();
            Transformer transformer = transFactory.newTransformer();
            //Construct source
            DOMSource source = new DOMSource(xmlDocument);
            //Transform the tree into an XML file
            OutputStream fos = new FileOutputStream(outputFile);
            OutputStream bos = new BufferedOutputStream(fos);
            OutputStreamWriter osw = new OutputStreamWriter(bos, encoding); //Latin-1 encoding
            //StreamResult result = new StreamResult( fos );
            StreamResult result = new StreamResult(osw);
            transformer.transform(source, result);
        } catch (TransformerConfigurationException tce) {
            System.err.println("ERROR: " + tce.toString());
        } catch (TransformerException te) {
            System.err.println("ERROR: " + te.toString());
        } catch (IOException ioe) {
            System.err.println("ERROR: " + ioe.toString());
        }
    }


    /**
     * A recursive method similar to getElementsByTagName from the DOM interface, which in contrast
     * skips nodes with name given by the parameter tagIgnored and does not search within the subtrees
     * represented by the nodes with the same tag name as tagSearched, hence in general tagSearched should be
     * different from tagIgnored.
     * @param node the node representing the tree to be searched
     * @param tagIgnored the tag name of the nodes to be skipped
     * @param tagSearched the tag name of the nodes to be returned
     * @return Vector A Vector of all the child nodes of node with name tagSearched, which are not children of nodes with name tagIgnored
     */
    public static Vector<Node> getElementsByTagNameNoEmbedding(Node node, String tagIgnored, String tagSearched) {
        Vector<Node> vResult = new Vector<Node>();
        if (node.getNodeName().equalsIgnoreCase(tagSearched))
            vResult.add(node);
        else {
            NodeList nl = node.getChildNodes();
            for (int i = 0; i < nl.getLength(); i++) {
                Node child = nl.item(i);
                if (!child.getNodeName().equalsIgnoreCase(tagIgnored))
                    vResult.addAll(getElementsByTagNameNoEmbedding(child, tagIgnored, tagSearched));
            }
        }
        return vResult;
    }


    /**
     * A recursive method similar to getElementsByTagName from the DOM interface, which in contrast
     * skips nodes with names contained in the Set tagsIgnored and does not search within the subtrees
     * represented by the nodes with the same tag name as tagSearched.
     * @param node the node representing the tree to be searched
     * @param tagsIgnored the tag names of the nodes to be skipped
     * @param tagSearched the tag name of the nodes to be returned
     * @return Vector A Vector of all the child nodes of node with name tagSearched, which are not children of nodes with name tagIgnored
     */
    public static Vector<Node> getElementsByTagNameNoEmbedding(Node node, Set tagsIgnored, String tagSearched) {
        Vector<Node> vResult = new Vector<Node>();
        if (node.getNodeName().equalsIgnoreCase(tagSearched))
            vResult.add(node);
        else {
            NodeList nl = node.getChildNodes();
            for (int i = 0; i < nl.getLength(); i++) {
                Node child = nl.item(i);
                if (!tagsIgnored.contains(child.getNodeName()))
                    vResult.addAll(getElementsByTagNameNoEmbedding(child, tagsIgnored, tagSearched));
            }
        }
        return vResult;
    }


    /**
     * This method returns the preceeding node of a given node (first parameter) in a preorder traversal of the
     * DOM tree, having that its tag name is equal to the second parameter of this method.
     * The third parameter specifies whether to search only the tree represented by the parent
     * of the reference node (tree points to non-null reference), or the whole DOM tree (tree is null).
     * (This method is a variant of the getPreviousSibling() method in the DOM API in the sense that it retrieves the previous
     * sibling by tag name).
     * @param node The reference node
     * @param tagName The name of the tags to be considered
     * @param tree The node delimiting the scope of the search
     * @return Node The node preceeding the reference node in a preorder traversal
     */
    public static Node getPreviousNodeByTagName(Node node, String tagName, Node tree) {
        Node result = null;
        Node parent = node;
        while (parent != null) {
            Node sibling = parent.getPreviousSibling();
            while (sibling != null) {
                if (tagName.equalsIgnoreCase(sibling.getNodeName())) {
                    result = sibling;
                    break; //exit inner loop
                } else if (sibling instanceof Element) {
                    NodeList nl = ((Element) sibling).getElementsByTagName(tagName);
                    int numberOfElements = nl.getLength();
                    if (numberOfElements > 0) {
                        result = nl.item(numberOfElements - 1); //last child
                        break; //exit inner loop
                    }
                }
                sibling = sibling.getPreviousSibling();
            } //end inner while
            parent = parent.getParentNode();
            if (result != null || parent == tree)break; //stop searching
        } //end outer while

        return result;
    }


    /**
     * This method returns the following node after a given node (first parameter) in a preorder traversal of the
     * DOM tree, having that its tag name is equal to the second parameter of this method.
     * The third parameter specifies whether to search only the tree represented by the parent
     * of the reference node (tree points to non-null reference), or the whole DOM tree (tree is null).
     * (This method is a variance of the getNextSibling() method in the DOM API in the sense that it retrieves the next
     * sibling by tag name).
     * @param node The reference node
     * @param tagName The name of the tags to be considered
     * @param tree The node delimiting the scope of the search
     * @return Node The node preceeding the reference node in a preorder traversal
     */
    public static Node getNextNodeByTagName(Node node, String tagName, Node tree) {
        Node result = null;
        Node parent = node;
        while (parent != null) {
            Node sibling = parent.getNextSibling();
            while (sibling != null) {
                if (tagName.equalsIgnoreCase(sibling.getNodeName())) {
                    result = sibling;
                    break; //exit inner loop
                } else if (sibling instanceof Element) {
                    NodeList nl = ((Element) sibling).getElementsByTagName(tagName);
                    int numberOfElements = nl.getLength();
                    if (numberOfElements > 0) {
                        result = nl.item(0); //first child
                        break; //exit inner loop
                    }
                }
                sibling = sibling.getNextSibling();
            } //end inner while
            parent = parent.getParentNode();
            if (result != null || parent == tree)break; //stop searching
        } //end outer while

        return result;
    }


    /**
     * This method returns a DOM tree node specified by its id, eg. id="ne233".
     * @param doc The root of the DOM tree
     * @param elementId The id of the element to look for
     * @return Element The element whose attribute id is equal to the given id
     */
    public static Element getElementById(Document doc, String elementId) {
        Element result = null;
        //Get the name of the elements to be searched
        StringTokenizer st = new StringTokenizer(elementId, "0123456789"); //eliminate all the digits
        String elementName = st.nextToken();
        NodeList list = doc.getElementsByTagName(elementName);
        for (int i = 0; i < list.getLength(); i++) {
            Element element = (Element) list.item(i);
            if (element.getAttribute(IOXMLUtils.ID_PROPERTY_NAME).equalsIgnoreCase(elementId)) result = element;
        }
        return result;
    }


    /**
     * This is a more efficient implementation of the method getElementById( Document, String ), as it makes use
     * of a pre-builded hash table (hashmap).
     * @param elementId The id of the element to look for
     * @param m The map (hash table) to be searched through
     * @return Element The element whose attribute id is equal to the given id
     */
    public static Element getElementById(String elementId, Map m) {
        Element result = (Element) m.get(elementId);
        return result;
    }


    /**
     * Builds an id-to-node map for a given tag name for fast retrieval.
     * @param doc The root of the DOM tree
     * @param tagName The id of the element to look for
     * @return HashMap The resulting Map
     */
    public static Map<String, Element> buildIdToNodeMap(Document doc, String tagName) {
        Map<String, Element> result = new HashMap<String, Element>();
        NodeList list = doc.getElementsByTagName(tagName);
        for (int i = 0; i < list.getLength(); i++) {
            Element element = (Element) list.item(i);
            result.put(element.getAttribute(IOXMLUtils.ID_PROPERTY_NAME), element);
        }
        return result;
    }


    /**
     * Converts a Vector of word Nodes to a Vector of Strings (all characters are lowercase).
     * @param wordNodes The Vector of word Nodes
     * @return Vector The vector of Strings
     */
    public static Vector toStringsVector(Vector wordNodes) {
        Vector<String> result = new Vector<String>();
        for (int i = 0; i < wordNodes.size(); i++)
            result.add(((Node) wordNodes.elementAt(i)).getFirstChild().getNodeValue().trim().toLowerCase());
        return result;
    }


    /**
     * Converts a Vector of word Nodes to a Vector of Strings. The original (lower, upper) case is kept.
     * @param wordNodes The Vector of word Nodes
     * @return Vector The vector of Strings
     */
    public static Vector toStringsVectorKeepLetterCase(Vector wordNodes) {
        Vector<String> result = new Vector<String>();
        for (int i = 0; i < wordNodes.size(); i++)
            result.add(((Node) wordNodes.elementAt(i)).getFirstChild().getNodeValue().trim());
        return result;
    }


    /*
     * Returns the string of words that compose the phrase held in the DOM node passed as a parameter.
     * @param cf The DOM node holding a constituent
     * @return String The string of words that compose the constituent
     */
    public static String getPhraseString(Node cf) {
        StringBuffer result = new StringBuffer();
        NodeList wordNodes = ((Element) cf).getElementsByTagName(IOXMLUtils.WORD_TAG);
        for (int i = 0; i < wordNodes.getLength(); i++)
            result.append(wordNodes.item(i).getFirstChild().getNodeValue().trim() + " ");
        if (result.length() > 0) result.deleteCharAt(result.length() - 1); //remove last white space
        return (result.toString());
    }


    /**
     * A recursive method that returns the number of nodes with tag name given as a parameter within
     * the tree represented by parameter node1 and before node2 if present in the latter tree.
     * @param node1 The node where the counting starts
     * @param node2 The node where the counting ends
     * @param tagName The name of the tags to be accounted for
     * @return Vector A Vector of two elements: 1- true if node2 is a child of node1 (false otherwise) 2- number of nodes with name=tagName
     */
    public static Vector getNumberOfMatchingNodes(Node node1, Node node2, String tagName) {
        Vector result = new Vector();
        boolean found = false;
        int numberOfMatchingNodes = 0;
        if (node1 == node2) {
            found = true;
            numberOfMatchingNodes = 1; //if we want to count node2 as well
        } else if (!node1.hasChildNodes()) {
            if (node1.getNodeName().equalsIgnoreCase(tagName))
                numberOfMatchingNodes = 1;
        } else {
            if (node1.getNodeName().equalsIgnoreCase(tagName))
                numberOfMatchingNodes = 1;
            NodeList childNodes = node1.getChildNodes();
            Vector resChildren = null;
            for (int i = 0; i < childNodes.getLength(); i++) {
                resChildren = getNumberOfMatchingNodes(childNodes.item(i), node2, tagName);
                found = ((Boolean) resChildren.elementAt(0)).booleanValue();
                numberOfMatchingNodes += ((Integer) resChildren.elementAt(1)).intValue();
                if (found)break; //stop searching
            }
        }

        result.add(new Boolean(found));
        result.add(new Integer(numberOfMatchingNodes));
        return (result);
    }


    /**
     * A method that returns the number of intermediate nodes with tag name given as a parameter
     * between node1 and node2 (in the DOM tree).
     * @param node1 The node where the counting starts
     * @param node2 The node where the counting ends
     * @param tagName The name of the tags to be accounted for
     * @return int Number of intermediate nodes
     */
    public static int getNumberOfIntermediateNodes(Node node1, Node node2, String tagName) {
        boolean found = false;
        int result = 0;
        Node node = node1;
        while (!found) {
            //Get next sibling
            Node nextSibling = null;
            while (nextSibling == null) {
                nextSibling = node.getNextSibling();
                node = node.getParentNode();
                if (node == null)break; //the end of the tree has been reached
            }
            if (nextSibling != null) {
                node = nextSibling;
                Vector aux = IOXMLUtils.getNumberOfMatchingNodes(nextSibling, node2, tagName);
                found = ((Boolean) aux.elementAt(0)).booleanValue();
                result += ((Integer) aux.elementAt(1)).intValue();
            } else break;
        } //end outer while

        return result;
    }


    /**
     * Performs a preorder traversal of the tree represented by the node parent, starting at node node1
     * and continues until either node2 is found or the end of the tree has been reached. Returns true
     * if node2 is found, false otherwise. The method assumes that neither node2 is part of the tree represented by node1,
     * nor node1 is part of the tree represented by node2.
     * @param node1 The first node
     * @param node2 The second node
     * @param parent The parent node
     * @return boolean True if node1 before node2 in tree represented by parent, false otherwise
     */
    public static boolean isBeforeNode(Node node1, Node node2, Node parent) {
        boolean result = false;
        Node sibling = node1;
        while (!result && (sibling = sibling.getNextSibling()) != null) {
            result = isNodeInTree(node2, sibling);
        }
        if (!result) {
            Node auxParent = node1.getParentNode();
            if (auxParent != parent) result = isBeforeNode(auxParent, node2, parent);
        }
        return result;
    }


    /**
     * Checks whether node (first parameter) is part of the tree represented by the second parameter.
     * @param node The node to be searched for
     * @param tree The tree defining the scope of the search
     * @return boolean True if node "node" is part of node "tree", false otherwise
     */
    public static boolean isNodeInTree(Node node, Node tree) {
        boolean result = false;
        if (node == tree) result = true;
        else {
            NodeList children = tree.getChildNodes();
            for (int i = 0; !result && i < children.getLength(); i++)
                result = isNodeInTree(node, children.item(i));
        }
        return result;
    }


    /*************** Code from PersonalPronoun ********************/
    /**
     * Returns the verb (if any) immediately to the left of the NP, given as a parameter.
     * @param ne The focal ne
     * @return String The verb to the left of ne, null if no such verb
     */
    public static String getLeftCollocate(NominalGroup ne) {
        String result = null;
        Node uttNode = ne.getUtterance().getNode();
        Node node = ne.getNode();
        Node parentNode = node.getParentNode();
        if (!parentNode.getNodeName().equalsIgnoreCase(IOXMLUtils.NP_TAG)) { //if the ne is not embedded into a more complex NP
            stop: { //compound statement used to exit from the main loop
                while (result == null) {
                    Node prevSibling = null;
                    while (prevSibling == null) {
                        prevSibling = node.getPreviousSibling();
                        node = node.getParentNode();
                        if (node == uttNode)break;
                    }
                    node = prevSibling;
                    if (prevSibling == null)break stop; //stop processing
                    if (prevSibling instanceof org.w3c.dom.Text
                        || prevSibling instanceof org.w3c.dom.Comment)continue; //just continue with next iteration
                    String nodeName = prevSibling.getNodeName();
                    if (nodeName.equalsIgnoreCase(IOXMLUtils.WORD_TAG))
                        result = processWord(prevSibling);
                    else if (!nodeName.equalsIgnoreCase(IOXMLUtils.NP_TAG)) { //anything except NP
                        NodeList nl = ((Element) prevSibling).getElementsByTagName(IOXMLUtils.WORD_TAG);
                        for (int i = nl.getLength() - 1; i >= 0; i--) {
                            result = processWord(nl.item(i));
                            if (isElementOfNP(nl.item(i), prevSibling)) {
                                result = null; //reset result
                                break stop; //stop processing, an adjacent NP was found
                            }
                            if (result != null)break stop; //stop processing
                        }
                    } else break stop; //stop processing, an adjacent NP was found
                } //end while
            } //stop statement
        }
        return result;
    }


    /*
     * Returns the verb (if any) immediately to the left of the NP held in node.
     * @param node The node holding the NP of interest
     * @param uttNode The utterance node of the NP held in node
     * @return String the verb to the left of the NP, null if no such verb
     */
    public static String getLeftCollocate(Node node, Node uttNode) {
        String result = null;
        Node parentNode = node.getParentNode();
        if (!parentNode.getNodeName().equalsIgnoreCase(IOXMLUtils.NP_TAG)) { //if the ne is not embedded into a more complex NP
            stop: { //compound statement used to exit from the main loop
                while (result == null) {
                    Node prevSibling = null;
                    while (prevSibling == null) {
                        prevSibling = node.getPreviousSibling();
                        node = node.getParentNode();
                        if (node == uttNode)break;
                    }
                    node = prevSibling;
                    if (prevSibling == null)break stop; //stop processing
                    if (prevSibling instanceof org.w3c.dom.Text
                        || prevSibling instanceof org.w3c.dom.Comment)continue; //just continue with next iteration
                    String nodeName = prevSibling.getNodeName();
                    if (nodeName.equalsIgnoreCase(IOXMLUtils.WORD_TAG))
                        result = processWord(prevSibling);
                    else if (!nodeName.equalsIgnoreCase(IOXMLUtils.NP_TAG)) { //anything except NP
                        NodeList nl = ((Element) prevSibling).getElementsByTagName(IOXMLUtils.WORD_TAG);
                        for (int i = nl.getLength() - 1; i >= 0; i--) {
                            result = processWord(nl.item(i));
                            if (isElementOfNP(nl.item(i), prevSibling)) {
                                result = null; //reset result
                                break stop; //stop processing, an adjacent NP was found
                            }
                            if (result != null)break stop; //stop processing
                        }
                    } else break stop; //stop processing, an adjacent NP was found
                } //end while
            } //stop statement
        } //End first if
        return result;
    }


    /**
     * Checks whether the word node, given as a parameter, contains a verb and if so this verb is returned.
     * @param word The word to be processed
     * @return String the verb if the word is a verb, null otherwise
     */
    public static String processWord(Node word) {
        String result = null;
        String pos = ((Element) word).getAttribute(TagSet.POS_PROPERTY_NAME);
        //if the word is a verb (pos is part of the list of verb tags)
        if (Local.containsIgnoreCase(TagSet.LIST_VERB_TAGS, pos)) {
            result = ((Element) word).getAttribute(IOXMLUtils.LEXEME_PROPERTY_NAME); //get lemma
            if (result == null || result.length() == 0)
                result = word.getFirstChild().getNodeValue().trim().toLowerCase(); //get surface form
        }
        return result;
    }


    /**
     * Tests whether the node, given as a parameter, is part of an NP.
     * Starts from node and goes up the DOM tree until either an NP node is found or upmostNode is reached.
     * @param node The starting node of the search
     * @param upmostNode The node, beyond which no search is done
     * @return boolean True if node is part of an NP, false otherwise
     */
    public static boolean isElementOfNP(Node node, Node upmostNode) {
        boolean result = false;
        Node parent = node;
        while (!result) {
            parent = parent.getParentNode();
            if (parent.getNodeName().equalsIgnoreCase(IOXMLUtils.NP_TAG)) result = true;
            if (parent == upmostNode)break;
        }
        return result;
    }


    /*
     * Returns the verb (if any) immediately to the right of the NP, given as a parameter.
     * @param ne The focal ne
     * @return String the verb to the right of ne, null if no such verb
     */
    public static String getRightCollocate(NominalGroup ne) {
        String result = null;
        Node uttNode = ne.getUtterance().getNode();
        Node node = ne.getNode();
        Node parentNode = node.getParentNode();
        if (!parentNode.getNodeName().equalsIgnoreCase(IOXMLUtils.NP_TAG)) { //if the ne is not embedded into a more complex NP
            Node firstVerbNode = null;
            stop: { //compound statement used to exit from the main loop
                while (result == null) {
                    Node nextSibling = null;
                    while (nextSibling == null) {
                        nextSibling = node.getNextSibling();
                        node = node.getParentNode();
                        if (node == uttNode)break;
                    }
                    node = nextSibling;
                    if (nextSibling == null)break stop; //stop processing
                    if (nextSibling instanceof org.w3c.dom.Text
                        || nextSibling instanceof org.w3c.dom.Comment)continue; //just continue with next iteration
                    String nodeName = nextSibling.getNodeName();
                    if (nodeName.equalsIgnoreCase(IOXMLUtils.WORD_TAG)) {
                        firstVerbNode = nextSibling;
                        result = processWord(nextSibling);
                    } else if (!nodeName.equalsIgnoreCase(IOXMLUtils.NP_TAG)) { //anything except NP
                        NodeList nl = ((Element) nextSibling).getElementsByTagName(IOXMLUtils.WORD_TAG);
                        for (int i = 0; i < nl.getLength(); i++) {
                            firstVerbNode = nl.item(i);
                            result = processWord(firstVerbNode);
                            if (isElementOfNP(nl.item(i), nextSibling)) {
                                result = null; //reset result
                                break stop; //stop processing, an adjacent NP was found
                            }
                            if (result != null)break stop; //stop processing
                        }
                    } else break stop; //stop processing, an adjacent NP was found
                } //end while
            } //stop statement
            if (result != null) result = findLastVerb(firstVerbNode);
        } //End first if
        return result;
    }


    /*
     * Returns the verb (if any) immediately to the right of the NP held in node.
     * @param node The node holding the NP of interest
     * @param uttNode The utterance node of the NP held in node
     * @return String the verb to the right of the NP, null if no such verb
     */
    public static String getRightCollocate(Node node, Node uttNode) {
        String result = null;
        Node parentNode = node.getParentNode();
        if (!parentNode.getNodeName().equalsIgnoreCase(IOXMLUtils.NP_TAG)) { //if the ne is not embedded into a more complex NP
            Node firstVerbNode = null;
            stop: { //compound statement used to exit from the main loop
                while (result == null) {
                    Node nextSibling = null;
                    while (nextSibling == null) {
                        nextSibling = node.getNextSibling();
                        node = node.getParentNode();
                        if (node == uttNode)break;
                    }
                    node = nextSibling;
                    if (nextSibling == null)break stop; //stop processing
                    if (nextSibling instanceof org.w3c.dom.Text
                        || nextSibling instanceof org.w3c.dom.Comment)continue; //just continue with next iteration
                    String nodeName = nextSibling.getNodeName();
                    if (nodeName.equalsIgnoreCase(IOXMLUtils.WORD_TAG)) {
                        firstVerbNode = nextSibling;
                        result = processWord(nextSibling);
                    } else if (!nodeName.equalsIgnoreCase(IOXMLUtils.NP_TAG)) { //anything except NP
                        NodeList nl = ((Element) nextSibling).getElementsByTagName(IOXMLUtils.WORD_TAG);
                        for (int i = 0; i < nl.getLength(); i++) {
                            firstVerbNode = nl.item(i);
                            result = processWord(firstVerbNode);
                            if (isElementOfNP(nl.item(i), nextSibling)) {
                                result = null; //reset result
                                break stop; //stop processing, an adjacent NP was found
                            }
                            if (result != null)break stop; //stop processing
                        }
                    } else break stop; //stop processing, an adjacent NP was found
                } //end while
            } //stop statement
            if (result != null) result = findLastVerb(firstVerbNode);
        } //End first if
        return result;
    }


    /**
     * Assuming that verbal constructions must be horizontally represented, in terms of XML, in the GNOME
     * Corpus, this method goes on traversing the siblings of node (parameter) until the last one is reached.
     * On the way any word whose part-of-speech is a verb is stored, hence at the end the rightmost verb, of
     * a possibly more complex verbal construction, is stored and returned.
     * @param node The node holding the first verb of a (possibly) more complex construction
     * @return String The rightmost verb
     */
    public static String findLastVerb(Node node) {
        String result = null;
        String safeResult = processWord(node);
        Node nextSibling = node.getNextSibling();
        while (nextSibling != null) {
            String nodeName = nextSibling.getNodeName();
            if (nodeName.equalsIgnoreCase(IOXMLUtils.WORD_TAG)) {
                String aux = processWord(nextSibling);
                if (aux != null) result = aux; //The word contains a verb
            } else break;
            nextSibling = nextSibling.getNextSibling();
        }
        return (result == null ? safeResult : result);
    }


    /**
     * Returns the first parent node with the specified tag name
     * @param node The node
     * @param topNode The node beyond which should not be searched
     * @param tagName String
     * @return Node The parent node with the specified tag name or if no such node the topNode is returned
     */
    public static Node getParentNodeByName(Node node, Node topNode, String tagName) {
        Node parent = node.getParentNode();
        while (!(parent.getNodeName().equalsIgnoreCase(tagName) || parent == topNode)) {
            parent = parent.getParentNode();
        }
        return parent;
    }

} //end class IOXMLUtils
