/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR;

import java.util.*;

import org.w3c.dom.*;

/**
 * A class that encapsulates the general functionality of a Noun Phrase (NP).
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 1.1
 */
public class NominalGroup extends Cf {
    //Constants
    //NPs
    public static final byte THE_NP = 0;
    public static final byte THAT_NP = 1;
    public static final byte THIS_NP = 2;
    public static final byte SUCH_NP = 3;
    public static final byte A_NP = 4;
    public static final byte ANOTHER_NP = 5;
    public static final byte NUM_NP = 6;
    public static final byte MEAS_NP = 7; //This is not used in findType()
    public static final byte BARE_NP = 8;
    public static final byte POSS_NP = 9;
    public static final byte Q_NP = 10;
    public static final byte PN = 11;
    public static final byte THE_PN = 12;
    public static final byte COORD_NP = 13;
    //Pronouns
    public static final byte POSS_PRO = 14;
    public static final byte PERS_PRO = 15;
    public static final byte WH_PRO = 16;
    public static final byte THIS_PRO = 17;
    public static final byte THAT_PRO = 18;
    public static final byte Q_PRO = 19;
    //Groups
    public static final byte[] DEFINITE = {0, 1, 2, 3, 12};
    public static final byte[] INDEFINITE = {4, 5, 6, 7, 8};
    public static final byte[] PRONOUNS = {14, 15, 16, 17, 18, 19};
    //Not used at the moment...
    public static final byte SUBJECT = 0;
    public static final byte DIRECT_OBJECT = 1;
    public static final byte INDIRECT_OBJECT = 2;

    //Instance variables
    private Node headNode; //A pointer to the DOM node holding the head of this NP
    private byte gramaticalFunction; //Subject, (in)direct object, etc.
    private Vector<Node> preModifiers, //A vector of DOM tree nodes
            postModifiers; //A vector of DOM tree nodes
    private Agreement agreement; //The agreement features of the ne
    private int[] antecedentIndicators; //For Mitkov's algorithm for pronoun resolution

    /**
     * Constructor. Sets the DOM node that holds the head of the NP (typically a word element - 'W'),
     * the premodifiers, postmodifiers, agreement features and the gramatical function (i.e. subject).
     * @param theNode The node that holds the Cf
     * @param theUtterance The Utterance of this Cf
     * @param type The type of the NP
     */
    public NominalGroup(Node theNode, Utterance theUtterance, byte type) {
        super(theNode, theUtterance, type);
        preModifiers = new Vector<Node>();
        postModifiers = new Vector<Node>();
        setHeadNode();
        setGramaticalFunction();
        setAgreementFeatures();
        setModifiers();
    }


    /**
     * A dummy method. No implementation at this level for the moment.
     */
    public Cf resolveAnaphor(DiscourseModel dm) {
        return null;
    }


    //Accessors

    /**
     * Returns the node holding the head of the NP.
     * @return Node the node holding the head of the NP
     */
    public Node getHeadNode() {
        return headNode;
    }


    /**
     * Returns the vector of pre-modifiers.
     * @return Vector The vector of pre-modifiers
     */
    public Vector getPreModifiers() {
        return preModifiers;
    }


    /**
     * Returns the vector of post-modifiers.
     * @return Vector The vector of post-modifiers
     */
    public Vector getPostModifiers() {
        return postModifiers;
    }


    /**
     * Returns the agreement (syntactic) features of this NP (namely person, number, gender).
     * @return Agreement the agreement features
     */
    public Agreement getAgreement() {
        return agreement;
    }


    /**
     * Returns the antecedent indicators of this NP (used in Mitkov's algorithm for pronoun resolution).
     * @return int[] The antecedent indicators
     */
    public int[] getAntecedentIndicators() {
        return antecedentIndicators;
    }


    /**
     * Sets the antecedent indicators of this NP.
     * @param indicators The antecedent indicators
     */
    public void setAntecedentIndicators(int[] indicators) {
        antecedentIndicators = indicators;
    }


    /*** PRIVATE METHODS ***/

    /**
     * Reads the head of this NP from input file and stores it internally.
     * The input file is expected to be MAS-XML compliant.
     */
    private void setHeadNode() {
        Node node = getNode();
        //NodeList nl = ((Element)node).getElementsByTagName( IOXMLUtils.NP_HEAD_TAG );
        Vector nl = IOXMLUtils.getElementsByTagNameNoEmbedding(node, IOXMLUtils.NP_TAG, IOXMLUtils.NP_HEAD_TAG);
        if (nl.size() > 0) headNode = (Node) nl.elementAt(0); //There is only one head
    }


    /**
     * Reads the gramatical function from input file and stores it internally.
     * The input file is expected to be MAS-XML compliant.
     * Note: At the moment MAS-XML does not include GF.
     */
    private void setGramaticalFunction() {
    }


    /**
     * Reads the agreement features (number, person, gender) from input file and stores them internally.
     * The input file is expected to be MAS-XML compliant.
     */
    private void setAgreementFeatures() {
        Node node = getNode();
        String strPerson = ((Element) node).getAttribute(IOXMLUtils.PERSON_PROPERTY_NAME);
        String strNumber = ((Element) node).getAttribute(IOXMLUtils.NUMBER_PROPERTY_NAME);
        String strGender = ((Element) node).getAttribute(IOXMLUtils.GENDER_PROPERTY_NAME);
        agreement = new Agreement(strPerson, strNumber, strGender);
    }


    /**
     * Reads the modifiers (pre- and post-) from input file and stores them internally.
     * The input file is expected to be MAS-XML compliant.
     */
    private void setModifiers() {
        Node node = getNode();
        //NodeList nl = ((Element)node).getElementsByTagName( IOXMLUtils.NP_MODIFIER_TAG );
        Vector nl = IOXMLUtils.getElementsByTagNameNoEmbedding(node, IOXMLUtils.NP_TAG, IOXMLUtils.NP_MODIFIER_TAG);
        if (nl.size() > 0) {
            boolean preMod = true;
            for (int i = 0; i < nl.size(); i++) {
                Node mod = (Node) nl.elementAt(i);
                if (preMod && (preMod = IOXMLUtils.isBeforeNode(mod, getHeadNode(), getUtterance().getNode())))
                    preModifiers.add(mod); //store mod nodes before head node
                else
                    postModifiers.add(mod); //store mod nodes after head node
            } //end for
        } //end if no <mod> tag found
    }


    /*** STATIC METHODS ***/

    /**
     * A static method that creates an appropriate object (implementation) for a given noun phrase,
     * depending on its type.
     * @param theNode The DOM node that holds the noun phrase to be instantiated
     * @param theUtterance The Utterance of the noun phrase to be instantiated
     * @return NominalGroup The reference to the newly created object
     */
    public static NominalGroup newInstance(Node theNode, Utterance theUtterance, byte dnClassifierType ) {
        NominalGroup result = null;
        byte type = NominalGroup.findType(theNode);
        switch (type) {
        case NominalGroup.THE_PN:
        case NominalGroup.THE_NP:
            result = new DefiniteDescription(theNode, theUtterance, type, dnClassifierType);
            break;
        case NominalGroup.POSS_PRO:
            result = new PossessivePronoun(theNode, theUtterance, type);
            break;
        case NominalGroup.PERS_PRO:
            result = new PersonalPronoun(theNode, theUtterance, type);
            break;
        case NominalGroup.THAT_NP:
        case NominalGroup.THIS_NP:
        case NominalGroup.SUCH_NP:
            result = new Demonstrative(theNode, theUtterance, type);
            break;
        case NominalGroup.PN:
            result = new ProperName(theNode, theUtterance, type);
            break;
        default:
            result = new NominalGroup(theNode, theUtterance, type);
        }

        return result;
    }


    /**
     * A static method that reads the type of a Noun Phrase from the source XML file.
     * @param node The DOM node that holds the noun phrase to be processesed
     * @return byte The type of the noun phrase (the-np, a-np, etc.)
     */
    public static byte findType(Node node) {
        Element ne = (Element) node;
        String npType = ne.getAttribute(IOXMLUtils.CATEGORY_PROPERTY_NAME);
        return (NominalGroup.getByteType(npType));
    }


    /**
     * Converts an NP type encoded in a byte format into a string representation.
     * Eg. 0 -> "the-np".
     * @param type The type as a byte
     * @return String The type as a String
     */
    public static String getStringType(byte type) {
        switch (type) {
        case NominalGroup.THE_NP:
            return (new String("the-np"));
        case NominalGroup.THAT_NP:
            return (new String("that-np"));
        case NominalGroup.THIS_NP:
            return (new String("this-np"));
        case NominalGroup.SUCH_NP:
            return (new String("such-np"));
        case NominalGroup.A_NP:
            return (new String("a-np"));
        case NominalGroup.ANOTHER_NP:
            return (new String("another-np"));
        case NominalGroup.NUM_NP:
            return (new String("num-np"));
        case NominalGroup.MEAS_NP:
            return (new String("meas-np"));
        case NominalGroup.BARE_NP:
            return (new String("bare-np"));
        case NominalGroup.POSS_NP:
            return (new String("poss-np"));
        case NominalGroup.Q_NP:
            return (new String("q-np"));
        case NominalGroup.PN:
            return (new String("pn"));
        case NominalGroup.THE_PN:
            return (new String("the-pn"));
        case NominalGroup.COORD_NP:
            return (new String("coord-np"));
        case NominalGroup.POSS_PRO:
            return (new String("poss-pro"));
        case NominalGroup.PERS_PRO:
            return (new String("pers-pro"));
        case NominalGroup.WH_PRO:
            return (new String("wh-pro"));
        case NominalGroup.THIS_PRO:
            return (new String("this-pro"));
        case NominalGroup.THAT_PRO:
            return (new String("that-pro"));
        case NominalGroup.Q_PRO:
            return (new String("q-pro"));
        default:
            return (new String(""));
        }
    }


    /**
     * Converts an NP type encoded in a String format into a byte representation.
     * Eg. "the-np" -> 0.
     * @param npType The type as String
     * @return byte The type as byte
     */
    public static byte getByteType(String npType) {
        byte type = -1;
        if (npType.equalsIgnoreCase("the-np")) type = NominalGroup.THE_NP;
        else if (npType.equalsIgnoreCase("that-np")) type = NominalGroup.THAT_NP;
        else if (npType.equalsIgnoreCase("this-np")) type = NominalGroup.THIS_NP;
        else if (npType.equalsIgnoreCase("such-np")) type = NominalGroup.SUCH_NP;
        else if (npType.equalsIgnoreCase("a-np")) type = NominalGroup.A_NP;
        else if (npType.equalsIgnoreCase("another-np")) type = NominalGroup.ANOTHER_NP;
        else if (npType.equalsIgnoreCase("num-np")) type = NominalGroup.NUM_NP;
        else if (npType.equalsIgnoreCase("meas-np")) type = NominalGroup.MEAS_NP;
        else if (npType.equalsIgnoreCase("bare-np")) type = NominalGroup.BARE_NP;
        else if (npType.equalsIgnoreCase("poss-np")) type = NominalGroup.POSS_NP;
        else if (npType.equalsIgnoreCase("q-np")) type = NominalGroup.Q_NP;
        else if (npType.equalsIgnoreCase("pn")) type = NominalGroup.PN;
        else if (npType.equalsIgnoreCase("the-pn")) type = NominalGroup.THE_PN;
        else if (npType.equalsIgnoreCase("coord-np")) type = NominalGroup.COORD_NP;
        else if (npType.equalsIgnoreCase("poss-pro")) type = NominalGroup.POSS_PRO;
        else if (npType.equalsIgnoreCase("pers-pro")) type = NominalGroup.PERS_PRO;
        else if (npType.equalsIgnoreCase("wh-pro")) type = NominalGroup.WH_PRO;
        else if (npType.equalsIgnoreCase("this-pro")) type = NominalGroup.THIS_PRO;
        else if (npType.equalsIgnoreCase("that-pro")) type = NominalGroup.THAT_PRO;
        else if (npType.equalsIgnoreCase("q-pro")) type = NominalGroup.Q_PRO;
        return type;
    }


    /**
     * Determines if to NominalGroups are equal
     * @param obj Object
     * @return boolean
     */
  /*  public boolean equals(Object obj) {
        Element objElement = (Element) ((Cf) obj).getNode();
        if (objElement == null)
            return false;
        String objID = objElement.getAttribute(IOXMLUtils.ID_PROPERTY_NAME);

        Element thisElement = (Element)this.getNode();
        if (thisElement == null)
            return false;
        String thisID = thisElement.getAttribute(IOXMLUtils.ID_PROPERTY_NAME);

        return thisID.equals(objID);
    }*/
} //End class NominalGroup
