/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR;

import java.util.*;

import org.w3c.dom.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.*;


/**
 * A class that holds the functionality for processing Possessive Pronouns (i.e. his, its).
 * It's mainly meant to provide the implementation for resolving the antecedent of a possessive pronoun.
 * We have reused Mitkov's algorithm, covered in detail in:
 * - Ruslan Mitkov. Robust Pronoun Resolution with Limited Knowledge. Proceedings of the 18th International Conference on Computational Linguistics (COLING'98)/ACL'98 Conference, 869-875. Montreal, Canada.
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 1.0
 */
public class PossessivePronoun extends NominalGroup {
    /** Specifies how many sentences back to look for antecedent */
    public static final int WINDOW_SIZE = 3; //How far back should we look for antecedent
    /** The list of indicating verbs.
     * (discuss, present, illustrate, identify, summarise, examine, describe, define, show, check,
     * develop, review, report, outline, consider, investigate, explore, assess, analyse, synthesise,
     * study, survey, deal, cover)
     */
    private static final String VERB_LIST[] = {"discuss", "present", "illustrate", "identify", "summarise", "examine",
            "describe", "define", "show", "check", "develop", "review",
            "report", "outline", "consider", "investigate", "explore", "assess",
            "analyse", "synthesise", "study", "survey", "deal", "cover"};
    /** Holds the first verb to the left of this NP (if any) */
    private String leftCollocate;
    /** Holds the first verb to the right of this NP (if any) */
    private String rightCollocate;

    /**
     * Constructor. Just calls the constructor of NominalGroup.
     */
    public PossessivePronoun(Node theNode, Utterance utt, byte theType) {
        super(theNode, utt, theType);
        leftCollocate = IOXMLUtils.getLeftCollocate(this);
        rightCollocate = IOXMLUtils.getRightCollocate(this);
    }


    /*
     public Cf resolveAnaphor( DiscourseModel dm ) {
     return null;
     }
     */

    /**
     * The implementation of Mitkov's algorithm for processing Personal Pronouns.
     * @param dm The Discourse Model to which this Cf belongs
     * @return Cf The proposed antecedent
     */
    //public Cf orgresolveAnaphor( DiscourseModel dm ) {
    public Cf resolveAnaphor(DiscourseModel dm) {
        Cf result = null;
        Set<PotentialAntecedent> potentialAntecedents = new TreeSet<PotentialAntecedent>(); //Ordered set for the antecedents
        //Determine potential antecedents based on syntactic agreement
        Utterance utterance = getUtterance();
	//Get the parent Cf of this Cf if any
        Cf thisCf = this; //default
        Cf parentCf = getParentCf(dm); //a method in class Cf
        if (parentCf != null) thisCf = parentCf;
        for (int distance = 0; utterance != null && distance < WINDOW_SIZE; distance++) { //loop over utterances backwards
            Vector vCfs = dm.getCfs(utterance);
            for (int j = 0; j < vCfs.size(); j++) { //loop over Cfs forwards
                Cf cfAnte = (Cf) vCfs.elementAt(j);
                if (cfAnte == thisCf) break; //Only for the first utterance, which holds this cf
                NominalGroup antecedent = (NominalGroup) cfAnte;
                if (this.getAgreement().equals(antecedent.getAgreement())) {
                    Local.out.println("PossessivePronoun", "UTT: " + ((Element) utterance.getNode()).getAttribute("id") + " cfAnte: " + ((Element) cfAnte.getNode()).getAttribute("id"));
                    //Map antecedent (NominalGroup) to ante (PotentialAntecedent)
                    PotentialAntecedent ante = new PotentialAntecedent(antecedent, dm, PotentialAntecedent.SCORE, utterance);
                    potentialAntecedents.add(ante);
                }
            } //end inner for
            utterance = dm.getPrevUtterance(utterance);
        } //end outer for
        //Select the preferred antecedent
        if (!potentialAntecedents.isEmpty()) {
            NominalGroup ne = selectAntecedent(potentialAntecedents, PotentialAntecedent.IMMEDIATE_REFERENCE);
            if (ne != null) {
                Local.out.println("PersonalPronoun", "ANTECEDENT: " + ne.getNode().getAttributes());
                Local.out.println("PersonalPronoun", "FOR: " + this.getNode().getAttributes());
                setAnaphoric(true);
                result = ne;
            } else
                System.err.println("ERROR! The method selectAntecedent() returned null!");
        } else
            setAnaphoric(false);
        Local.out.println("PossessivePronoun", "PP END...");
        return result;
    } //end resolve anaphor


    /**
     * A baseline algorithm for processing Personal Pronouns. Takes as an antecedent the first Cf
     * that syntactically agrees with the anaphor. The processing goes backwards utterance by utterance
     * and from left to right over the list of Cfs of a given utterance.
     * @param dm The Discourse Model to which this Cf belongs
     * @return Cf The proposed antecedent
     */
    public Cf firstMatchresolveAnaphor(DiscourseModel dm) {
        //public Cf resolveAnaphor( DiscourseModel dm ) {
        Cf result = null;
        setAnaphoric(false); //default
        Utterance utterance = getUtterance();
        for (int distance = 0; utterance != null && distance < WINDOW_SIZE; distance++) { //loop over utterances backwards
            Vector vCfs = dm.getCfs(utterance);
            for (int j = 0; j < vCfs.size(); j++) { //loop over Cfs forwards
                Cf cfAnte = (Cf) vCfs.elementAt(j);
                NominalGroup antecedent = (NominalGroup) cfAnte;
                if (this.getAgreement().equals(antecedent.getAgreement())) {
                    setAnaphoric(true);
                    result = antecedent;
                }
            } //end inner for
            utterance = dm.getPrevUtterance(utterance);
        } //end outer for
        return result;
    } //end resolve anaphor


    /**
     * A gold standard algorithm for processing Personal Pronouns based on corpus annotation.
     * @param dm The Discourse Model to which this Cf belongs
     * @return Cf The antecedent
     */
    public Cf goldStandardresolveAnaphor(DiscourseModel dm) {
        //public Cf resolveAnaphor( DiscourseModel dm ) {
        Cf result = null;
        DiscourseModel flatDM = dm.getFlatDM();
        if (flatDM == null) {
            Vector<String> categoryName = new Vector<String>();
            categoryName.add("pers-pro");
            categoryName.add("the-np");
            dm.setFlatDM(new DiscourseModelImplementer(dm.getDOMDocument(), "ante", "cat", categoryName));
            flatDM = dm.getFlatDM();
        }
        Node anaphor = getNode();
        String anaphorId = ((Element) anaphor).getAttribute(IOXMLUtils.ID_PROPERTY_NAME);
        String anteId = (String) flatDM.getAntecedent(anaphorId);
        //Look for the Cf with the given id
        result = findCfById(anteId, dm); //a method in class Cf
        setAnaphoric(result != null);
        return result;
    } //end resolve anaphor


    /**
     * A recursive method to select the preferred antecedent.
     * @param anteSet the sorted Set of antecedents to be searched
     * @param nextCriteria the score criteria to be considered in case of a tie
     * @return NominalGroup the resulting antecedent
     */
    private NominalGroup selectAntecedent(Set anteSet, int nextCriterion) {
        if (anteSet.size() == 1)
            return (((PotentialAntecedent) anteSet.iterator().next()).getCorrespondingNE());
        else {
            TreeSet<PotentialAntecedent> newAntecedents = new TreeSet<PotentialAntecedent>();
            PotentialAntecedent firstElement = (PotentialAntecedent) anteSet.iterator().next();
            int maxScore = firstElement.getScore(nextCriterion - 1);
            firstElement.setCriterion(nextCriterion);
            newAntecedents.add(firstElement);
            boolean stop = false;
            Iterator it = anteSet.iterator();
            while (!stop && it.hasNext()) {
                PotentialAntecedent currentElement = (PotentialAntecedent) it.next();
                if (currentElement.getScore(nextCriterion - 1) != maxScore) stop = true;
                else {
                    currentElement.setCriterion(nextCriterion);
                    newAntecedents.add(currentElement);
                }
            }
            if (newAntecedents.size() == 1)
                return (((PotentialAntecedent) newAntecedents.first()).getCorrespondingNE());
            else { //There is a tie
                if (nextCriterion < PotentialAntecedent.POSSIBLE_CRITERIA - 1)
                    return (selectAntecedent(newAntecedents, nextCriterion + 1));
            }
        } //end main else
        return null; //It is not supposed to get here
    }


    ////////////////////////////////////////// INTERNAL CLASSES /////////////////////////////////////////
    /**
     * An internal class to store the antecedent preferences ("indicators") and to provide the sorting
     * functionality for the potential antecedents.(Note: here I have placed and adapted code from class NounGroup
     * from my MSc project, so in some way NounGroup maps to this class)
     */
    class PotentialAntecedent implements Comparable {
        //For the score criteria returned by getScore()
        /** Index of the antecedent indicator of that name */
        public static final int DEFINITENESS = 0;
        /** Index of the antecedent indicator of that name */
        public static final int GIVENNESS = 1;
        /** Index of the antecedent indicator of that name */
        public static final int INDICATING_VERBS = 2;
        /** Index of the antecedent indicator of that name */
        public static final int LEXICAL_REITERATION = 3;
        /** Index of the antecedent indicator of that name */
        public static final int SECTION_HEADING_PREFERENCE = 4;
        /** Index of the antecedent indicator of that name */
        public static final int NON_PREPOSITIONAL_NP = 5;
        /** Index of the antecedent indicator of that name */
        public static final int COLLOCATION_PATTERN = 6;
        /** Index of the antecedent indicator of that name */
        public static final int IMMEDIATE_REFERENCE = 7;
        /** Index of the antecedent indicator of that name */
        public static final int REFERENTIAL_DISTANCE = 8;
        /** Index of the antecedent indicator of that name */
        public static final int TERM_PREFERENCE = 9;
        /** Number of antecedent indicators */
        public static final int NUMBER_OF_INDICATORS = 10;
        /** Index of the aggregate score (sum of antecedent indicators) */
        public static final int SCORE = 10;
        /** Number of possible criteria for sorting potential antecedents */
        public static final int POSSIBLE_CRITERIA = 5;

        //Instance variables
        /** The aggregate score of this antecedent (sum of all indicators) */
        private int score;
        /** The criterion to be used for sorting the antecedents */
        private int sortCriterion;
        /** The original Cf from which this object was created */
        private NominalGroup ne;
        /** The array of antecedent indicators */
        private int[] antecedentIndicators;
        /** The Discourse Model */
        private DiscourseModel dModel;
        /** Pointer to the parent utterance of this potential antecedent */
        Utterance parentUtterance;

        /**
         * Creates an instance of PotentialAntecedent and sets the antecedent indicators.
         * @param theNE The original ne, which will be instantiated as a PotentialAntecedent
         * @param dm The Discourse Model
         * @param theCriterion The sorting criterion
         */
        public PotentialAntecedent(NominalGroup theNE, DiscourseModel dm, int theCriterion, Utterance utt) {
            ne = theNE;
            dModel = dm;
            sortCriterion = theCriterion;
            parentUtterance = utt;
            String anteLeftCollocate = IOXMLUtils.getLeftCollocate(ne);
            String anteRightCollocate = IOXMLUtils.getRightCollocate(ne);
            String preposition = getPreposition();
            int[] indicators = ne.getAntecedentIndicators();
            if (indicators != null)
                antecedentIndicators = indicators;
            else {
                antecedentIndicators = new int[NUMBER_OF_INDICATORS];
                setFixedAntecedentIndicators(anteLeftCollocate, preposition, getTimesMentioned()); //Initialise independent indicators
            }
            //Initialise dynamic indicators - those that depend upon the anaphor context
            setDynamicAntecedentIndicators(anteLeftCollocate, anteRightCollocate);
            updateScore();
            ne.setAntecedentIndicators(antecedentIndicators);
        }


        /**
         * Returns the ne corresponding to this PotentialAntecedent instance.
         * @return NominalGroup the corresponding NE.
         */
        public NominalGroup getCorrespondingNE() {
            return ne;
        }


        /**
         * Sets the sorting criterion of this Potential Antecedent.
         * @param criterion The sorting criterion
         */
        public void setCriterion(int criterion) {
            sortCriterion = criterion;
        }


        /**
         * Sums antecedent indicators up to obtain the aggregate score.
         */
        public void updateScore() {
            score = 0; //This is an instance (global) variable
            for (int i = 0; i < antecedentIndicators.length; i++) score += antecedentIndicators[i];
        } //end update score


        /**
         * Returns the current aggregate score of this potential antecedent or the value of a specific indicator.
         * @param criterion Specifies what score to be returned
         * @return int The aggregate score or value of specific indicator
         */
        public int getScore(int criterion) {
            int result = -1;
            if (criterion >= antecedentIndicators.length) result = score;
            else result = antecedentIndicators[criterion];
            return result;
        }


        /**
         * Override the compareTo method to achieve a desired natural order of the elements of this class.
         * The desired order is controlled by the instance variable sortCriterion and is either of the following:
         * aggregate score, immediate reference, collocational pattern, indicating verbs and referential distance
         * (see Mitkov's paper for details).
         */
        public int compareTo(Object obj) {
            if (obj instanceof PotentialAntecedent) {
                Integer thisScore = new Integer(getScore(sortCriterion));
                Integer objScore = new Integer(((PotentialAntecedent) obj).getScore(sortCriterion));
                //For sorting into descending order, hence * (-1)
                return (thisScore.compareTo(objScore) * ( -1));
            } else
                throw new ClassCastException("PotentialAntecedent != " + obj.getClass().getName());
        } //end compareTo


        /////////////////////////////////////////////////////////////////////////////////////////////////
        ////////////////////// METHODS FOR SETTING THE ANTECEDENT INDICATORS ////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////////////

        /*** Fixed Antecedent Indicators (do not dependent on anaphor's context) ***/

        /*
         * An auxiliary method used only for holding the calls to methods that set the following
         * antecedent indicators: definiteness, givenness, indicating verbs, section heading,
         * non-prepositional np, lexical reiteration and term preference.
         * @param anteLeftCollocate The first verb to the left of this potential antecedent
         * @param preposition The preceeding preposition if any
         * @param timesMentioned The number of mentions of the entity, to which this antecedent refers to
         */
        private void setFixedAntecedentIndicators(String anteLeftCollocate, String preposition, int timesMentioned) {
            Local.out.println("PersonalPronoun", "FIXED INDICATORS BEGINS...");
            //Possible scores (-1, 0, 1, 2)
            setDefiniteness();
            Local.out.println("PersonalPronoun", "DEFINITENESS FINISHED");
            setGivenness(anteLeftCollocate, preposition);
            Local.out.println("PersonalPronoun", "GIVENNESS FINISHED");
            setIndicatingVerbs(anteLeftCollocate);
            Local.out.println("PersonalPronoun", "INDICATING VERBS FINISHED");
            setSectionHeadingPreference();
            Local.out.println("PersonalPronoun", "SECTION HEADING FINISHED");
            setNon_prepositionalNP(preposition);
            Local.out.println("PersonalPronoun", "NONPREPOSITIONAL FINISHED");
            setLexicalReiteration(timesMentioned);
            Local.out.println("PersonalPronoun", "LEXICAL REITERATION FINISHED");
            setTermPreference();
            Local.out.println("PersonalPronoun", "FIXED... END... ");
        }


        /*
         * Sets the indicator for definiteness. NPs containing the, this, that, or Possessive pronoun
         * score 0, all other NPs score -1 (eg. presence of a, an).
         */
        private void setDefiniteness() {
            byte type = ne.getType();
            if (Arrays.binarySearch(NominalGroup.DEFINITE, type) >= 0 ||
                type == NominalGroup.POSS_NP)
                antecedentIndicators[DEFINITENESS] = 0;
            else
                antecedentIndicators[DEFINITENESS] = -1;
        }


        /*
         * Sets the indicator for givenness. First noun phrase in a non-imperative sentence scores 1, any other 0.
         * @param anteLeftCollocate The first verb to the left of this potential antecedent
         * @param preceedingPreposition The preceeding preposition if any
         */
        private void setGivenness(String anteLeftCollocate, String preceedingPreposition) {
            //If there is no other NP, preposition, or verb before this NP
            int indexCf = 0;
            Vector vCfs = dModel.getCfs(parentUtterance);
            for (Cf cf = (Cf) vCfs.elementAt(0); cf != ne; indexCf++, cf = (Cf) vCfs.elementAt(indexCf));
            if ((indexCf == 0 && anteLeftCollocate == null && preceedingPreposition == null)) //theme
                antecedentIndicators[GIVENNESS] = 1;
            else //rheme
                antecedentIndicators[GIVENNESS] = 0;
        }


        /**
         * Sets the indicator for "indicating verbs".
         * If the verb to the left of this NP is a member of a predefined set of verbs (see VERB_LIST[]).
         * @param anteLeftCollocate The first verb to the left of this potential antecedent
         */
        private void setIndicatingVerbs(String anteLeftCollocate) {
            Vector<String> verbList = new Vector<String>();
            for (int i = 0; i < VERB_LIST.length; i++)
                verbList.addElement(VERB_LIST[i]);
            antecedentIndicators[INDICATING_VERBS] = 0; //default
            if (anteLeftCollocate != null)
                if (verbList.contains(anteLeftCollocate))
                    antecedentIndicators[INDICATING_VERBS] = 1;
        }


        /**
         * Sets the indicator for "section heading preference".
         * If the NP occurs in the heading of a section (in our corpus title tag).
         */
        private void setSectionHeadingPreference() {
            //If ne occurs in the heading of the section
            //Consider the title tags if available
            Node node = ne.getNode();
            boolean partOfTitle = false;
            while (!partOfTitle && node != null) {
                if (node.getNodeName() == IOXMLUtils.TITLE_TAG) partOfTitle = true;
                node = node.getParentNode();
            }
            if (partOfTitle)
                antecedentIndicators[SECTION_HEADING_PREFERENCE] = 1;
            else
                antecedentIndicators[SECTION_HEADING_PREFERENCE] = 0;
        }


        /*
         * Sets the indicator for non-prepositional NPs. A non-prepositional NP is given higher preference.
         * @param preceedingPreposition The preceeding preposition if any
         */
        private void setNon_prepositionalNP(String preceedingPreposition) {
            if (preceedingPreposition != null)
                antecedentIndicators[NON_PREPOSITIONAL_NP] = -1;
            else
                antecedentIndicators[NON_PREPOSITIONAL_NP] = 0;
        }


        /**
         * Sets the indicator for "lexical reiteration".
         * The salience of a referent that was accessed (refered to) more than once is boosted.
         * @param timesMentioned The number of mentions of the entity, to which this antecedent refers to
         */
        private void setLexicalReiteration(int timesMentioned) {
            switch (timesMentioned) {
            case 0:
                antecedentIndicators[LEXICAL_REITERATION] = 0;
                break;
            case 1:
                antecedentIndicators[LEXICAL_REITERATION] = 1;
                break;
            default: //2 or more
                antecedentIndicators[LEXICAL_REITERATION] = 2;
            }
        }


        /**
         * Sets the indicator for "term preference".
         * NPs representing terms in the field (technical terminology) are preferred, but for the
         * moment no implementation is available.
         */
        private void setTermPreference() {
            antecedentIndicators[TERM_PREFERENCE] = 0;
        }


        /**
         * Counts the number of times this PotentialAntecedent was mentioned in the previous sentences.
         * A fixed window of sentences is used to limit the search.
         * @return int Number of lexical reiterations
         */
        private int getTimesMentioned() {
            return (dModel.getTimesMentioned(dModel.getDiscourseEntity(ne)));
        }


        /**
         * Find the preceeding word of this PotentialAntecedent, and if it is a preposition return it,
         * otherwise return null.
         * @return String The preceeding preposition, if no such, null
         */
        private String getPreposition() {
            /********** I need to change "ante" for a global constant in IOXMLUtils **************/
            String result = null;
            Node uttNode = ne.getUtterance().getNode();
            Node node = ne.getNode();
            Local.out.println("PersonalPronoun", "FRASEorg: " + ((Element) node).getAttribute("id"));
            Node prevSibling = null;
            while (prevSibling == null) {
                prevSibling = node.getPreviousSibling();
                if (prevSibling instanceof org.w3c.dom.Text
                    || prevSibling instanceof org.w3c.dom.Comment
                    || (prevSibling != null && prevSibling.getNodeName().equals("ante"))) {
                    node = prevSibling;
                    prevSibling = null; //keep on looping
                } else
                    node = node.getParentNode();
                if (node == uttNode)break;
            }
            if (prevSibling != null) {
                Node wordNode = prevSibling;
                String nodeName = prevSibling.getNodeName();
                if (!nodeName.equalsIgnoreCase(IOXMLUtils.WORD_TAG)) { //if not a word, look for one
                    NodeList nl = ((Element) prevSibling).getElementsByTagName(IOXMLUtils.WORD_TAG);
                    int numWords = nl.getLength();
                    Local.out.println("PersonalPronoun", "numwords: " + numWords);
                    wordNode = nl.item(numWords - 1); //Get last word
                }
                Local.out.println("PersonalPronoun", "ORACION: " + ((Element) uttNode).getAttribute("id"));
                Local.out.println("PersonalPronoun", "previousName: " + prevSibling.getNodeName());
                Local.out.println("PersonalPronoun", "previous: " + ((Element) prevSibling).getAttribute("id"));
                Local.out.println("PersonalPronoun", "FRASE: " + ((Element) node).getAttribute("id"));
                String pos = ((Element) wordNode).getAttribute(TagSet.POS_PROPERTY_NAME);
                //if the word is a preposition
                if (Local.containsIgnoreCase(TagSet.PREPOSITION, pos))
                    result = wordNode.getFirstChild().getNodeValue().trim().toLowerCase();
            }
            return result;
        }


        /*** Dynamic Antecedent Indicators (dependent on anaphor's context) ***/

        /*
         * An auxiliary method used only for holding the calls to methods that set the following
         * antecedent indicators: collocation pattern, immediate reference and referential distance.
         * @param anteLeftCollocate The verb to the left of this antecedent
         * @param anteRightCollocate The verb to the right of this antecedent
         */
        private void setDynamicAntecedentIndicators(String anteLeftCollocate,
                String anteRightCollocate) {
            //Possible scores (-1, 0, 1, 2)
            Local.out.println("PersonalPronoun", "DYNAMIC INDICATORS BEGINS...");
            setCollocationPattern(anteLeftCollocate, anteRightCollocate); //These are the anaphor's left and right collocates
            Local.out.println("PersonalPronoun", "COLLOCATIONAL PATTERN FINISHED");
            setImmediateReference();
            Local.out.println("PersonalPronoun", "IMMEDIATE REFERENCE FINISHED");
            setReferentialDistance();
            Local.out.println("PersonalPronoun", "DYNAMIC... END...");
        }


        /**
         * Sets the indicator for "collocation pattern".
         * If the NP occurs in the same collocational pattern.
         * Receives as parameters the verb to the left and to the right of this antecedent (if any),
         * and compares them with the verb to the left and right (if any) of the anaphor, which are
         * stored as instance variables of PersonalPronoun.
         * @param anteLeftCollocate The verb to the left of this antecedent
         * @param anteLeftCollocate The verb to the right of this antecedent
         */
        private void setCollocationPattern(String anteLeftCollocate,
                                           String anteRightCollocate) {
            antecedentIndicators[COLLOCATION_PATTERN] = 0;
            if (anteLeftCollocate != null && leftCollocate != null) {
                if (anteLeftCollocate.compareTo(leftCollocate) == 0)
                    antecedentIndicators[COLLOCATION_PATTERN] = 2;
            }

            if (antecedentIndicators[COLLOCATION_PATTERN] == 0) {
                if (anteRightCollocate != null && rightCollocate != null) {
                    if (anteRightCollocate.compareTo(rightCollocate) == 0)
                        antecedentIndicators[COLLOCATION_PATTERN] = 2;
                }
            }
        }


        /**
         * Sets the indicator for "immediate reference".
         * This indicator deals with constructions of the form (You)V1NP...conjV2it conj...,
         * but for the moment no implementation is available.
         */
        private void setImmediateReference() {
            //That is related
            antecedentIndicators[IMMEDIATE_REFERENCE] = 0;
        }


        /**
         * Sets the indicator for "referential distance".
         * This is the recency preference. If the potential antecedent is within the same utterance of
         * the anaphor it scores 2, previous utterance scores 1, one before that 0, and further back -1.
         */
        private void setReferentialDistance() {
            int option = 0;
            //getUtterance() returns the utterance of the pronoun being resolved
            //parentUtterance is the utterance of this potential antecedent
            for (Utterance utt = getUtterance(); utt != null && utt != parentUtterance; utt = dModel.getPrevUtterance(utt), option++);
            switch (option) {
            case 0: //same sentence
                antecedentIndicators[REFERENTIAL_DISTANCE] = 2;
                break;
            case 1: //previous sentence
                antecedentIndicators[REFERENTIAL_DISTANCE] = 1;
                break;
            case 2: //sentence before the previous one
                antecedentIndicators[REFERENTIAL_DISTANCE] = 0;
                break;
            default: //further before
                antecedentIndicators[REFERENTIAL_DISTANCE] = -1;
            }
        }

///////////////////////////////////////////////////////////
    } //end internal class PotentialAntecedent

} //End class PersonalPronoun
