/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR;

import java.util.*;

import org.w3c.dom.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.*;

/**
 * A class that encapsulates the functionality for processing Proper Nouns.
 *
 * Examples of Proper Noun to be considered:
 *    Herman Melville; Mrs. Hacket; Snoopy; Oreo; Orlando; Paco's Mexican Cafe; Declaration of Independence; University of Virginia;...
 *
 * @author Philippe J. Goux
 * @version 1.0
 */
public class ProperName extends NominalGroup {
    /**
     * Constructor. Just calls the constructor of NominalGroup.
     * @param theNode Node
     * @param theUtterance Utterance
     * @param theType the type
     */
    public ProperName(Node theNode, Utterance theUtterance, byte theType) {
        super(theNode, theUtterance, theType);
    }


    /**
     * The implementation of ...'s algorithm for processing Proper nouns.
     * @param dm The Discourse Model to which this Cf belongs
     * @return Cf The proposed antecedent
     */
    public Cf resolveAnaphor(DiscourseModel dm) {
        Local.out.println("ProperName", "\n===================================");
        Local.out.println("ProperName", "PROPER NAME RESOLUTION BEGINS...");

        // Get Cf to work on
        Cf thisCf = this; // Default Cf

        // If head not null, get discourse entity
        DiscourseEntity deAnaphor = null;
        if (getHeadNode() != null)
            deAnaphor = dm.getDiscourseEntity(thisCf); // used to be this

        // Get element from node of default Cf
        Element thisElement = (Element)this.getNode();
        Local.out.println("ProperName", "- Cf:\t\t\"" + IOXMLUtils.getPhraseString(thisElement) + "\"");
        Local.out.println("ProperName", "- Name set:\t" + deAnaphor.getNames().size() + deAnaphor.getNames());
        Local.out.println("ProperName", "- Type:\t\t" + NominalGroup.getStringType(thisCf.getType()));
        Local.out.println("ProperName", "- ID:\t\t" + thisElement.getAttribute(IOXMLUtils.ID_PROPERTY_NAME));

        // Get current utterance
        Utterance utterance = this.getUtterance();

        // Start utterance loop
        Local.out.println("ProperName", "\nLOOKING FOR ANTECEDANT...");
        for (int distance = 0; utterance != null; distance++) { //loop over utterances backwards
            // Get vector of Cfs creating utterance
            Vector<Cf> vCfs = dm.getCfs(utterance);

            // Start Cf loop forwards
            for (int j = 0; j < vCfs.size(); j++) {
                Cf cfAnte = (Cf) vCfs.elementAt(j);

                // If the Cf analysed is the same as this Cf, break out
                if (cfAnte.equals(thisCf))
                    break;

                byte anteCfType = cfAnte.getType();
                if (anteCfType == NominalGroup.PN || anteCfType == NominalGroup.THE_PN) {
                    DiscourseEntity deAnte = dm.getDiscourseEntity(cfAnte);

                    Local.out.print("ProperName", "- Checking:\t\t" + deAnte.getNames().size() + deAnte.getNames());
                    Local.out.println("ProperName", "; in: \"" + IOXMLUtils.getPhraseString((Element) cfAnte.getNode()) + "\""); // get first element in utterance

                    boolean isAntecedant = passRulesForAll(deAnaphor, deAnte);
                    if (!isAntecedant)
                        isAntecedant = passRulesForOrganisationsAndPersons(deAnaphor, deAnte);

                    if (isAntecedant) {
                        Element anteCfElement = (Element) cfAnte.getNode();
                        Local.out.print("ProperName", "ANTECEDANT FOUND.");
                        Local.out.println("ProperName", " ID: " + anteCfElement.getAttribute(IOXMLUtils.ID_PROPERTY_NAME));
                        Local.out.println("ProperName", "===================================\n");

                        setAnaphoric(true);
                        return cfAnte; // Return antecendant found
                    }
                }
            } // End Cf loop

            // Get previous utterance if any
            utterance = dm.getPrevUtterance(utterance);
        } // End utterance loop

        Local.out.println("ProperName", "NO ANTECEDANT FOUND.");
        Local.out.println("ProperName", "===================================\n");
        return null;
    }


    /**
     * Here are the rules that apply for all types of named entities.
     * @param de DiscourseEntity
     * @param deAntecedant DiscourseEntity
     * @return boolean
     */
    private boolean passRulesForAll(DiscourseEntity de, DiscourseEntity deAntecedant) {
        // Retreive names for names for current DiscourseEntity and antecedant DiscourseEntity
        Set<Set<Property>> propertiesSet = de.getNames();
        Set<Set<Property>> propertiesSetAnte = deAntecedant.getNames();

        for (Iterator<Set<Property>> i = propertiesSet.iterator(); i.hasNext(); ) {
            Set<Property> properties = i.next();
            String[] names = setToStrings(properties);
            String name = setToString(properties);
            for (Iterator<Set<Property>> j = propertiesSetAnte.iterator(); j.hasNext(); ) {
                Set<Property> propertiesAnte = j.next();
                String[] namesAnte = setToStrings(propertiesAnte);
                String nameAnte = setToString(propertiesAnte);

                // ERROR CORRECTION: some cfs may be PN but the name Propriety is null.
                // Verify that antecedants and current name are not null.
                if (names.length < 1 || namesAnte.length < 1)
                    return false;

                /**
                 *  exact match: two identical occurrences of the same named entity corefer.
                 **/
                if (name.equalsIgnoreCase(nameAnte)) {
                    return true;
                }

                /**
                 *  equivalent, as defined in a synonym list: this rule is used to handle matching
                 *  of names like IBM and The Big Blue. This synonym list can be extended easily with
                 *  new equivalent names without recompiling the orthomatcher.
                 **/

                /**
                 *  possessives: handles named entities in possessive form, e.g., New York and New Yorks.
                 **/

                /**
                 *  spurious, as defined in a list of spurious names. This rule prevents matching
                 *  entities which have similar names but are otherwise different. Most frequently
                 *  this is needed for companies, where a daughter companys name contains the parent
                 *  company name, e.g., BT Cellnet and BT.
                 **/

		/**
		 * MAK: This is the first match rule. see next method.
		 */
                if (name.startsWith(nameAnte) ||
                    nameAnte.startsWith(name)) {
                    return true;
                }
            }
        }
        return false;
    }


    /**
     * Here are some of the rules that apply to organisations and persons.
     * @param de DiscourseEntity
     * @param deAntecedant DiscourseEntity
     * @return boolean
     */
    private boolean passRulesForOrganisationsAndPersons(DiscourseEntity de, DiscourseEntity deAntecedant) {
        // Retreive names for names for current DiscourseEntity and antecedant DiscourseEntity
        Set<Set<Property>> propertiesSet = de.getNames();
        Set<Set<Property>> propertiesSetAnte = deAntecedant.getNames();

        for (Iterator<Set<Property>> i = propertiesSet.iterator(); i.hasNext(); ) {
            Set<Property> properties = i.next();
            String[] names = setToStrings(properties);
            String name = setToString(properties);
            for (Iterator<Set<Property>> j = propertiesSetAnte.iterator(); j.hasNext(); ) {
                Set<Property> propertiesAnte = j.next();
                String[] namesAnte = setToStrings(propertiesAnte);
                String nameAnte = setToString(propertiesAnte);

                // ERROR CORRECTION: some cfs may be PN but the name Propriety is null.
                // Verify that antecedants and current name are not null.
                if (names.length < 1 || namesAnte.length < 1)
                    return false;

                //  word token match: do all word tokens match, ignoring punctuation and word order,
                // e.g., Kalina Bontcheva and Bontcheva, Kalina.
                if (namesAnte.length == names.length) {
                    if (propertiesAnte.containsAll(properties))
                        return true;
                }

                //  first token match: does the first token in one name match the first token in
                //  the other, e.g., Peter Smith and Peter. This rule, however, which was originally
                //  developed in LaSIE to apply for both people and organisations, had to be modified
                //  in order to work correctly for people, because it also matched wrongly Peter Kline
                //  and Peter Smith. This problem was corrected by allowing this rule to fire for persons,
                //  only if the shorter name has one token.
                if (namesAnte.length < names.length) {
                    if (name.startsWith(nameAnte))
                        return true;
                } else if (names.length < namesAnte.length) {
                    if (nameAnte.startsWith(name))
                        return true;
                }

                //  acronyms (organisations only): handles acronyms like
                // International Business Machines and IBM.
                if (namesAnte.length == 1) {
                    if (isAcronymOf(nameAnte, properties))
                        return true;
                } else if (names.length == 1) {
                    if (isAcronymOf(name, propertiesAnte))
                        return true;
                }

                //  last token match: does the last token in one name match the other name (which must be one token only),
                //  e.g., John Smith and Smith. Shallow Methods for Named Entity Coreference Resolution
                if (namesAnte.length == 1) {
                    if (nameAnte.equals(names[names.length - 1]))
                        return true;
                } else if (names.length == 1) {
                    if (name.equals(namesAnte[namesAnte.length - 1]))
                        return true;
                }

                //  prepositional phrases: matches organisation names which are inverted around a preposition,
                //  e.g., University of Sheffield and Sheffield University.
                if (namesAnte.length == 3 && names.length == 2) {
                    String strPOS = getLexicalWord(properties, 1).getPOSTag();
                    if (Local.containsIgnoreCase(TagSet.PREPOSITION, strPOS) &&
                        namesAnte[0].equalsIgnoreCase(names[1]) &&
                        namesAnte[2].equalsIgnoreCase(names[0]))
                        return true;
                } else if (names.length == 3&& namesAnte.length == 2) {
                    String strPOS = getLexicalWord(propertiesAnte, 1).getPOSTag();
                    if (Local.containsIgnoreCase(TagSet.PREPOSITION, strPOS) &&
                        names[0].equalsIgnoreCase(namesAnte[1]) &&
                        names[2].equalsIgnoreCase(namesAnte[0]))
                        return true;
                }

                //  abbreviations: matches organisation names, where one name is an abbreviation of the other,
                //  e.g., Pan American and Pan Am.

                //  multi-word name matching: there are several new rules for matching person names when they have
                //  more than two tokens and also multi-word company names. The most important rule here is whether
                //  all tokens in the shorter name match tokens in the longer name,
                //  e.g., Second Force Recon Company and Force Recon Company.
            }
        }
        return false;
    }


    /**
     * Attempt to determine if string acronym is an acronym for the name contained
     * in the set names. Only consider PN for acronyms!
     * eg. ACER = Advisory Committee on Environmental Resources
     * @param acronym String
     * @param names Set
     * @return boolean
     */
    private boolean isAcronymOf(String acronym, Set<Property> names) {
        int length = acronym.length(), position = 0;
        Iterator iterator = names.iterator();
        while (iterator.hasNext() && position < length) {
            LexicalWord lexicalWord = (LexicalWord) ((Property) iterator.next()).getPredicate();
            if (Local.containsIgnoreCase(TagSet.VALID_PROPER_NOUN_TAGS, lexicalWord.getPOSTag())) {
                if (acronym.charAt(position) != lexicalWord.getWord().charAt(0))
                    return false;
            }
        }
        return true;
    }


    /**
     * Retreive name <code>String</code> from <code>Set<Property></code> object
     * @param names Set<Property>
     * @return String
     */
    private String setToString(Set<Property> names) {
        String name = "";
        for (Iterator<Property> i = names.iterator(); i.hasNext(); ) {
            name += ((LexicalWord) i.next().getPredicate()).getWord() +
                    ((i.hasNext()) ? " " : "");
        }
        return name;
    }


    private String[] setToStrings(Set<Property> names) {
        String[] name = new String[names.size()];
        int j = 0;
        for (Iterator<Property> i = names.iterator(); i.hasNext(); j++) {
            name[j] = ((LexicalWord) i.next().getPredicate()).getWord();
        }
        return name;
    }


    private LexicalWord getLexicalWord(Set<Property> properties, int position) {
        LexicalWord lexicalWord = null;
        Iterator<Property> iterator = properties.iterator();
        for(int i = 0; iterator.hasNext(); i++) {
            Property property = iterator.next();
            if(i == position)
                lexicalWord = (LexicalWord) property.getPredicate();
        }
        return lexicalWord;
    }


} //End class NominalGroup
