/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR.eval;

//For DOM Tree manipulation
import java.util.*;

import org.w3c.dom.*;
import uk.ac.essex.malexa.nlp.NLPEvaluationMeasures;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.Local;

/**
 * A class that encapsulates the functionality for evaluating the performance of
 * the Anaphora Resolution Engine running over an annotated corpus.
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 1.2
 */
public class AnaphoraResolutionEvaluation {
    private static int[] performance = new int[] {0, 0, 0, 0, 0, 0};

    int correctlyResolved = 0;
    Map<String, Object> wrongAnaphoricLinks = new HashMap<String, Object>(); //the anaphor and antecedent are not in the same coref chain
    Map<String, Object> spuriousAnaphoricLinks = new HashMap<String, Object>(); //the the-np is not anaphoric
    Map<String, Object> missingAnaphoricLinks = new HashMap<String, Object>(); //no antecedent was proposed
    DiscourseModel dmOriginal = null;
    DiscourseModel dmAlgorithm = null;

    /*   public AnaphoraResolutionEvaluation() {
//open a database connection
       }
     */

    /**
     * Given a corpus file, which contains original annotation and automatic annotation, this method
     * computes and displays performance statistics of the latter with reference to the former.
     * @param inputFileName The name of the file to be processed
     * @param categoryName A vector of types of NPs to be considered (pers-pro, the-np, etc.)
     */
    public void processFile(String inputFileName, Vector<String> categoryName) {
        Document domDocument = IOXMLUtils.load(inputFileName); //Load XML file
        process(domDocument, inputFileName, categoryName);
    }


    /**
     * Given a corpus file, which contains original annotation and automatic annotation, this method
     * computes and displays performance statistics of the latter with reference to the former.
     * @param domDocument Dom Document of the XML to process
     * @param inputFileName The name of the file to be processed (Used in log only)
     * @param categoryName A vector of types of NPs to be considered (pers-pro, the-np, etc.)
     */
    public void process(Document domDocument, String inputFileName, Vector<String> categoryName) {
        String cathegory = "";
        for (Iterator<String> i = categoryName.iterator(); i.hasNext(); )
            cathegory += "\"" + i.next() + ((i.hasNext()) ? "\" & " : "\"");

        displaySeparator();
        Local.out.println("Performance Statistics for file \"" + inputFileName +
                          "\" and type" + ((categoryName.size() > 1) ? "s " : " ") + cathegory + Local.NEWLINE);

        // initialise discourse models
        initialiseDiscourseModels(domDocument, domDocument, categoryName);

        // initialise the list of Cf
        initialiseCfList(domDocument, categoryName);

        // Display results
        displayResults(domDocument, categoryName);
    }


    private void displayResults(Document domDocument, Vector<String> categoryName) {
        String cathegory = "";
        for (Iterator<String> i = categoryName.iterator(); i.hasNext(); )
            cathegory += "\"" + i.next() + ((i.hasNext()) ? "\" & " : "\"");

        Local.out.println(Local.NEWLINE + "A. Classifying Anaphoric " + cathegory + ":");
        displayAnaphoric();

        Local.out.println(Local.NEWLINE + "B.1. General resolution performance:");
        displayFilePerformance(cathegory);

        Local.out.println(Local.NEWLINE + "B.2. Detail of mis-matches:");
        Map<String, Element> mapIdToNode = IOXMLUtils.buildIdToNodeMap(domDocument, "ne");

        displayWrongMatches(mapIdToNode);
        displayNoMatches(mapIdToNode);
        displaySpuriousMatches(mapIdToNode);

        dmOriginal.printEquivalenceClassesStatistics(dmAlgorithm);
    }


    private void displayAnaphoric() {
        Local.out.println("Precision, " + NLPEvaluationMeasures.precision(dmAlgorithm.getSetOfAnaphors(), dmOriginal.getSetOfAnaphors()));
        Local.out.println("Recall, " + NLPEvaluationMeasures.recall(dmAlgorithm.getSetOfAnaphors(), dmOriginal.getSetOfAnaphors()));
        Local.out.println("F Measure, " + NLPEvaluationMeasures.fMeasure(dmAlgorithm.getSetOfAnaphors(), dmOriginal.getSetOfAnaphors()));
    }


    private void displayFilePerformance(String cathegory) {
        int[] filePerformance = new int[] {
                                correctlyResolved,
                                dmOriginal.getNumberOfAnaphoricReferences(),
                                dmAlgorithm.getNumberOfAnaphoricReferences(),
                                missingAnaphoricLinks.size(),
                                wrongAnaphoricLinks.size(),
                                spuriousAnaphoricLinks.size()};
        Local.out.print("Correctly Resolved " + cathegory + ", " + filePerformance[0]);
        Local.out.println(", , , , , ," + filePerformance[0] + "," + filePerformance[1] + "," +
                          filePerformance[2] + "," + filePerformance[3] + "," +
                          filePerformance[4] + "," + filePerformance[5]);
        Local.out.println("Anaphoric " + cathegory + " from corpus, " + filePerformance[1]);
        Local.out.println("Anaphoric " + cathegory + " from algorithm, " + filePerformance[2]);

        for (int i = 0; i < performance.length; i++)
            performance[i] += filePerformance[i];
    }


    private void displayNoMatches(Map<String, Element> mapIdToNode) {
        Local.out.println(Local.NEWLINE +
                          "B.2.2. No matches (NM) - No antecedent was proposed");
        Local.out.println("[NEID] Indicates id of the antecedent");
        Local.out.println("NEID,PHRASE,NPTYPE,AANPTYPE,HEAD,MODIFIERS,UTTID");
        for (Iterator<String> it = missingAnaphoricLinks.keySet().iterator(); it.hasNext(); ) {
            String ana = it.next();
            String ante = (String) dmOriginal.getAntecedent(ana);
            Element element = IOXMLUtils.getElementById(ana, mapIdToNode);
            String anaDetails = getInformationString(element);
            element = IOXMLUtils.getElementById(ante, mapIdToNode);
            String anteDetails = getInformationString(element);
            Local.out.println(ana + "," + anaDetails);
            Local.out.println("[" + ante + "]" + "," + anteDetails);
        }
    }


    private void displaySpuriousMatches(Map<String, Element> mapIdToNode) {
        Local.out.println(Local.NEWLINE +
                          "B.2.3. Spurious matches (SM) - An antecedent was proposed for a non-anaphoric NP");
        Local.out.println("[AA_NEID] Indicates id of the proposed antecedent");
        Local.out.println("NEID,PHRASE,NPTYPE,AANPTYPE,HEAD,MODIFIERS,UTTID");
        for (Iterator<String> it = spuriousAnaphoricLinks.keySet().iterator(); it.hasNext(); ) {
            String ana = it.next();
            String aaAnte = (String) spuriousAnaphoricLinks.get(ana);
            Element element = IOXMLUtils.getElementById(ana, mapIdToNode);
            String anaDetails = getInformationString(element);
            element = IOXMLUtils.getElementById(aaAnte, mapIdToNode);
            String aaAnteDetails = getInformationString(element);
            Local.out.println(ana + "," + anaDetails);
            Local.out.println("[AA_" + aaAnte + "]" + "," + aaAnteDetails);
        }
    }


    private void displayWrongMatches(Map<String, Element> mapIdToNode) {
        Local.out.println(Local.NEWLINE +
                          "B.2.1. Wrong matches (WM) - The anaphor and the proposed antecedent are not in the same coref chain");
        Local.out.println("[NEID] Indicates id of the antecedent; [AA_NEID] - proposed antecedent");
        Local.out.println("NEID,PHRASE,NPTYPE,AANPTYPE,HEAD,MODIFIERS,UTTID");
        for (Iterator<String> it = wrongAnaphoricLinks.keySet().iterator(); it.hasNext(); ) {
            String ana = it.next();
            String ante = (String) dmOriginal.getAntecedent(ana);
            String aaAnte = (String) wrongAnaphoricLinks.get(ana);
            Element element = IOXMLUtils.getElementById(ana, mapIdToNode);
            String anaDetails = getInformationString(element);
            element = IOXMLUtils.getElementById(ante, mapIdToNode);
            String anteDetails = getInformationString(element);
            element = IOXMLUtils.getElementById(aaAnte, mapIdToNode);
            String aaAnteDetails = getInformationString(element);
            Local.out.println(ana + "," + anaDetails);
            Local.out.println("[" + ante + "]" + "," + anteDetails);
            Local.out.println("[AA_" + aaAnte + "]" + "," + aaAnteDetails);
        }
    }


    private void initialiseDiscourseModels(Document original, Document resolved, Vector categoryName) {
        Local.out.print("Original Classes:,");
        dmOriginal = new DiscourseModelImplementer(original, "ante", "AAcat", categoryName);
        //DiscourseModel dmOriginal = new DiscourseModelImplementer( domDocument, "ante", "cat", categoryName );

        Local.out.print("Found Classes:,");
        dmAlgorithm = new DiscourseModelImplementer(resolved, IOXMLUtils.ANTE_TAG, "AAcat", categoryName);
        //DiscourseModel dmAlgorithm = new DiscourseModelImplementer( domDocument, IOXMLUtils.ANTE_TAG, "cat", categoryName );
        //Retrieve Cfs (NEs) in preorder (equivalent to a sequential order in the original text)
    }


    private void initialiseCfList(Document domDocument, Vector<String> categoryName) {
        NodeList cfList = domDocument.getElementsByTagName(IOXMLUtils.NP_TAG);
        // Print table headers
        Local.out.println(Local.NEWLINE +
                          "PhraseId,Phrase,OriginalType,OriginalCorefChain,Type,CorefChain");
        for (int i = 0; i < cfList.getLength(); i++) {
            Element cf = (Element) cfList.item(i);

            // Get phrase ID
            String phraseId = cf.getAttribute(IOXMLUtils.ID_PROPERTY_NAME);

            // Get antecedent proposed by algorithm
            Object anteAlgorithm = dmAlgorithm.getAntecedent(phraseId);

            // Check if anaphoric and if correct
            if (dmOriginal.isAnaphoric(phraseId)) {
                // Get original equivalence class
                Set eqClass = dmOriginal.getEquivalenceClass(phraseId);
                if (eqClass.contains(anteAlgorithm))
                    correctlyResolved++;
                else if (anteAlgorithm != null)
                    wrongAnaphoricLinks.put(phraseId, anteAlgorithm);
                else
                    missingAnaphoricLinks.put(phraseId, anteAlgorithm);
            } else if (dmAlgorithm.isAnaphoric(phraseId))
                spuriousAnaphoricLinks.put(phraseId, anteAlgorithm);

            // Get original type
            String originalType = cf.getAttribute("cat");
            // Get found type
            String algorithmType = cf.getAttribute(IOXMLUtils.CATEGORY_PROPERTY_NAME);
            // Get original coreference chain
            String originalCorefChain = dmOriginal.getEquivalenceClassString(phraseId);
            // Get found coreference chain
            String algorithmCorefChain = dmAlgorithm.getEquivalenceClassString(phraseId);
            // Get the corresponding phrase
            String phrase = IOXMLUtils.getPhraseString(cf);

            // Print out information if necessary
            for (Iterator<String> it = categoryName.iterator(); it.hasNext(); ) {
                if (it.next().equalsIgnoreCase(algorithmType)) {
                    Local.out.println(phraseId + "," + "\"" + phrase + "\"" + "," + originalType + "," + "\"" + originalCorefChain + "\"" + "," + algorithmType + "," + "\"" + algorithmCorefChain + "\"");
                    break;
                }
            }
        } //end for

    }


    /**
     * A method that given a DOM node of an NP, extracts 6 features:
     * 1. NP string,
     * 2. NP type (original),
     * 3. NP type (by heuristics),
     * 4. Head,
     * 5. Modifiers,
     * 6. Utterance id
     * @param element The DOM node holgind the NP to be processed
     * @return String The list of extracted features separated by ","
     */
    private String getInformationString(Element element) {
        String phrase = IOXMLUtils.getPhraseString(element);
        String type = element.getAttribute("cat");
        String AAtype = element.getAttribute("AAcat");
        //Get the head
        String head = "";
        Vector nl = IOXMLUtils.getElementsByTagNameNoEmbedding(element,
                IOXMLUtils.NP_TAG, IOXMLUtils.NP_HEAD_TAG);
        if (nl.size() > 0) head = IOXMLUtils.getPhraseString((Node) nl.
                elementAt(0));
        //Get the modifiers
        String modifiers = "";
        nl = IOXMLUtils.getElementsByTagNameNoEmbedding(element,
                IOXMLUtils.NP_TAG, IOXMLUtils.NP_MODIFIER_TAG);
        for (int i = 0; i < nl.size(); i++) {
            modifiers = modifiers + "[" + IOXMLUtils.getPhraseString((Node) nl.elementAt(i)) + "]";
        }
        //get the utterance id
        String uttId = ((Element) DiscourseModelImplementer.findUtteranceNode(
                element)).getAttribute("id");
        return ("\"" + phrase + "\"" + "," + type + "," + AAtype + "," +
                "\"" + head + "\"" + "," + "\"" + modifiers + "\"" + "," + "\"" + uttId + "\"");
    }


    private void displaySeparator() {
        Local.out.println(Local.NEWLINE + Local.NEWLINE +
                          "=========================================================================");
    }


    public void displayPerformance() {
        displaySeparator();
        Local.out.println("RESULTS:");
        Local.out.println("Correctly Resolved," + performance[0]);
        Local.out.println("Original Anaphoric References," + performance[1]);
        Local.out.println("Algorithm Anaphoric References," + performance[2]);
        Local.out.println("Missing Anaphoric Links," + performance[3]);
        Local.out.println("Wrong Anaphoric Links," + performance[4]);
        Local.out.println("Spurious Anaphoric Links," + performance[5]);
        Local.out.println(Local.NEWLINE +
                          "Performance (%)," +
                          (((double) performance[0] * 100) / performance[1]));
    }


    /*** NEW ***/
    /**
     * Given a corpus file and an automatically parsed and annotated file (1.txtToXML, 2.ARE_Runner), this method,
     * by first aligning the corpus nes with the parsed ones, computes and displays performance statistics for the
     * automatic annotation in terms of the parsed nes with reference to the corpus annotation.
     * @param corpusFileName The name of the file from which corpus annotation will be extracted
     * @param inputFileName The name of the automatically parsed and annotated file
     * @param categoryName A vector of types of NPs to be considered (pers-pro, the-np, etc.)
     */
    public void processFile(String corpusFileName, String inputFileName, Vector<String> categoryName) {
        Document domDocumentCorpus = IOXMLUtils.load(corpusFileName); //Load corpus file
        Document domDocumentParsed = IOXMLUtils.load(inputFileName); //Load parsed file

        // Process File
        process(domDocumentCorpus, corpusFileName, domDocumentParsed, inputFileName, categoryName);
    }


    /**
     * Given a corpus file and an automatically parsed and annotated file (1.txtToXML, 2.ARE_Runner), this method,
     * by first aligning the corpus nes with the parsed ones, computes and displays performance statistics for the
     * automatic annotation in terms of the parsed nes with reference to the corpus annotation.
     * @param domDocumentCorpus Dom Document of the XML from which corpus annotation will be extracted
     * @param corpusFileName The name of the file from which corpus annotation will be extracted
     * @param domDocumentParsed Dom Document of the automatically parsed and annotated XML file
     * @param inputFileName The name of the automatically parsed and annotated file
     * @param categoryName A vector of types of NPs to be considered (pers-pro, the-np, etc.)
     */
    public void process(Document domDocumentCorpus, String corpusFileName,
                        Document domDocumentParsed, String inputFileName,
                        Vector<String> categoryName) {
        displaySeparator();
        Local.out.println("PERFORMANCE STATISTICS FOR FILES: " + inputFileName + "; " + corpusFileName);

        // initialise discourse models
        initialiseDiscourseModels(domDocumentCorpus, domDocumentParsed, categoryName);

        // initialise the list of Cf
        initialiseCfList(domDocumentParsed, categoryName);

        // Display results
        displayResults(domDocumentParsed, categoryName);

        /*
                 NodeList cfList = domDocumentParsed.getElementsByTagName(IOXMLUtils.NP_TAG);
                 Local.out.println(Local.NEWLINE +
                          "PhraseId,Phrase,OriginalType,OriginalCorefChain,Type,CorefChain");
                 for (int i = 0; i < cfList.getLength(); i++) {
            Element cf = (Element) cfList.item(i);
            // Get phrase ID
            String phraseId = cf.getAttribute(IOXMLUtils.ID_PROPERTY_NAME);

            // Search dmOriginal for
            Vector candidates = dmOriginal.getVectors(phraseId);
            if (candidates != null) {
                for (int k = 0; k < candidates.size(); k++) {
                    Vector key = (Vector) candidates.elementAt(k);

                    //Check if anaphoric and if correct
                    if (dmOriginal.isAnaphoric(key)) {
                        //Get original equivalence class
                        Set eqClass = dmOriginal.getEquivalenceClass(key);

                        //Get antecedent proposed by algorithm
                        Object anteAlgorithm = dmAlgorithm.getAntecedent(phraseId);
                        if (eqClass.contains(anteAlgorithm)) {
                            correctlyResolved++;
                            break; //stop looping
                        }
                    }
                }
            }
         */
        /*****************
          if( anaphoricNEsOriginal.containsKey( phraseId ) ) {
           //Get original equivalence class
         int index = ((Integer)equivClassesMapOriginal.get( phraseId )).intValue();
         Set eqClass = (Set)equivalenceClassesOriginal.elementAt( index );
           //Get antecedent proposed by algorithm
         String anteAlgorithm = (String)anaphoricNEsAlgorithm.get( phraseId );
           if( eqClass.contains( anteAlgorithm ) ) correctlyResolved++;
           else if( anteAlgorithm != null ) wrongAnaphoricLinks.put( phraseId, anteAlgorithm );
           else missingAnaphoricLinks.put( phraseId, anteAlgorithm );
          } else if( anaphoricNEsAlgorithm.containsKey( phraseId ) ) {
         String anteAlgorithm = (String)anaphoricNEsAlgorithm.get( phraseId );
           spuriousAnaphoricLinks.put( phraseId, anteAlgorithm );
          }
         ****************/
        /*
                     // Get original type
                     String originalType = cf.getAttribute("cat");
                     // Get found type
                     String algorithmType = cf.getAttribute(IOXMLUtils.CATEGORY_PROPERTY_NAME);
                     // Get original coreference chain
                     String originalCorefChain = dmOriginal.getEquivalenceClassString(phraseId);
                     // Get found coreference chain
                     String algorithmCorefChain = dmAlgorithm.getEquivalenceClassString(phraseId);
                     // Get the corresponding phrase
                     String phrase = IOXMLUtils.getPhraseString(cf);

                     Local.out.println(phraseId + "," + "\"" + phrase + "\"" + "," +
                          originalType + "," + "\"" + originalCorefChain + "\"" + "," +
                          algorithmType + "," + "\"" + algorithmCorefChain + "\"");
                     //Store record in database
                 } //end for

                 // Display results
                 Local.out.println(Local.NEWLINE +
                      "A. Classifying Anaphoric (the)-nps: " + categoryName);
                 Local.out.println("Precision, " + NLPEvaluationMeasures.precision(dmAlgorithm.getSetOfAnaphors(), dmOriginal.getSetOfAnaphors()));
                 Local.out.println("Recall, " + NLPEvaluationMeasures.recall(dmAlgorithm.getSetOfAnaphors(), dmOriginal.getSetOfAnaphors()));
                 Local.out.println("F Measure, " + NLPEvaluationMeasures.fMeasure(dmAlgorithm.getSetOfAnaphors(), dmOriginal.getSetOfAnaphors()));
                 Local.out.println(Local.NEWLINE +
                      "B.1. General resolution performance:");
                 Local.out.print("Correctly resolved (the)-nps" + categoryName + ", " + correctlyResolved);
                 Local.out.println(", , , , ," + correctlyResolved + "," +
                      dmOriginal.getNumberOfAnaphoricReferences() + "," +
                      dmAlgorithm.getNumberOfAnaphoricReferences());
                 Local.out.println("Anaphoric (the)-nps " + categoryName + " from corpus, " + dmOriginal.getNumberOfAnaphoricReferences());
                 Local.out.println("Anaphoric (the)-nps " + categoryName + " from algorithm, " + dmAlgorithm.getNumberOfAnaphoricReferences());
                 Local.out.println("B.2. Detail of mis-matches:");
                 Local.out.println("B.2.1. Wrong matches (WM) - The anaphor and the proposed antecedent are not in the same coref chain");
                 Local.out.println("PHRASE,ANTE");
                 Iterator it = wrongAnaphoricLinks.keySet().iterator();
                 while (it.hasNext()) {
                     String ne = (String) it.next();
                     Local.out.println(ne + "," + wrongAnaphoricLinks.get(ne));
                 }
                 Local.out.println(
            "B.2.2. No matches (NM) - No antecedent was proposed");
                 Local.out.println("PHRASE,ANTE");
                 it = missingAnaphoricLinks.keySet().iterator();
                 while (it.hasNext()) {
                     String ne = (String) it.next();
                     Local.out.println(ne + "," + missingAnaphoricLinks.get(ne));
                 }
                 Local.out.println("B.2.3. Spurious matches (SM) - An antecedent was proposed for a non-anaphoric NP");
                 Local.out.println("PHRASE,ANTE");
                 it = spuriousAnaphoricLinks.keySet().iterator();
                 while (it.hasNext()) {
                     String ne = (String) it.next();
                     Local.out.println(ne + "," + spuriousAnaphoricLinks.get(ne));
                 }
         */
        //printEquivalenceClassesStatistics();
    }
} //end class AnaphoraResolutionEvaluation

