/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR.gui;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import java.io.*;
import java.util.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.DiscourseModelImplementer;
import org.w3c.dom.Document;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.IOXMLUtils;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.eval.AnaphoraResolutionEvaluation;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.eval.GenerateErrorsInHTML;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.prepro.PreProSyntacticHeuristics;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.LexicalWord;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.*;
import libsvm.svm;
import libsvm.svm_model;
import libsvm.SVMScalingModel;

/**
 * <b>Description:</b> Used for the GuiTAR tool's GUI.<br>
 * <b>Copyright:</b> Copyright (c) 2005<br>
 * <b>Company:</b> Essex University<br>
 * @author Philippe J. Goux
 * @version 1.0
 */

public class MainFrame extends JFrame implements Gui {
    // ===========================================================================
    // CONSTANTS
    private static final String ID = "Main Frame";

    // ===========================================================================
    // VARIABLES
    public static JFrame mainFrame = null;
    private String outputDirectory = "";
    private BrowsePanel filesPanel;
    private JButton jButton_Process = new JButton();
    private JButton jButton_Close = new JButton();
    private OptionPanel options = new OptionPanel();
    private ResultPanel results = new ResultPanel();
    private JTabbedPane jTabbedPane = new JTabbedPane();


    // ===========================================================================
    // CONSTRUCTORS
    public MainFrame() throws Exception {
        // Set variables
        boolean packFrame = true;
        mainFrame = this;

        // Create history and language file if necessary
        History.setCreateFile(true);
        Language.setCreateFile(true);
        TagSet.setCreateFile(true);

        try {
            UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
        } catch (Exception ex) {
            Local.err.printException(ex);
        }

        enableEvents(AWTEvent.WINDOW_EVENT_MASK);

        initialiseGUI();
        initialiseHistory();
        initialiseLanguage();
        addListeners();

        // Validate frames that have preset sizes
        // Pack frames that have useful preferred size info, e.g. from their layout
        if (packFrame) {
            mainFrame.pack();
        } else {
            mainFrame.validate();
        }

        // Center the window
        Utils.centerWindow(mainFrame);

        // Display window
        mainFrame.setVisible(true);
    }


    // ===========================================================================
    // PUBLIC FUNCTIONS

    //Overridden so we can exit when window is closed
    protected void processWindowEvent(WindowEvent e) {
        super.processWindowEvent(e);
        if (e.getID() == WindowEvent.WINDOW_CLOSING) {
            System.exit(0);
        }
    }


    // ===========================================================================
    // PRIVATE FUNCTIONS
    private JMenuBar createMenu() throws Exception {
        // Create file menu
        JMenu jMenuFile = new JMenu(Language.FILEMENU_FILE);
        JMenuItem jMenuFileExit = new JMenuItem(Language.FILEMENU_EXIT);
        jMenuFileExit.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                System.exit(0);
            }
        });

        // Create language menu
        JMenu jMenuLanguage = new JMenu(Language.FILEMENU_LANGUAGE);
        String[] languages = Language.getLanguages();
        for (int i = 0; i < languages.length; i++) {
            JRadioButtonMenuItem jRadioButtonLanguage = new JRadioButtonMenuItem(languages[i]);
            jRadioButtonLanguage.setActionCommand(languages[i]);
            // Get Icon
            String flag = Language.getValueInLanguage(languages[i], "FLAG_FILENAME", "flag_Uk.png");
            if (!flag.equals(""))
                jRadioButtonLanguage.setIcon(Utils.getImageIcon(flag));
            // Set selected or not
            if (languages[i].equals(Language.getLanguage()))
                jRadioButtonLanguage.setSelected(true);
            else
                jRadioButtonLanguage.setSelected(false);
            // Add actions
            jRadioButtonLanguage.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    // Change Language
                    Language.setLanguage(e.getActionCommand());

                    // Update mainFrame
                    mainFrame.dispose();

                    try {
                        new MainFrame();
                    } catch (Exception ex) {
                        Local.err.printException(ex);
                    }
                }
            });
            jMenuLanguage.add(jRadioButtonLanguage);
        }

        // Create help menu
        JMenu jMenuHelp = new JMenu(Language.FILEMENU_HELP);
        JMenuItem jMenuHelpAbout = new JMenuItem(Language.FILEMENU_ABOUT);
        jMenuHelpAbout.setIcon(Utils.getImageIcon("about16.gif"));
        jMenuHelpAbout.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                AboutDialog dlg = new AboutDialog(mainFrame);
                Utils.showDialogBox(dlg);
            }
        });

        // Link menus
        JMenuBar jMenuBar = new JMenuBar();
        jMenuBar.add(jMenuFile);
        {
            jMenuFile.add(jMenuFileExit);
        }
        jMenuBar.add(jMenuLanguage);
        {
            // Languages already linked
        }
        jMenuBar.add(jMenuHelp);
        {
            jMenuHelp.add(jMenuHelpAbout);
        }
        return jMenuBar;
    }


    private void setSize() {
        Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
        int width = (int) (screen.getWidth() * 9 / 10);
        int height = (int) (screen.getHeight() * 8 / 10);
        Dimension tool = new Dimension(width, height);
        ((JPanel)this.getContentPane()).setPreferredSize(tool);
        this.setSize(tool);
    }


    private JPanel createButtonPanel() {
        jButton_Process.setPreferredSize(new Dimension(90, 25));
        jButton_Close.setPreferredSize(new Dimension(90, 25));

        JPanel jPanel_Buttons = new JPanel();
        jPanel_Buttons.setLayout(new GridBagLayout());
        jPanel_Buttons.add(jButton_Process, new GridBagConstraints(0, 0, 1, 1, 0.0, 0.0,
                GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(5, 5, 5, 5), 0, 0));
        jPanel_Buttons.add(jButton_Close, new GridBagConstraints(1, 0, 1, 1, 0.0, 0.0,
                GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(5, 5, 5, 5), 0, 0));
        return jPanel_Buttons;
    }


    private BrowsePanel createBrowsePanel(String label, String tip, String id) {
        BrowsePanel browsePanel = new BrowsePanel(label, true, id);
        browsePanel.textField.setEditable(false);
        browsePanel.textField.setBackground(Color.WHITE);
        browsePanel.label.setToolTipText(tip);
        browsePanel.button.setMinimumSize(new Dimension(120, 25));
        browsePanel.label.setMinimumSize(new Dimension(150, 25));
        browsePanel.setFilterTypes(new byte[] {FileBrowser.FILTER_ACCEPT_XML_FILES, FileBrowser.FILTER_ACCEPT_ALL_FILES});
        browsePanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
        return browsePanel;
    }


    private void initialyseOutputDirectory() {
        if (outputDirectory.equals("")) {
            File directory = options.getDirectory();
            if (directory == null)
                outputDirectory = System.getProperty("user.dir");
            else {
                outputDirectory = directory.getPath();
                if (!directory.exists())
                    directory.mkdir();
            }
            outputDirectory += Local.FILE_SEPARATOR;
        }
    }


    private Document preProcess(Document domDocument, FileObject file) {
        if (options.runPreprocess()) {
            if (options.logInformation())
                Local.out.setOut(outputDirectory + Local.FILE_LOG_PREPROCESS.getName(), true, false);
            PreProSyntacticHeuristics ppsh = new PreProSyntacticHeuristics();
            ppsh.process(domDocument, file.getName());
            domDocument = ppsh.getDomDocument();
        }
        return domDocument;
    }


    private Document resolveAnaphors(Document domDocument, int queryCounter, svm_model svmModel, SVMScalingModel svmScalingModel) {
        if (options.logInformation())
            Local.out.setOut(outputDirectory + Local.FILE_LOG_PROCESS.getName(), true, false);
        DiscourseModelImplementer dmi = new DiscourseModelImplementer(domDocument);
        dmi.setClassifier(queryCounter, svmModel, svmScalingModel);
        dmi.process();
        return dmi.getDOMDocument();
    }


    private void evaluate(Document domDocument, FileObject file, boolean isLast) {
        if (options.runEvaluation()) {
            Local.out.setOut(outputDirectory + Local.FILE_LOG_EVALUATION.getName(), true, true);
            AnaphoraResolutionEvaluation are = new AnaphoraResolutionEvaluation();
            if (options.parsedEvaluation()) {
                File parsedFile = new File(file.getParent() + Local.FILE_SEPARATOR + "Parsed" + file.getName());
                if (!parsedFile.exists()) {
                    System.err.println(Language.ERROR_PARSEDFILENOTFOUND + ":" + Local.NEWLINE + parsedFile.getPath());
                    System.exit(1);
                }
                Document parsedDomDocument = IOXMLUtils.load(parsedFile.getPath());
                if (parsedDomDocument == null)
                    System.exit(1);
                are.process(domDocument, file.getName(), parsedDomDocument, parsedFile.getName(), options.evaluationTypes());
            }
            are.process(domDocument, file.getName(), options.evaluationTypes());
            if (isLast)
                are.displayPerformance();
        }

    }


    private void process(ActionEvent e) {
        if (isReady()) {
            // Load gnmvpc.libsvm.* files
            svm_model svmModel = null;
            SVMScalingModel svmScalingModel = null;
            try {
                svmModel = svm.svm_load_model("gnmvpc.libsvm" + ".model"); //modelFileName = "gnmvpc.libsvm"
                svmScalingModel = new SVMScalingModel("gnmvpc.libsvm" + ".range");
            } catch (FileNotFoundException fnfe) {
                Local.err.printException(fnfe);
                System.exit(1);
            } catch (IOException ioe) {
                Local.err.printException(ioe);
                System.exit(1);
            }

            results.clearFiles();
            int queryCounter = 0;
            for (Iterator files = filesPanel.getSelectedFiles().iterator(); files.hasNext(); ) {
                FileObject file = (FileObject) files.next();
                file.setHtml(new File(outputDirectory + file.getHtml().getName()));
                File processed = new File(outputDirectory + file.getProcessed().getName());

                // Load XML files
                Document domDocument = IOXMLUtils.load(file.getPath());

                //Producing MASXML files
                domDocument = preProcess(domDocument, file);

                // Running GuiTAR on MASXML files...
                domDocument = resolveAnaphors(domDocument, queryCounter, svmModel, svmScalingModel);
                queryCounter++;

                // Running evaluation...
                evaluate(domDocument, file, !files.hasNext());

                // Producing html files for error analysis...
                Local.out.setOut(file.getHtml(), false, true);
                (new GenerateErrorsInHTML()).transformDomDocument(domDocument, file.getName(), "AAante", "ante");

                // Return DOM document to XML form
                IOXMLUtils.save(domDocument, processed.getPath());
                file.setProcessed(processed);

                // Close html file output
                Local.out.toDefault(false);

                // Add file to result panel
                results.addFile(file);
            }
            jTabbedPane.setSelectedIndex(1);
        }
    }


    private boolean isReady() {
        // Initialise output directory
        initialyseOutputDirectory();

        // Set inifile if any
        File file = options.getTagSet();
        if (file != null)
            TagSet.setTagSet(file);

        // Set verbose mode
        Local.out.setVerbose(options.logInformation());

        // Set log types
        Local.out.setLogTypes(options.logTypes());

        // Check if files fave been selected
        if (filesPanel.getSelectedFiles().size() < 1)
            return false;

        return true;
    }


    // ===========================================================================
    // INTERFACE IMPLEMENTATIONS
    public void addListeners() {
        jButton_Close.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                System.exit(0);
            }
        });
        jButton_Process.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                process(e);
            }
        });
    }


    public void initialiseGUI() throws Exception {
        this.setTitle("GuiTAR");
        this.setJMenuBar(createMenu());
        this.setSize();
        this.setIconImage(Utils.getImageIcon("guitar_small.png").getImage());

        filesPanel = createBrowsePanel(Language.BROWSEPANEL_XMLFILES, Language.BROWSEPANEL_XMLFILES_TIP, "XML files");

        options.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
        results.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

        jTabbedPane.add(options, Language.LABEL_OPTION);
        jTabbedPane.add(results, Language.LABEL_RESULT);

        JPanel contentPane = (JPanel)this.getContentPane();
        contentPane.setMinimumSize(new Dimension(700, 600));
        contentPane.setLayout(new GridBagLayout());
        contentPane.add(filesPanel, new GridBagConstraints(0, 0, 1, 1, 1.0, 0.0,
                GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL, new Insets(5, 5, 5, 5), 0, 0));
        contentPane.add(jTabbedPane, new GridBagConstraints(0, 1, 1, 1, 1.0, 1.0,
                GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(5, 5, 5, 5), 0, 0));
        contentPane.add(createButtonPanel(), new GridBagConstraints(0, 2, 1, 1, 0.0, 0.0,
                GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(5, 5, 5, 5), 0, 0));
    }


    public String getID() {
        return ID;
    }


    public void initialiseHistory() {
    }


    public void initialiseLanguage() {
        jButton_Process.setText(Language.BUTTON_PROCESS);
        jButton_Process.setToolTipText(Language.BUTTON_PROCESS_TIP);
        jButton_Close.setText(Language.BUTTON_CLOSE);
        jButton_Close.setToolTipText(Language.BUTTON_CLOSE_TIP);
    }

}
