/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR.prepro; //General Tool for Anaphora Resolution

import java.io.*;
import java.util.*;

import org.w3c.dom.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.*;
import uk.ac.essex.malexa.nlp.dp.GuiTAR.util.*;

/**
 * A class that implements heuristics for identifying NP-type, NP-AgreementFeatures, NP-head, and NP-modifiers.
 * @author Mijail A. Kabadjov
 * @author Philippe J. Goux
 * @version 1.2
 */
public class PreProSyntacticHeuristics {
    //Instance variables
    private Document domDocument;
    private int headCounter, modifierCounter;

    /**
     * This the main method which triggers various other methods for identifying and setting NP syntactic features.
     * @param inputFileName The name of the file to be processed
     */
    public void processFile(String inputFileName) {
        //Load XML document
        Document domDoc = IOXMLUtils.load(inputFileName);

        //Launch process
        process(domDoc, inputFileName);

        //Store the result
        IOXMLUtils.save(domDocument, "masxml." + inputFileName); //Store modified DOM Tree to file
    }


    /**
     * This the main method which triggers various other methods for identifying and setting NP syntactic features.
     * @param domDoc Dom Document of the XML to process
     * @param inputFileName The name of the file to be processed (Used in log only)
     */
    public void process(Document domDoc, String inputFileName) {
        headCounter = 1;
        modifierCounter = 1;
        domDocument = domDoc;
        NodeList nounPhrases = domDocument.getElementsByTagName(IOXMLUtils.NP_TAG);
        Local.out.println("PreProSyntacticHeuristics", "");
        Local.out.println("PreProSyntacticHeuristics", "File: " + inputFileName);
        Local.out.println("PreProSyntacticHeuristics", "Number of NEs: " + nounPhrases.getLength());
        //Process Cfs
        for (int i = 0; i < nounPhrases.getLength(); i++) {
            Node np = nounPhrases.item(i);
            Node head = findHead(np);
            if (head == null) {
                /*** COORD_NPs do not have head so this type is never used?! ***/
                Local.out.println("PreProSyntacticHeuristics", "Null head for NP: " + ((Element) np).getAttribute("id"));
                byte npType = setType(np, head); //This will create empty attribute
                setAgreementFeatures(np, head, npType); //This will create empty attributes
                continue;
            }
            if (!head.getNodeName().equalsIgnoreCase(IOXMLUtils.NP_HEAD_TAG))
                head = setHead(np, head);
            NodeList wordList = ((Element) head).getElementsByTagName(IOXMLUtils.WORD_TAG);
            //Only last word is taken as the head (if more than one words)
            Node headWordNode = wordList.item(wordList.getLength() - 1);
            byte npType = setType(np, headWordNode);
            setAgreementFeatures(np, headWordNode, npType);
            setModifiers(np, head);
        }
    }


    /**
     * A method that identifies the head of a Noun Phrase. Essentially the rightmost
     * common or proper noun, or personal, possessive or reflexive pronoun, or determiner this/these (if any)
     * of the NP is considered to be its head.
     * @param node The DOM node that holds the noun phrase to be processesed
     * @return Node The DOM node holding the head of the noun phrase
     */
    public Node findHead(Node node) {
        Node result = null;
        //Get all words not in embedded nes
        Vector wordNodes = IOXMLUtils.getElementsByTagNameNoEmbedding(node, node.getNodeName(), IOXMLUtils.NP_HEAD_TAG);
        if (wordNodes != null && wordNodes.size() > 0)
            return ((Node) wordNodes.elementAt(0)); //Though check what to do with double heads
        wordNodes = IOXMLUtils.getElementsByTagNameNoEmbedding(node, node.getNodeName(), IOXMLUtils.WORD_TAG);
        int numberOfWords = wordNodes.size();
        if (numberOfWords == 1) {
            Node word = (Node) wordNodes.elementAt(0);
            String p = ((Element) word).getAttribute(TagSet.POS_PROPERTY_NAME);
            if (Local.containsIgnoreCase(TagSet.PERSONAL_PRONOUN, p) ||
                Local.containsIgnoreCase(TagSet.PERSONAL_PRONOUN_PLURAL, p) ||
                Local.containsIgnoreCase(TagSet.PERSONAL_PRONOUN_OBJECT, p) ||
                Local.containsIgnoreCase(TagSet.POSSESSIVE_PRONOUN, p) ||
                Local.containsIgnoreCase(TagSet.REFLEXIVE_PRONOUN, p) ||
                Local.containsIgnoreCase(TagSet.REFLEXIVE_PRONOUN_PLURAL, p) ||
                Local.containsIgnoreCase(TagSet.DETERMINER, p) ||
                Local.containsIgnoreCase(TagSet.DETERMINER2, p))
                return word;
        }
        Node pronoun = null;
        for (int i = numberOfWords - 1; i >= 0; i--) { //go backwards
            Node word = (Node) wordNodes.elementAt(i);
            String p = ((Element) word).getAttribute(TagSet.POS_PROPERTY_NAME);
            if (Local.containsIgnoreCase(TagSet.COMMON_NOUN_SINGULAR, p) ||
                Local.containsIgnoreCase(TagSet.COMMON_NOUN_PLURAL, p) ||
                Local.containsIgnoreCase(TagSet.PROPER_NOUN_SINGULAR, p) ||
                Local.containsIgnoreCase(TagSet.PROPER_NOUN_PLURAL, p)) {
                result = word;
                break;
            } else if (Local.containsIgnoreCase(TagSet.PERSONAL_PRONOUN, p) ||
                       Local.containsIgnoreCase(TagSet.PERSONAL_PRONOUN_PLURAL, p) ||
                       Local.containsIgnoreCase(TagSet.PERSONAL_PRONOUN_OBJECT, p))
                pronoun = word;
            else if (pronoun != null) {
                result = pronoun;
                break;
            }
        }
        return result;
    }


    /**
     * Sets the NP-type (i.e. the-np, pers-pro, etc.).
     * @param npNode The DOM node that holds the noun phrase to be processesed
     * @param headNode The head of the given noun phrase
     * @return byte The type (if any) to which this NP was set
     */
    private byte setType(Node npNode, Node headNode) {
        byte type = -1;
        if (headNode != null) {
            type = findType(npNode, headNode);
            String strType = NominalGroup.getStringType(type);
            ((Element) npNode).setAttribute(IOXMLUtils.CATEGORY_PROPERTY_NAME, strType);
        } else
            ((Element) npNode).setAttribute(IOXMLUtils.CATEGORY_PROPERTY_NAME, "");
        return type;
    }


    /**
     * Sets the agreement features (i.e. person, number, and gender).
     * @param npNode The DOM node that holds the noun phrase to be processesed
     * @param headNode The head of the given noun phrase
     * @param npType The type of the NP - only used to set 'coord-np' to plural
     */
    private void setAgreementFeatures(Node npNode, Node headNode, byte npType) {
        if (headNode != null) {
            String pos = ((Element) headNode).getAttribute(TagSet.POS_PROPERTY_NAME);
            String strHead = headNode.getFirstChild().getNodeValue().trim().toLowerCase();
            Agreement agreement = Agreement.newInstance(pos, strHead);
            String person = "";
            int per = agreement.getPerson();
            if (per == Agreement.FIRSTPERSON) person = IOXMLUtils.FIRSTPERSON_PROPERTY_VALUE;
            else if (per == Agreement.SECONDPERSON) person = IOXMLUtils.SECONDPERSON_PROPERTY_VALUE;
            else if (per == Agreement.THIRDPERSON) person = IOXMLUtils.THIRDPERSON_PROPERTY_VALUE;
            else person = IOXMLUtils.ANY_PROPERTY_VALUE;
            String number = "";
            int num = agreement.getNumber();
            if (num == Agreement.SINGULAR) number = IOXMLUtils.SINGULAR_PROPERTY_VALUE;
            else if (num == Agreement.PLURAL) number = IOXMLUtils.PLURAL_PROPERTY_VALUE;
            else number = IOXMLUtils.ANY_PROPERTY_VALUE;
            String gender = "";
            int gen = agreement.getGender();
            if (gen == Agreement.FEMENINE) gender = IOXMLUtils.FEMENINE_PROPERTY_VALUE;
            else if (gen == Agreement.MASCULINE) gender = IOXMLUtils.MASCULINE_PROPERTY_VALUE;
            else if (gen == Agreement.IMPERSONAL) gender = IOXMLUtils.IMPERSONAL_PROPERTY_VALUE;
            else gender = IOXMLUtils.ANY_PROPERTY_VALUE;
            ((Element) npNode).setAttribute(IOXMLUtils.PERSON_PROPERTY_NAME, person);
            ((Element) npNode).setAttribute(IOXMLUtils.NUMBER_PROPERTY_NAME, number);
            ((Element) npNode).setAttribute(IOXMLUtils.GENDER_PROPERTY_NAME, gender);
        } else {
            ((Element) npNode).setAttribute(IOXMLUtils.PERSON_PROPERTY_NAME, "");
            ((Element) npNode).setAttribute(IOXMLUtils.NUMBER_PROPERTY_NAME, "");
            ((Element) npNode).setAttribute(IOXMLUtils.GENDER_PROPERTY_NAME, "");
        }
    }


    /**
     * Marks-up the head of a given NP.
     * @param npNode The DOM node that holds the noun phrase to be processesed
     * @param headNode The head of the given noun phrase
     * @return Node The newly created DOM Node that holds the NP head
     */
    private Node setHead(Node npNode, Node headNode) {
        //headNode is a word node
        Node parent = headNode.getParentNode();
        Element npHead = domDocument.createElement(IOXMLUtils.NP_HEAD_TAG);
        Node oldNode = parent.replaceChild(npHead, headNode);
        npHead.appendChild(oldNode);
        npHead.setAttribute(IOXMLUtils.ID_PROPERTY_NAME, "AAh" + headCounter);
        headCounter++;
        return npHead;
    }


    /**
     * Marks-up the modifiers of a given NP.
     * @param npNode The DOM node that holds the noun phrase to be processesed
     * @param headNode The head of the given noun phrase
     */
    private void setModifiers(Node npNode, Node headNode) {
        //headNode could be nphead node or word node
        Vector wordList = IOXMLUtils.getElementsByTagNameNoEmbedding(npNode,
                npNode.getNodeName(), IOXMLUtils.NP_MODIFIER_TAG);
        if (wordList.size() == 0) { //if mod not marked-up yet
            if (headNode != null) {
                //In case headNode is embedded into special tag (i.e. <emph> )
                Node parent = headNode.getParentNode();
                Node refNode = headNode;
                while (parent != npNode) {
                    refNode = parent;
                    parent = parent.getParentNode();
                }
                Vector modifiers = getPreModifiers(npNode, refNode);
                if (modifiers.size() > 0) markUpModifiers(modifiers, refNode, true);
                modifiers = getPostModifiers(npNode, refNode);
                if (modifiers.size() > 0) {
                    refNode = refNode.getNextSibling();
                    markUpModifiers(modifiers, refNode, false);
                }
            } //end if head not null
        }
    }


    /**
     * Wraps up the original DOM nodes holding modifiers within a <mod> node.
     * @param modifiers A Vector of DOM nodes holding the modifiers
     * @param refNode A DOM node before which the new <mod> node will be inserted
     * @param premod True if marking premodifiers, false otherwise
     */
    private void markUpModifiers(Vector modifiers, Node refNode, boolean premod) {
        Node parent = refNode.getParentNode();
        Element npModifier = domDocument.createElement(IOXMLUtils.NP_MODIFIER_TAG);
        parent.insertBefore(npModifier, refNode);
        for (int i = 0; i < modifiers.size(); i++) {
            Node mod = (Node) modifiers.elementAt(i);
            //parent = mod.getParentNode();
            //Node oldNode = parent.remove( mod ); Maybe not needed
            npModifier.appendChild(mod);
        }
        npModifier.setAttribute(IOXMLUtils.ID_PROPERTY_NAME, "AAm" + modifierCounter);
        if (premod) npModifier.setAttribute(IOXMLUtils.CATEGORY_PROPERTY_NAME, "AApre");
        else npModifier.setAttribute(IOXMLUtils.CATEGORY_PROPERTY_NAME, "AApost");
        modifierCounter++;
    }


    /**
     * Gets the premodifiers of a Noun Phrase. Starts from the NP head goes to the left, sibling by sibling, storing
     * every node on the way, except determiners.
     * @param npNode The DOM node that holds the noun phrase to be processesed
     * @param headNode The head of the given noun phrase
     * @return Vector The Vector of DOM nodes holding the premodifiers
     */
    private Vector<Node> getPreModifiers(Node npNode, Node headNode) {
        Vector<Node> preModifiers = new Vector<Node>();
        Node node = headNode;
        while ((node = node.getPreviousSibling()) != null) {
            String pos = ((Element) node).getAttribute(TagSet.POS_PROPERTY_NAME);
            /*** A more extensive list of what is not a premodifier is needed ***/
            if (!(node instanceof Element) ||
                (node.getNodeName().equalsIgnoreCase(IOXMLUtils.WORD_TAG) &&
                 (Local.containsIgnoreCase(TagSet.DETERMINER, pos) ||
                  Local.containsIgnoreCase(TagSet.DETERMINER2, pos))))
                break; //continue; skip determiner
            preModifiers.add(0, node);
        }
        return preModifiers;
    }


    /**
     * Gets the postmodifiers of a Noun Phrase. First it gets the DOM node holding the NP, then retrieves
     * all the nodes holding words and finally stores all the words after the head as postmodifiers.
     * @param npNode The DOM node that holds the noun phrase to be processesed
     * @param headNode The head of the given noun phrase
     * @return Vector The Vector of DOM nodes holding the postmodifiers
     */
    private Vector<Node> getPostModifiers(Node npNode, Node headNode) {
        Vector<Node> postModifiers = new Vector<Node>();
        Node node = headNode;
        while ((node = node.getNextSibling()) != null) {
            if (!(node instanceof Element))continue;
            postModifiers.add(node);
        }
        return postModifiers;
    }


    /**
     * A method that identifies the type of a Noun Phrase, depending on its structure,
     * and words and their part-of-speech.
     * @param node The DOM node that holds the noun phrase to be processesed
     * @param headNode The head of the given noun phrase
     * @return byte The type of the noun phrase (the-np, a-np, etc.)
     */
    public byte findType(Node node, Node headNode) {
        byte type = -1;
        //word to type mappings, eg. the -> the-np
        Map determinerMap = createDeterminerMap();
        Map determinerNoHeadMap = createDeterminerNoHeadMap();
        Map adjectiveMap = createAdjectiveMap();
        Map adjectiveSuperlativeMap = createAdjectiveSuperlativeMap();
        //Get the words (Nodes) that constitute the NP
        //Vector wordNodes = IOXMLUtils.getElementsByTagNameNoEmbedding( node, node.getNodeName(), IOXMLUtils.WORD_TAG );
        Set<String> tagsToIgnore = new HashSet<String>();
        tagsToIgnore.add(IOXMLUtils.NP_TAG); //ne
        tagsToIgnore.add(IOXMLUtils.UNIT_TAG); //unit
        Vector wordNodes = IOXMLUtils.getElementsByTagNameNoEmbedding(node, tagsToIgnore, IOXMLUtils.WORD_TAG);
        String headPOS = ((Element) headNode).getAttribute(TagSet.POS_PROPERTY_NAME);
        //First try some straightforward ones
        if (Local.containsIgnoreCase(TagSet.PERSONAL_PRONOUN, headPOS) ||
            Local.containsIgnoreCase(TagSet.PERSONAL_PRONOUN_PLURAL, headPOS) ||
            Local.containsIgnoreCase(TagSet.PERSONAL_PRONOUN_OBJECT, headPOS)) {
            type = NominalGroup.PERS_PRO;
            return type;
        } else if (Local.containsIgnoreCase(TagSet.POSSESSIVE_PRONOUN, headPOS)) {
            type = NominalGroup.POSS_PRO;
            return type;
        }
        boolean headIsCommonNoun = (Local.containsIgnoreCase(TagSet.COMMON_NOUN_SINGULAR, headPOS) ||
                                    Local.containsIgnoreCase(TagSet.COMMON_NOUN_PLURAL, headPOS));
        int numberOfWords = wordNodes.size();
        for (int i = 0; i < numberOfWords; i++) {
            Element wordNode = (Element) wordNodes.elementAt(i);
            String pos = wordNode.getAttribute(TagSet.POS_PROPERTY_NAME);
            String word = wordNode.getFirstChild().getNodeValue().toLowerCase();
            //Conditions at the part-of-speech level
            if (Local.containsIgnoreCase(TagSet.DETERMINER, pos) ||
                Local.containsIgnoreCase(TagSet.DETERMINER2, pos)) {
                if (headIsCommonNoun) type = getTypeMapping(word, determinerMap);
                else type = getTypeMapping(word, determinerNoHeadMap);
                if (type == NominalGroup.BARE_NP) type = -1; //the determiner is not in the map
                //break; no exit in order to catch the-pn
            } else if (Local.containsIgnoreCase(TagSet.CARDINAL, pos)) {
                if (type == -1) type = NominalGroup.NUM_NP;
                break; //exit loop
            } else if (Local.containsIgnoreCase(TagSet.ADJECTIVE_GENERAL, pos)) {
                /*** WHAT HAPPENS WHEN THERE IS NO HEAD, EG THE BEAUTIFULLY DECORATED ***/
                if (type == -1) type = getTypeMapping(word, adjectiveMap);
                break; //exit loop
            } else if (Local.containsIgnoreCase(TagSet.ADJECTIVE_SUPERLATIVE, pos)) {
                if (type == -1) type = getTypeMapping(word, adjectiveSuperlativeMap);
                break; //exit loop
            } else if (Local.containsIgnoreCase(TagSet.COMMON_NOUN_SINGULAR, pos) ||
                       Local.containsIgnoreCase(TagSet.COMMON_NOUN_PLURAL, pos)) {
                if (type == -1) type = NominalGroup.BARE_NP; //if type not set yet
                //break; //exit loop Perhaps no break here, in case a PN follows the noun
            } else if (Local.containsIgnoreCase(TagSet.PROPER_NOUN_SINGULAR, pos) ||
                       Local.containsIgnoreCase(TagSet.PROPER_NOUN_PLURAL, pos)) {
                if (type == -1 || type == NominalGroup.BARE_NP) type = NominalGroup.PN;
                else if (type == NominalGroup.THE_NP && !headIsCommonNoun) type = NominalGroup.THE_PN;
                break; //exit loop
            } else if (Local.containsIgnoreCase(TagSet.CON_COORDINATING, pos)) {
                //else if( Local.contains( LexicalWord.CON_COORDINATING ) || Local.contains( LexicalWord.COMMA )) {
                if (type == -1) type = NominalGroup.COORD_NP;
                break; //exit loop
            } else if (Local.containsIgnoreCase(TagSet.WH_PRONOUN, pos)) {
                if (type == -1) type = NominalGroup.WH_PRO;
                break; //exit loop
            }
        } //end for

        //Check for possessives
        if (headNode != null) {
            Element previousNE = (Element) IOXMLUtils.getPreviousNodeByTagName(headNode, IOXMLUtils.NP_TAG, node);
            if (previousNE != null) {
                NodeList nl = previousNE.getElementsByTagName(IOXMLUtils.WORD_TAG);
                int n = nl.getLength();
                Element wordNode = (Element) nl.item(n - 1);
                if (wordNode != null) {
                    String pos = wordNode.getAttribute(TagSet.POS_PROPERTY_NAME);
                    if (Local.containsIgnoreCase(TagSet.POSSESSIVE, pos) ||
                        Local.containsIgnoreCase(TagSet.POSSESSIVE_PRONOUN, pos))
                        type = NominalGroup.POSS_NP;
                }
            }
        }

        return type;
    }


    /**
     * An auxiliary method that wraps a general functionality for retrieving a type of an NP
     * associated with a given word. If no mapping is found for the word the type is set to bare-np.
     * @param word The key word
     * @param wordToTypeMap The map to be searched
     * @return byte The type of the NP containing the input word
     */
    private byte getTypeMapping(String word, Map wordToTypeMap) {
        byte type = -1;
        Byte preType = (Byte) wordToTypeMap.get(word);
        if (preType != null) type = preType.byteValue();
        else type = NominalGroup.BARE_NP;
        return type;
    }


    /**
     * An auxiliary method that creates word-to-type mappings for a determiner. The mappings are hard-coded.
     * @return Map The map containing the corresponding associations
     */
    private Map<String, Byte> createDeterminerMap() {
        Map<String, Byte> theMap = new HashMap<String, Byte>();
        theMap.put(new String("the"), new Byte(NominalGroup.THE_NP));
        theMap.put(new String("that"), new Byte(NominalGroup.THAT_NP));
        theMap.put(new String("those"), new Byte(NominalGroup.THAT_NP));
        theMap.put(new String("this"), new Byte(NominalGroup.THIS_NP));
        theMap.put(new String("these"), new Byte(NominalGroup.THIS_NP));
        theMap.put(new String("a"), new Byte(NominalGroup.A_NP));
        theMap.put(new String("an"), new Byte(NominalGroup.A_NP));
        theMap.put(new String("another"), new Byte(NominalGroup.ANOTHER_NP));
        theMap.put(new String("any"), new Byte(NominalGroup.Q_NP));
        theMap.put(new String("both"), new Byte(NominalGroup.Q_NP));
        theMap.put(new String("each"), new Byte(NominalGroup.Q_NP));
        theMap.put(new String("every"), new Byte(NominalGroup.Q_NP));
        theMap.put(new String("no"), new Byte(NominalGroup.Q_NP));
        theMap.put(new String("some"), new Byte(NominalGroup.Q_NP));
        return theMap;
    }


    /**
     * An auxiliary method that creates word-to-type mappings for a determiner when the NP has no head.
     * The mappings are hard-coded.
     * @return Map The map containing the corresponding associations
     */
    private Map<String, Byte> createDeterminerNoHeadMap() {
        Map<String, Byte> theMap = new HashMap<String, Byte>();
        theMap.put(new String("the"), new Byte(NominalGroup.THE_NP));
        theMap.put(new String("that"), new Byte(NominalGroup.THAT_PRO));
        theMap.put(new String("those"), new Byte(NominalGroup.THAT_PRO));
        theMap.put(new String("this"), new Byte(NominalGroup.THIS_PRO));
        theMap.put(new String("these"), new Byte(NominalGroup.THIS_PRO));
        theMap.put(new String("any"), new Byte(NominalGroup.Q_PRO));
        theMap.put(new String("both"), new Byte(NominalGroup.Q_PRO));
        theMap.put(new String("some"), new Byte(NominalGroup.Q_PRO));
        return theMap;
    }


    /**
     * An auxiliary method that creates word-to-type mappings for an adjective.
     * The mappings are hard-coded.
     * @return Map The map containing the corresponding associations
     */
    private Map<String, Byte> createAdjectiveMap() {
        Map<String, Byte> theMap = new HashMap<String, Byte>();
        theMap.put(new String("such"), new Byte(NominalGroup.SUCH_NP));
        theMap.put(new String("other"), new Byte(NominalGroup.ANOTHER_NP));
        theMap.put(new String("many"), new Byte(NominalGroup.Q_NP));
        return theMap;
    }


    /**
     * An auxiliary method that creates word-to-type mappings for a superlative adjective.
     * The mappings are hard-coded.
     * @return Map The map containing the corresponding associations
     */
    private Map<String, Byte> createAdjectiveSuperlativeMap() {
        Map<String, Byte> theMap = new HashMap<String, Byte>();
        theMap.put(new String("most"), new Byte(NominalGroup.Q_NP));
        return theMap;
    }


    /**
     * Return DOM document modified during ProcessFile() method.
     * @return Document
     */
    public Document getDomDocument() {
        return domDocument;
    }


    /* ******** MAIN METHOD ******** */
    public static void main(String args[]) throws IOException {
        PreProSyntacticHeuristics ppsh = new PreProSyntacticHeuristics();
        ppsh.run(args);
    }


    public void run(String arguments[]) throws IOException {
        Local.out.print("PreProSyntacticHeuristics", ""); // Leave this to make sure the tag set is initialysed
        if (arguments.length < 1) {
            BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
            String fileName = in.readLine();
            while (fileName != null) {
                processFile(fileName);
                fileName = in.readLine();
            }
        } else {
            processFile(arguments[0]);
        } //end if-else
    }
} //end PreProSyntacticHeuristics
