/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR.util;

import java.io.*;
import java.util.*;

/**
 * <b>Description:</b> The IniFile class has been designed to be able to retrieve
 * or set a specific value from a given INI file. These files are text based
 * and can be easily modified by a user. This may be useful for cases in
 * which the user may be required to change that file but it may also cause
 * trouble if that file is corrupted by the user.<br>
 * <b>Copyright:</b> Copyright (c) 2005<br>
 * <b>Company:</b> Essex University<br>
 * @author Philippe J. Goux
 * @version 1.0
 */

public class IniFile {
    // ===========================================================================
    // CONSTANTS
    public static final String NEWLINE = System.getProperty("line.separator");

    protected static final String CATHEGORY_OPEN = "[";
    protected static final String CATHEGORY_END = "]";
    protected static final String COMMENT_TAG = ":";
    protected static final String SEPARATOR = "=";
    protected static final String SEPARATOR_ARRAY = "; ";

    // ===========================================================================
    // VARIABLES
    private boolean createFile = true;
    protected Map<String, Map<String, String>> iniFileContent = new LinkedHashMap<String, Map<String, String>>();
    protected File iniFile = null;


    // ===========================================================================
    // CONSTRUCTORS
    public IniFile(File file, boolean createFile) {
        this.createFile = createFile;
        setIniFile(file);
    }


    public IniFile(File file) {
        this(file, true);
    }


    public IniFile() {
        this(null);
    }


    // ===========================================================================
    // PUBLIC FONCTIONS
    /**
     * Set if the file should be created if not found.
     * @param b boolean
     */
    public void setCreateFile(boolean b) {
        createFile = b;
        if(!iniFile.exists())
            initialise();
    }


    /**
     * Set the file to be used. This function also calls the <code>initialise()</code>
     * method.
     * @param file File
     */
    public void setIniFile(File file) {
        iniFile = file;
        iniFileContent.clear();
        initialise();
    }


    /**
     * Get the value for the given key. If the value is not found in the file,
     * return default value and set it in the file.
     * @param cathegory String
     * @param key String
     * @param defaultValue String
     * @return String
     */
    public String getKeyValue(String cathegory, String key, String defaultValue) {
        // Make sure that file has been read
        if (iniFileContent.size() == 0) {
            // Check if file is readable
            if (!readContent() || !iniFile.exists()) {
                // If error in file, return default value instead of crashing
                if (defaultValue == null)
                    return "";
                else
                    return defaultValue;
            }
        }

        // Retreive value from read information
        Map<String, String> cathegoryContent = iniFileContent.get(cathegory);
        if (cathegoryContent == null)
            cathegoryContent = new HashMap<String, String>();

        // Retreive value from read information
        if (cathegoryContent.containsKey(key)) {
            return cathegoryContent.get(key);
        }
        // If no value retreived, return default value
        else {
            if (defaultValue != null)
                setKeyValue(cathegory, key, defaultValue);
            return defaultValue;
        }
    }


    /**
     * Get the value for the given key. If the value is not found in the file,
     * return default value and set it in the file.
     * @param cathegory String
     * @param key String
     * @param defaultValue String
     * @return String
     */
    public String[] getKeyValues(String cathegory, String key, String[] defaultValue) {
        // Make sure that file has been read
        if (iniFileContent.size() == 0) {
            // Check if file is readable
            if (!readContent() || !iniFile.exists()) {
                // If error in file, return default value instead of crashing
                if (defaultValue == null)
                    return new String[0];
                else
                    return defaultValue;
            }
        }

        // Retreive value from read information
        Map<String, String> cathegoryContent = iniFileContent.get(cathegory);
        if (cathegoryContent == null)
            cathegoryContent = new HashMap<String, String>();

        // Retreive value from read information
        if (cathegoryContent.containsKey(key)) {
            return cathegoryContent.get(key).split(SEPARATOR_ARRAY);
        }
        // If no value retreived, return default value
        else {
            if (defaultValue != null)
                setKeyValues(cathegory, key, defaultValue);
            return defaultValue;
        }
    }


    /**
     * Set the value for the given key in the INI file.
     * @param cathegory String
     * @param key String
     * @param value String
     * @return boolean
     */
    public boolean setKeyValue(String cathegory, String key, String value) {
        // Make sure that file has been read
        if (iniFileContent.size() == 0) {
            // Check if file is readable
            if (!readContent())
                // If file is unreadable, return false
                return false;
        }

        // Set value to file
        // Try to find existing value corresponding to key within file
        Map<String, String> cathegoryContent = iniFileContent.get(cathegory);
        if (cathegoryContent == null)
            cathegoryContent = new HashMap<String, String>();

        if (cathegoryContent.containsKey(key)) {
            cathegoryContent.remove(key);
        }

        // remove old values and replace them by the new
        cathegoryContent.put(key, value);
        iniFileContent.remove(cathegory);
        iniFileContent.put(cathegory, cathegoryContent);

        // Modify file
        return writeContent(cathegory, key, value);
    }


    /**
     * Set the value for the given key in the INI file
     * @param cathegory String
     * @param key String
     * @param values String
     * @return boolean
     */
    public boolean setKeyValues(String cathegory, String key, String[] values) {
        // Make sure that file has been read
        if (iniFileContent.size() == 0) {
            // Check if file is readable
            if (!readContent())
                // If file is unreadable, return false
                return false;
        }

        // Set value to file
        // Try to find existing value corresponding to key within file
        Map<String, String> cathegoryContent = iniFileContent.get(cathegory);
        if (cathegoryContent == null)
            cathegoryContent = new HashMap<String, String>();

        if (cathegoryContent.containsKey(key)) {
            cathegoryContent.remove(key);
        }

        // If value is new, add it to list
        String value = "";
        for (int j = 0; j < values.length; j++)
            value += values[j] + (((j + 1) < values.length) ? SEPARATOR_ARRAY : "");
        cathegoryContent.put(key, value);
        iniFileContent.remove(cathegory);
        iniFileContent.put(cathegory, cathegoryContent);

        // Modify file
        return writeContent(cathegory, key, value);
    }


    /**
     * Return an array of all the cathegories within the ini file.
     * @return String[]
     */
    public String[] getCathegories() {
        // Make sure that file has been read
        if (iniFileContent.size() == 0) {
            // Check if file is readable
            if (!readContent())
                // If file is unreadable, return false
                return new String[0];
        }

        Set<String> set = iniFileContent.keySet();
        return set.toArray(new String[set.size()]);
    }


    /**
     * Returns an array of the values contained in the given cathegory.
     * @param cathegory String
     * @return String[]
     */
    public String[] getKeys(String cathegory) {
        // Make sure that file has been read
        if (iniFileContent.size() == 0) {
            // Check if file is readable
            if (!readContent())
                // If file is unreadable, return false
                return new String[0];
        }

        // Make sure that cathegory exist
        if (iniFileContent.containsKey(cathegory)) {
            Set<String> set = iniFileContent.get(cathegory).keySet();
            return set.toArray(new String[set.size()]);
        } else
            return new String[0];

    }


    // ===========================================================================
    // PROTECTED FONCTIONS
    /**
     * This function is by default empty. Nevertheless, it may be overriden to
     * be called each time the <code>setIniFile()</code> method is called. This
     * also include the constructor call.
     */
    protected void initialise() {
        /* Do nothing */
    }

    /**
     * Read the content of the file.
     * Returns true if file doesn't exist or if file has been read
     * @return boolean
     */
    protected boolean readContent() {
        try {
            // Create file if it doesn't exist
            if (!iniFile.exists()) {
                if (createFile)
                    iniFile.createNewFile();
                return true;
            }

            // Read file
            BufferedReader inBuff = new BufferedReader(new FileReader(iniFile));

            String cathegory = "", str = "";
            Map<String, String> cathegoryContent = null;
            while ((str = inBuff.readLine()) != null) {
                // If comment or new line
                if (str.startsWith(COMMENT_TAG) || str.equals(""))
                    continue;
                // If cathegory
                else if (str.startsWith(CATHEGORY_OPEN)) {
                    // If cathegory is set then store content
                    if (cathegoryContent != null)
                        iniFileContent.put(cathegory, cathegoryContent);

                    // Get new cathegory
                    cathegory = str.substring(CATHEGORY_OPEN.length(),
                                              str.length() - CATHEGORY_END.length());

                    // Clear list to store new content
                    cathegoryContent = new HashMap<String, String>();
                }
                // If content
                else {
                    if (cathegoryContent != null) {
                        String[] values = str.split(SEPARATOR);
                        if (values.length > 1) {
                            String key = values[0];
                            String value = "";
                            // Make sure to have complete content of line
                            for (int i = 1; i < values.length; i++)
                                value += values[i] + (((i + 1) < values.length) ? SEPARATOR : "");
                            cathegoryContent.put(key, value);
                        }
                    }
                }
            }
            // Don't forget to store last read category
            if (cathegoryContent != null)
                iniFileContent.put(cathegory, cathegoryContent);
            inBuff.close();
        } catch (IOException e) {
            iniFileContent.clear();
            e.printStackTrace();
            return false;
        }
        return true;
    }


    /**
     * Write content of information stored in Map _fileContent to file.
     * @param cathegory String
     * @param key String
     * @param content String
     * @return boolean
     */
    protected boolean writeContent(String cathegory, String key, String content) {
        // Check if file should be created
        if (!createFile)
            return true;

        try {
            // Read content from file to add content to add in correct cathegory
            BufferedReader inBuff = new BufferedReader(new FileReader(iniFile));
            boolean isCathegoryFound = false, isKeyFound = false;
            String str = "", wholeFile = "";
            while ((str = inBuff.readLine()) != null) {
                if (!isCathegoryFound && str.equalsIgnoreCase(CATHEGORY_OPEN + cathegory + CATHEGORY_END)) {
                    wholeFile += ((wholeFile.equals("")) ? "" : NEWLINE) + str;
                    isCathegoryFound = true;
                } else if (isCathegoryFound && str.startsWith(key + SEPARATOR)) {
                    wholeFile += NEWLINE + key + SEPARATOR + content;
                    isKeyFound = true;
                } else if (isCathegoryFound && !isKeyFound && str.startsWith(CATHEGORY_OPEN)) {
                    while(wholeFile.substring(wholeFile.length() - NEWLINE.length(), wholeFile.length()).equals(NEWLINE)) {
                        wholeFile = wholeFile.substring(0, wholeFile.length() - NEWLINE.length());
                    }
                    wholeFile = wholeFile + NEWLINE +
                                key + SEPARATOR + content + NEWLINE + NEWLINE + str;
                    isKeyFound = true;
                } else
                    wholeFile += ((wholeFile.equals("")) ? "" : NEWLINE) + str;
            }
            inBuff.close();

            // Write in file
            BufferedWriter outBuff = new BufferedWriter(new FileWriter(iniFile));
            // If no information was found in file, create new cathegory and content
            if (wholeFile.equals("")) {
                wholeFile += CATHEGORY_OPEN + cathegory + CATHEGORY_END +
                        NEWLINE + key + SEPARATOR + content;
            } else if (!isCathegoryFound) {
                wholeFile += NEWLINE + NEWLINE + CATHEGORY_OPEN + cathegory + CATHEGORY_END +
                        NEWLINE + key + SEPARATOR + content;
            } else if (!isKeyFound) {
                wholeFile += NEWLINE + key + SEPARATOR + content;
            }
            // Write content from whole file
            outBuff.write(wholeFile);
            outBuff.close();
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }
}
