/**************************************************************************
*  GuiTAR - A General Tool for Anaphora Resolution
*  Copyright (C) 2004-2007
*  Mijail Kabadjov, Massimo Poesio, Olivia Sanchez-Graillet, Philippe Goux
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, write to the Free Software Foundation, Inc.,
*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**************************************************************************/

package uk.ac.essex.malexa.nlp.dp.GuiTAR.util;

import java.io.*;
import java.util.*;

/**
 * <b>Description:</b> Enables the GuiTAR user to select which type of traces
 * to have displayed.<br>
 * <b>Copyright:</b> Copyright (c) 2005<br>
 * <b>Company:</b> Essex University
 * @author Philippe J. Goux
 * @version 1.0
 */
public class Log {
    // =========================================================================
    // CONSTANTS
    private final PrintStream DEFAULT;


    // =========================================================================
    // VARIABLES
    private PrintStream printStream;
    private boolean verbose = true;
    private boolean mustVerbose = false;
    private String[] logTypes = new String[0];


    // =========================================================================
    // PRIVATE CONSTRUCTOR
    public Log(PrintStream defaultStream) {
        DEFAULT = defaultStream;
        printStream = defaultStream;
    }


    // =========================================================================
    // PUBLIC METHODS
    public void addLogType(String logType) {
        List<String> types = new ArrayList<String>();
        for (int i = 0; i < logTypes.length; i++) {
            types.add(logTypes[i]);
        }
        types.add(logType.toLowerCase());
        logTypes = (String[]) types.toArray(new String[types.size()]);
    }


    public void clearTypes() {
        logTypes = new String[0];
    }


    public void print(Object str) {
        this.print("", str);
    }


    public void print(String type, Object str) {
        if (isPrintable(type)) {
            print(str, false);
        }
    }


    public void printException(Exception e) {
        StackTraceElement[] elements = e.getStackTrace();
        String[] stackTrace = new String[elements.length];
        String error = "";
        for (int i = 0; i < elements.length; i++) {
            error += elements[i].toString() + Local.NEWLINE;
            stackTrace[i] = elements[i].toString();
        }
        History.storeError(stackTrace);
        print(error, true);
    }


    /**
     * Print an empty string and then terminate the line.  This method behaves as
     * though it invokes <code>{@link #println(java.lang.Object)}</code>
     * with an empty <code>String</code>.
     */
    public void println() {
        this.println("", "");
    }


    public void println(Object str) {
        this.println("", str);
    }


    public void println(String type, Object str) {
        if (isPrintable(type)) {
            print(str, true);
        }
    }


    public void reset() {
        mustVerbose = false;
        this.printStream = DEFAULT;
    }


    public boolean setOut(File file, boolean append, boolean verbose) {
        mustVerbose = verbose;
        try {
            return setOut(new PrintStream(new FileOutputStream(file, append)), verbose);
        } catch (FileNotFoundException ex) {
            return false;
        }
    }


    public boolean setOut(PrintStream printStream, boolean verbose) {
        mustVerbose = verbose;
        this.printStream = printStream;
        return true;
    }


    public boolean setOut(String filename, boolean append, boolean verbose) {
        return setOut(new File(filename), append, verbose);
    }

    public void toDefault(boolean verbose) {
        setOut(DEFAULT, verbose);
    }



    public void setVerbose(boolean v) {
        verbose = v;
    }


    public void setLogTypes(String[] logTypes) {
        this.logTypes = logTypes;
    }


    // =========================================================================
    // PRIVATE METHODS
    protected void print(Object str, boolean newLine) {
        if (newLine)
            printStream.println(str);
        else
            printStream.print(str);
    }


    protected boolean isPrintable(String type) {
        if (mustVerbose)
            return true;
        else if (verbose) {
            if (logTypes.length < 1 || type.length() < 1)
                return true;
            else {
                for (int i = 0; i < logTypes.length; i++) {
                    if (logTypes[i].equalsIgnoreCase(type))
                        return true;
                }
            }
        }
        return false;
    }
}
